
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
open Types;;
open Printer;;

(* ----------------
   union : a' list -> a' list -> a' list
   ---------------- *)
let rec union l1 l2 = match l1 with
    [] -> l2
  | a::reste -> if List.mem a l2
                  then
                    union reste l2
                else
                  union reste (a::l2)
;; 

(* ----------------
   getAllIds : string -> (int*string*string*msg) list -> string list
   ---------------- *)
let rec getAllIds s msgs = 
 let rec getMsgIds msg = match msg with
    Id str -> [str]
  | Int _ -> []
  | Func (str,m) -> union [str] (getMsgIds m) 
  | Crypt (m1,m2) -> union (getMsgIds m1) (getMsgIds m2)
  | Exp (m1,m2) -> union (getMsgIds m1) (getMsgIds m2)
  | Xor (m1,m2) -> union (getMsgIds m1) (getMsgIds m2)
  | Pair (m1,m2) -> union (getMsgIds m1) (getMsgIds m2)
  | Inv str -> [str]
  | TableElt (s1,s2) -> [s1;s2]
  | TableEltInv (s1,s2) -> [s1;s2]
 in match msgs with
   [] -> []
 | (_,_,e,d,m)::reste -> 
	if e=s || d=s then union (getMsgIds m) (getAllIds s reste)
	else getAllIds s reste
;;

(* ----------------
   stack2list : a' stack -> a' list
   ---------------- *)
let stack2list st = 
	let l = ref [] in
	Stack.iter (fun e -> l:=[e]@(!l)) st;
	!l
;;

(* ----------------
   add_secret : string -> (int * goal_type) list -> string
   ---------------- *)
let rec add_secret s goals = match goals with
   [] -> ""
 | (i,Secrecy(m,l))::reste -> if m=s then "secret("^s^"',sec_"^(string_of_int i)^",{"^(String.concat "," l)^"})" else add_secret s reste
 | (_,Authentication(_,_,_))::reste -> add_secret s reste
 | (_,Weak_authentication (_,_,_))::reste -> add_secret s reste
;;

(* ----------------
   add_witness : string -> (int * goal_type) list -> string
   ---------------- *)
let rec add_witness ag nonce goals = match goals with
   [] -> ""
 | (_,Secrecy(_,_))::reste -> add_witness ag nonce reste
 | (i,Authentication(a,b,c))::reste -> if a=ag && c=nonce then "witness("^a^","^b^",auth_"^(string_of_int i)^","^c^"')" else add_witness ag nonce reste
 | (i,Weak_authentication (a,b,c))::reste -> if a=ag && c=nonce then "witness("^a^","^b^",wauth_"^(string_of_int i)^","^c^"')" else add_witness ag nonce reste
;;

(* ----------------
   add_request : string -> (int * goal_type) list -> string
   ---------------- *)
let rec add_request ag nonce goals = match goals with
   [] -> ""
 | (_,Secrecy(_,_))::reste -> add_request ag nonce reste
 | (i,Authentication(a,b,c))::reste -> if b=ag && c=nonce then "request("^b^","^a^",auth_"^(string_of_int i)^","^c^"')" else add_witness ag nonce reste
 | (i,Weak_authentication (a,b,c))::reste -> if b=ag && c=nonce then "wrequest("^b^","^a^",wauth_"^(string_of_int i)^","^c^"')" else add_witness ag nonce reste
;;

(* ----------------
   getTransitions : string -> string list -> (int*string*srting*msg) list -> goal_type list -> (int * string list * string list) list
   ---------------- *)
let getTransitions id vars msgs goals optS = 
  let varNonInit = ref vars and varPrim = ref [] and secretVar = ref [] and  nbNW = ref 1 and nbP = ref 1 in


  let sndMsg m t e d= 
	let resul = ref [] in
	let rec sndMsg_rec msg = (
	match msg with
	  Id s ->
		if List.mem s !varNonInit then (
		  resul := !resul@[s^"':=new()"];
		  (let sec=add_secret s goals in if sec<>"" then (resul := !resul@[sec];secretVar:=!secretVar@[s]));
		  (let wit=add_witness id s goals in if wit<>"" then resul := !resul@[wit]);
		  varNonInit := List.filter (fun e -> e<>s) !varNonInit;
		  varPrim := !varPrim@[s]
		);
		(let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
		if List.mem s !varPrim then s^"'" else s
	 | Int i -> string_of_int i
	 | Func (s,m1) -> 
	     (let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
	     s^"("^(sndMsg_rec m1)^")"
	 | Crypt (m1,m2) -> "{"^(sndMsg_rec m1)^"}_"^(sndMsg_rec m2)
	 | Xor (m1,m2) -> "xor("^(sndMsg_rec m1)^","^(sndMsg_rec m2)^")"
	 | Exp (m1,m2) -> "exp("^(sndMsg_rec m1)^","^(sndMsg_rec m2)^")"
	 | Pair (m1,m2) -> (sndMsg_rec m1)^"."^(sndMsg_rec m2)
	 | Inv s -> 
		if List.mem s !varNonInit then (
		  resul := !resul@[s^"':=new()"];
		  (let sec=add_secret s goals in if sec<>"" then (resul := !resul@[sec];secretVar:=!secretVar@[s]));
		  (let wit=add_witness id s goals in if wit<>"" then resul := !resul@[wit]);
		  varNonInit := List.filter (fun e -> e<>s) !varNonInit;
		  varPrim := !varPrim@[s]
		);
		(let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
		if List.mem s !varPrim then "inv("^s^"')" else "inv("^s^")"
	 | TableElt (t,a) -> 
	    (let sec = add_secret t goals in if optS && not (List.mem t !secretVar) && sec<>"" then (secretVar := !secretVar@[t];resul := !resul@[sec]));
	    if List.mem a !varNonInit then (
		  resul := !resul@[a^"':=new()"];
		  (let sec=add_secret a goals in if sec<>"" then (resul := !resul@[sec];secretVar:=!secretVar@[a]));
		  (let wit=add_witness id a goals in if wit<>"" then resul := !resul@[wit]);
		  varNonInit := List.filter (fun e -> e<>a) !varNonInit;
		  varPrim := !varPrim@[a]
	    );
	    (let sec = add_secret a goals in if optS && not (List.mem a !secretVar) && sec<>"" then (secretVar := !secretVar@[a];resul := !resul@[sec]));
	    t^(if List.mem a !varPrim then "("^a^"')" else "("^a^")")
	 | TableEltInv (t,a) -> 
	    (let sec = add_secret t goals in if optS && not (List.mem t !secretVar) && sec<>"" then (secretVar := !secretVar@[t];resul := !resul@[sec]));
	    if List.mem a !varNonInit then (
		  resul := !resul@[a^"':=new()"];
		  (let sec=add_secret a goals in if sec<>"" then (resul := !resul@[sec];secretVar:=!secretVar@[a]));
		  (let wit=add_witness id a goals in if wit<>"" then resul := !resul@[wit]);
		  varNonInit := List.filter (fun e -> e<>a) !varNonInit;
		  varPrim := !varPrim@[a]
	    );
	    (let sec = add_secret a goals in if optS && not (List.mem a !secretVar) && sec<>"" then (secretVar := !secretVar@[a];resul := !resul@[sec]));
	    "inv("^t^(if List.mem a !varPrim then "("^a^"')" else "("^a^")")^")"
	) in let snd = sndMsg_rec m in match t with
		  Dolev_Yao -> !resul@["SND("^snd^")"]
		| Non_writable -> !resul@["Nonce_"^(string_of_int !nbNW)^"':=new()";
					"Nonce_set_"^e^"_"^d^"':=cons(Nonce_"^(string_of_int !nbNW)^"',Nonce_set_"^e^"_"^d^")";
					"SND("^snd^".Hash_0(Nonce_"^(string_of_int !nbNW)^"'."^snd^"))"]
		| Protected -> !resul@["Key_"^(string_of_int !nbP)^"':=new()";
					"Key_set_"^e^"_"^d^"':=cons(Key_"^(string_of_int !nbP)^"',Key_set_"^e^"_"^d^")";
					"SND({"^snd^"}_Key_"^(string_of_int !nbP)^"')"]

  and rcvMsg m t e d = 
	let resul = ref [] in
	let rec rcvMsg_rec msg = (
	match msg with
	  Id s -> 
		if List.mem s !varNonInit then (
		  varNonInit := List.filter (fun e -> e<>s) !varNonInit;
		  varPrim := !varPrim@[s];
		  (let req=add_request id s goals in if req<>"" then resul := !resul@[req])
		);
		(let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
		if List.mem s !varPrim then s^"'" else s
	 | Int i -> string_of_int i
	 | Func (s,m1) -> 
	    (let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
	    s^"("^(rcvMsg_rec m1)^")"  
	 | Crypt (m1,m2) -> "{"^(rcvMsg_rec m1)^"}_"^(rcvMsg_rec m2)
	 | Xor (m1,m2) -> "xor("^(rcvMsg_rec m1)^","^(rcvMsg_rec m2)^")"
	 | Exp (m1,m2) -> "exp("^(rcvMsg_rec m1)^","^(rcvMsg_rec m2)^")"
	 | Pair (m1,m2) -> (rcvMsg_rec m1)^"."^(rcvMsg_rec m2)
	 | Inv s -> 
		if List.mem s !varNonInit then (
		  varNonInit := List.filter (fun e -> e<>s) !varNonInit;
		  varPrim := !varPrim@[s];
		  (let req=add_request id s goals in if req<>"" then resul := !resul@[req])
		);
		(let sec = add_secret s goals in if optS && not (List.mem s !secretVar) && sec<>"" then (secretVar := !secretVar@[s];resul := !resul@[sec]));
		if List.mem s !varPrim then "inv("^s^"')" else "inv("^s^")"
	 | TableElt (t,a) -> 
	    (let sec = add_secret t goals in if optS && not (List.mem t !secretVar) && sec<>"" then (secretVar := !secretVar@[t];resul := !resul@[sec]));
	    if List.mem a !varNonInit then (
		  varNonInit := List.filter (fun e -> e<>a) !varNonInit;
		  varPrim := !varPrim@[a];
		  (let req=add_request id a goals in if req<>"" then resul := !resul@[req])
		);
	    (let sec = add_secret a goals in if optS && not (List.mem a !secretVar) && sec<>"" then (secretVar := !secretVar@[a];resul := !resul@[sec]));
	    t^"("^(if List.mem a !varPrim then a^"'" else a)^")"
	 | TableEltInv (t,a) -> 
	   (let sec = add_secret t goals in if optS && not (List.mem t !secretVar) && sec<>"" then (secretVar := !secretVar@[t];resul := !resul@[sec]));
	   if List.mem a !varNonInit then (
		  varNonInit := List.filter (fun e -> e<>a) !varNonInit;
		  varPrim := !varPrim@[a];
		  (let req=add_request id a goals in if req<>"" then resul := !resul@[req])
	    );
	    (let sec = add_secret a goals in if optS && not (List.mem a !secretVar) && sec<>"" then (secretVar := !secretVar@[a];resul := !resul@[sec]));
	    "inv("^t^"("^(if List.mem a !varPrim then a^"'" else a)^"))"
	) in let rcv = rcvMsg_rec m in match t with 
		  Dolev_Yao -> (["RCV("^rcv^")"],!resul)
		| Non_writable -> (["in(Nonce_"^(string_of_int !nbNW)^"',Nonce_set_"^e^"_"^d^")";"RCV("^rcv^".Hash_0(Nonce_"^(string_of_int !nbNW)^"'."^rcv^"))"],
				("Nonce_set_"^e^"_"^d^"':=delete(Nonce_"^(string_of_int !nbNW)^"',Nonce_set_"^e^"_"^d^")")::!resul)
		| Protected -> (["in(Key_"^(string_of_int !nbP)^"',Key_set_"^e^"_"^d^")";"RCV({"^rcv^"}_Key_"^(string_of_int !nbP)^"')"],
				("Key_set_"^e^"_"^d^"':=delete(Key_"^(string_of_int !nbP)^"',Key_set_"^e^"_"^d^")")::!resul)




  in let resul = Stack.create() and nbTrans = ref 0 in
  let rec getTransRec msges = match msges with
     [] -> ()
   | (t,n,e,d,m)::reste when e=id -> 
	(if Stack.is_empty resul then ( incr nbTrans;Stack.push (n,["State=0";"RCV(start)"],["State':=1"]@(sndMsg m t e d)) resul )
	 else ( let (i,l1,l2)=Stack.pop resul in Stack.push (i,l1,l2@(sndMsg m t e d)) resul )
	);
	(match t with
	  Dolev_Yao -> ()
	| Non_writable -> incr nbNW
	| Protected -> incr nbP);
	getTransRec reste
   | (t,n,e,d,m)::reste when d=id -> 
	varPrim := [];
	secretVar := [];
	(let r = rcvMsg m t e d in (Stack.push (n,["State="^(string_of_int !nbTrans)]@(fst r),["State':="^(string_of_int (!nbTrans+1))]@(snd r)) resul));
	incr nbTrans;
	(match t with
	  Dolev_Yao -> ()
	| Non_writable -> incr nbNW
	| Protected -> incr nbP);
	getTransRec reste
   | (_,_,_,_,_)::reste -> getTransRec reste
   in
	getTransRec msgs;
	stack2list resul
;;

(* ----------------
   getSets : channel_type*int*string*string*msg list -> (string list*int)*(string list*int) 
   ---------------- *)
let rec getSets s msgs = match msgs with
  [] -> (([],0),([],0))
| (Dolev_Yao,_,_,_,_)::reste -> getSets s reste
| (Non_writable,_,e,d,_)::reste -> let ((h,hi),(k,ki))=getSets s reste in 
	if e=s || d=s then ((union ["Nonce_set_"^e^"_"^d] h,hi+1),(k,ki)) else ((h,hi),(k,ki))
| (Protected,_,e,d,_)::reste -> let ((h,hi),(k,ki))=getSets s reste in 
	if e=s || d=s then ((h,hi),(union ["Key_set_"^e^"_"^d] k,ki+1)) else ((h,hi),(k,ki))
;;

(* ----------------
   getAgents : struct_protocol -> role_agent list
   ---------------- *)
let getAgents proto optS =
	let rec hash_key_vars s i = match i with
	   0 -> []
	 | n -> (s^(string_of_int n))::(hash_key_vars s (n-1))
	in let params s = (
	  try 
		union [s] (snd (List.find (fun p -> let (id,_)=p in s=id) proto.knowledge))
	  with Not_found -> [s]
	) in let computeRoleAg s = (
	  let parameters = params s and sets = getSets s proto.messages in
	  let vars = List.filter (fun e -> not (List.mem e parameters)) (getAllIds s proto.messages) in
	  let trans = getTransitions s vars proto.messages proto.goals optS in
	  {
	  name_ag=s;
	  params_ag=parameters;
	  nonce_set_ag=(fst sets);
	  key_set_ag=(snd sets);
	  var_ag=vars;
	  transitions=trans;
	  }
	) in let rec getAgents_rec idents = ( match idents with 
	  [] -> []
	| (s,t)::reste when t=User-> [(computeRoleAg s)]@(getAgents_rec reste)
	| _::reste -> getAgents_rec reste
	) in getAgents_rec proto.idents
;;

(* ----------------
   getSessions : struct_protocol -> role_agent list -> role_session list
   ---------------- *)
let getSessions proto agents = 
 let rec getParams s ag = match ag with 
	  [] -> ([],[],[]) (* Si tout se passe bien on ne passe pas par ici *)
	| a::reste -> if a.name_ag=s then (a.params_ag,fst a.nonce_set_ag,fst a.key_set_ag) else getParams s reste
 in let rec getSession_rec sessions = (
   let agents_list = ref [] and param_list = ref [] and hash_list = ref [] and key_list = ref [] in
	match sessions with [] -> []
	| l::reste -> 
		List.iter (fun e ->
		 if (getType proto.idents (fst e))=User then
			let (p,h,k)=(getParams (fst e) agents) 
			and n = (string_of_int ((List.length !agents_list)+1)) in
			let ha = if List.length h >0 then ",Hash_0" else "" in
			agents_list:=("role_"^(fst e)^"("^(String.concat "," (p@k@h))^ha^",SND"^n^",RCV"^n^")")::(!agents_list);
			param_list:= union !param_list p;
			hash_list:= union !hash_list h;
			key_list:= union !key_list k
		) l;
		({params_s=(!param_list);nonce_set_s=(!hash_list);key_set_s=(!key_list);comp_s=(!agents_list);names_s=l})::(getSession_rec reste)
 ) in getSession_rec proto.sessions
;;

(* ----------------
   getEnvironment : struct_protocol -> role_session list -> role_env list
   ---------------- *)
let getEnvironment proto sessions =
 let cons = ref [] and comp = ref [] and nbS = ref 1 and nb = ref 1 in
	List.iter (fun s ->
		let tmp = ref [] in
		List.iter (fun p ->
		  try let c = snd (List.find (fun pp -> let (id,_)=pp in p=id) s.names_s) in
		     if c<>"i" then cons := union !cons [(c^":"^(print_type_var_hlpsl (getType proto.idents p)))];
		     tmp  := (!tmp)@[c]
		  with Not_found -> 
		    cons := !cons@["const_"^(string_of_int !nb)^":"^(print_type_var_hlpsl (getType proto.idents p))];
		    tmp  := (!tmp)@["const_"^(string_of_int !nb)]
		) s.params_s;
		(let rec emptySets n = match n with 0 -> () | i -> tmp := (!tmp)@["{}"];emptySets (i-1) in emptySets ( (List.length s.nonce_set_s)+(List.length s.key_set_s) ));
		(if (List.length s.nonce_set_s)>0 then tmp := (!tmp)@["hash_0"];cons := union ["hash_0:function"] !cons);
		comp := ("session"^(string_of_int !nbS)^"("^(String.concat "," !tmp)^")")::(!comp);
		incr nbS
	) sessions;
	List.iter ( fun g -> match g with
	   (i,Secrecy (_,_)) -> cons := !cons@["sec_"^(string_of_int i)^":protocol_id"]
  	 | (i,Authentication (_,_,_)) -> cons := !cons@["auth_"^(string_of_int i)^":protocol_id"] 
  	 | (i,Weak_authentication (_,_,_)) -> cons := !cons@["wauth_"^(string_of_int i)^":protocol_id"]
	) proto.goals;
 {cons_env=(!cons);comp_env=(!comp);intruder_env=proto.intruder}
;;

(* ----------------
   getGoals : (int*goal_type) list -> string list
   ---------------- *)
let getGoals goals = 
  let resul = ref [] in
  List.iter ( fun g -> match g with
	  (i,Secrecy (_,_)) -> resul := !resul@["secrecy_of sec_"^(string_of_int i)]
  	| (i,Authentication (_,_,_)) -> resul := !resul@["authentication_on auth_"^(string_of_int i)] 
  	| (i,Weak_authentication (_,_,_)) -> resul := !resul@["weak_authentication_on wauth_"^(string_of_int i)]
	
  ) goals;
  !resul
;;

(* ----------------
   protocol2Hlspl : struct_protocol -> bool -> struct_hlpsl
   ---------------- *)
let protocol2Hlspl proto optS = 
 let ag = getAgents proto !optS in let se = getSessions proto ag in
{	idents_func=
		(fun s -> try 
			snd (List.find (fun p -> let (id,_)=p in s=id) proto.idents) 
			with Not_found -> Error);
	agents=ag;
	sessions_h=se;
	environment=getEnvironment proto se;
	goals_h=getGoals proto.goals;
};; 
