
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
open Types;;

(*
let prerr_string=print_string
let prerr_endline=print_string
let prerr_int=print_int
*)


(* ----------------
   print_type_var : type_var -> string
   ---------------- *)
let print_type_var var = match var with 
    User -> "Agent"
  | SyKey -> "Symmetric key"
  | PcKey -> "Public key"
  | Table -> "Table"
  | Number -> "Number"
  | Function -> "Function"
  | Error -> "Error"
;;

(* ----------------
   print_type_var_hlpsl : type_var -> string
   ---------------- *)
let print_type_var_hlpsl var = match var with 
    User -> "agent"
  | SyKey -> "symmetric_key"
  | PcKey -> "public_key"
  | Table -> "agent->public_key"
  | Number -> "text"
  | Function -> "function"
  | Error -> raise (Call "ERROR : internal error.")
;;

(* ----------------
   print_msg : msg -> string
   ---------------- *)
let rec print_msg msg = match msg with
    Id s -> s
  | Int i -> (string_of_int i)
  | Func (s,m) -> s^"("^(print_msg m)^")"
  | Crypt (m1,m2) -> "{"^(print_msg m1)^"}_"^(print_msg m2)
  | Exp (m1,m2) -> "exp("^(print_msg m1)^","^(print_msg m2)^")"
  | Xor (m1,m2) -> "xor("^(print_msg m1)^","^(print_msg m2)^")"
  | Pair (m1,m2) -> (print_msg m1)^","^(print_msg m2)
  | Inv s -> "inv("^s^")"
  | TableElt (s1,s2) -> s1^"["^s2^"]"
  | TableEltInv (s1,s2) -> "inv("^s1^"["^s2^"])"
;;

(* ----------------
   print_knowledge : (string * string list)-> string
   ---------------- *)
let print_knowledge k = let (id,l)=k in
	let resul = ref (id^" :") in (
		List.iter (fun s -> resul:=!resul^" "^s) l;
		!resul
	)
;;

(* ----------------
   print_session : (string * string) list -> string
   ---------------- *)
let print_session sess = let resul = ref "" in
	List.iter (fun s -> resul:=!resul^(fst s)^":"^(snd s)^" ") sess;
	!resul
;;

(* ----------------
   print_goal : type_goal -> string
   ---------------- *)
let print_goal g = match g with
    Secrecy (s,_) -> "Secrecy of "^s
  | Authentication(a,b,c) -> a^" authenticates "^b^" on "^c
  | Weak_authentication(a,b,c) -> a^" authenticates (weak) "^b^" on "^c
;;

(* ----------------
   print_arrow : channel_type -> string
   ---------------- *)
let print_arrow t = match t with
    Dolev_Yao -> "->"
  | Protected -> "=>"
  | Non_writable -> "~>"
;;

(* ----------------
   print_struct_protocol : struct_protocol -> unit
   ---------------- *)
let print_struct_protocol proto = 
	let {name=n;idents=id;messages=m;knowledge=k;sessions=s;intruder=i;goals=g}=proto in
	print_string ("Name : "^n^"\n");
	print_string "Identifiers : \n";
	List.iter (fun i -> let (s,t)=i in print_string ("\t"^s^" : "^(print_type_var t)^"\n")) id;
	print_string "Messages : \n";
	List.iter (fun msg -> let (t,n,e,d,m)=msg in print_string ("\t"^(string_of_int n)^". "^e^" "^(print_arrow t)^" "^d^" : "^(print_msg m)^"\n")) m;
	print_string "Knowledge : \n";
	List.iter (fun know -> print_string ("\t"^(print_knowledge know)^"\n")) k;
	print_string "Sessions : \n";
	List.iter (fun sess -> print_string ("\t"^(print_session sess)^"\n")) s;
	print_string "Intruder knowledge : \n\t";
	List.iter (fun know -> print_string (know^" ")) i;
	print_string "\nGoals : \n";
	List.iter (fun goal -> print_string ("\t"^(print_goal (snd goal))^"\n")) g
;; 

(* ----------------
   id2idType : (string -> type_var) -> string list -> string list
   ---------------- *)
let rec id2idType func_t l = match l with
   [] -> []
 | id::reste -> (id^":"^(print_type_var_hlpsl (func_t id)))::(id2idType func_t reste)
;;

(* ----------------
   session_local : int -> string
   ---------------- *)
let rec session_local n =
 match n with
  0 -> ""
| 1 -> "SND1,RCV1:channel(dy)"
| i -> "SND"^(string_of_int i)^",RCV"^(string_of_int i)^","^(session_local (i-1))
;;

(* ----------------
   addNonceSet : string list -> string list
   ---------------- *)
let addNonceSet l =
 let rec addNonceSet_rec l =
	match l with
	   [] -> []
	 | a::reste -> (a^":(text) set")::(addNonceSet_rec reste)
 in (addNonceSet_rec l)@(if (List.length l) > 0 then ["Hash_0:function"] else [])
;;

(* ----------------
   addKeySet : string list -> string list
   ---------------- *)
let rec addKeySet l =
 match l with
   [] -> []
 | a::reste -> (a^":(symmetric_key) set")::(addKeySet reste)
;;

(* ----------------
   addHashKeyVars : string list -> string list
   ---------------- *)
let rec addHashKeyVars i j = match i with
   0  -> ( match j with 0  -> []
		   | nj -> ("Key_"^(string_of_int nj)^":symmetric_key")::(addHashKeyVars 0 (nj-1)))
 | ni -> ("Nonce_"^(string_of_int ni)^":text")::(addHashKeyVars (ni-1) j)
;;

(* ----------------
   print_struct_hlspl : struct_hlspl -> channel_out -> unit
   ---------------- *)
let print_struct_hlspl struct_h out = 
 let {idents_func=getTypeIdent;agents=agent_list;sessions_h=session_list;environment=env;goals_h=goals}=struct_h and nbSess = ref 1 in
	List.iter (fun agent -> 
	    Printf.fprintf out "role role_%s(%s,SND,RCV:channel(dy))\nplayed_by %s\ndef=\n\tlocal\n\t\t%s\n\tinit\n\t\tState := 0\n\ttransition\n"
	    agent.name_ag
	    (String.concat "," ((id2idType getTypeIdent agent.params_ag)@(addKeySet (fst agent.key_set_ag))@(addNonceSet (fst agent.nonce_set_ag))))
	    agent.name_ag
	    (String.concat "," (("State:nat"::(id2idType getTypeIdent agent.var_ag))@(addHashKeyVars (snd agent.nonce_set_ag) (snd agent.key_set_ag))));
		List.iter (fun tr -> let (n,g,d)=tr in
			Printf.fprintf out "\t\t%i. %s =|> %s\n" n (String.concat " /\\ " g) (String.concat " /\\ " d) 
		) agent.transitions;
		Printf.fprintf out "end role\n\n"
	) agent_list;
	List.iter (fun sess -> 
		Printf.fprintf out "role session%i(%s)\ndef=\n\tlocal\n\t\t%s\n\tcomposition\n\t\t%s\nend role\n\n" !nbSess (String.concat "," ((id2idType getTypeIdent sess.params_s)@(addKeySet sess.key_set_s)@(addNonceSet sess.nonce_set_s))) (session_local (List.length sess.comp_s)) (String.concat " /\\ " sess.comp_s);
		incr nbSess
	) session_list;
	
	Printf.fprintf out "role environment()\ndef=\n\tconst\n\t\t%s\n\tintruder_knowledge = {%s}\n\tcomposition\n\t\t%s\nend role\n\n" (String.concat "," env.cons_env) (String.concat "," env.intruder_env) (String.concat " /\\ " env.comp_env);
	
	if not(goals=[]) then Printf.fprintf out "goal\n\t%s\nend goal\n\n" (String.concat "\n\t" goals);
	
	Printf.fprintf out "environment()"
;;
