
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
(* This module can be considered as an Instantiation sub-module 
   It generates the transition rules that will be displayed in IF format
 *)
open Types;;
open Globals;;
open Utils;;
open Error_handler;;
open Interface;;

(*
   We need the following function to transform HLPSL predicates to IF predicates.
  \begin{center}
  type:
   [val get_if_term : Types.predicate -> Types.if_term]
  \end{center}
*)
let rec get_if_term = function
    Not(p) -> Ifnot(get_if_term p)
  | Equal(t1,t2) -> Ifequal(t1,t2)
  | Leq(t1,t2) -> Ifleq(t1,t2)
  | In(t1,t2) -> Ifcontains(t1,t2)
  | Event(Const(i) as e,[]) ->
(*      if (i = Globals.global_start) then
	Iknows(Base(e))
      else*)
      Unknown(Base(e),[])
  | New(xid) -> Ifnew(xid)
  | Event((Var(i) | Const(i)) as e,l) ->
      try (
	match (Globals.type_table#get_type i) with
      	  Base(Channel)
	| Base(Channel_ota)
	| Base(Channel_dy) ->
	    Iknows(List.hd l)
      	| _ ->
	    Unknown(Base(e),l))
      with Not_found -> Unknown(Base(e),l)
;;

(*
  \begin{center}
  type:
   [val apply_subst :
     (Types.atoms_type Types.term_structure *
      Types.atoms_type Types.term_structure) list ->
     Types.atoms_type Types.term_structure ->
     Types.atoms_type Types.term_structure]
  \end{center}
*)
let apply_subst subst =
  let rec apply_subst_on_term = function
      Pair(t1,t2) ->
	Pair(apply_subst_on_term t1, apply_subst_on_term t2)
    | Scrypt(t1,t2) ->
	Scrypt(apply_subst_on_term t1, apply_subst_on_term t2)
    | Crypt(t1,t2) ->
	Crypt(apply_subst_on_term t1, apply_subst_on_term t2)
    | Cons(t1,t2) ->
 	Cons(apply_subst_on_term t1, apply_subst_on_term t2)
    | Delete(t1,t2) ->
 	Delete(apply_subst_on_term t1, apply_subst_on_term t2)
    | Inv(t) ->
 	Inv(apply_subst_on_term t)
    | Function(t1,l) ->
 	Function(apply_subst_on_term t1,
		 List.map apply_subst_on_term l)
    | Set(l) ->
 	Set(List.map apply_subst_on_term l)
(*    | Lst(l) ->
 	Lst(List.map apply_subst_on_term l)*)
    | Base(Var(x)) as t ->
 	if List.mem_assoc t subst then
	  List.assoc t subst
	else t
    | Prime(Var(x)) as t ->
 	if List.mem_assoc t subst then
	  List.assoc t subst
	else t
    | t -> t
  in apply_subst_on_term
;;

(*
  \begin{center}
  type:
   [val get_if_lhs :
     Types.predicate list ->
     (Types.atoms_type Types.term_structure * Types.term) list *
     Types.if_term list * Types.if_term list]
  \end{center}
*)
let get_if_lhs hlpsl_terms = 
  let subst = ref []
  and lcond = ref []
  and lmsg = ref [] in
  List.iter
    (function
	Equal(t1,t2) ->
	  (match (apply_subst !subst t1, apply_subst !subst t2) with
	    (Base(Var(_)) as t1, t2)
	  | (t2, (Base(Var(_)) as t1)) ->
	      let p = [(t1,t2)] in
	      subst := (List.map (fun (s,t) -> (s, apply_subst p t)) !subst)@p
	  | (Prime(Var(_)) as t1, t2)
	  | (t2, (Prime(Var(_)) as t1)) ->
	      let p = [(t1,t2)] in
	      subst := (List.map (fun (s,t) -> (s, apply_subst p t)) !subst)@p
	  | _ -> lcond := !lcond@[Ifequal(t1,t2)])
      | Not(p) -> lcond := !lcond@[Ifnot(get_if_term p)]
      | Leq(t1,t2) -> lcond := !lcond@[Ifleq(t1,t2)]
      | In(t1,t2) -> lcond := !lcond@[Ifcontains(t1,t2)]
      | New _ -> ()
      | event -> lmsg := !lmsg@[get_if_term event])
    hlpsl_terms;
  lcond := List.map (function
      Ifequal(t1,t2) ->
	Ifequal(apply_subst !subst t1, apply_subst !subst t2)
    | Ifleq(t1,t2) ->
	Ifleq(apply_subst !subst t1, apply_subst !subst t2)
    | Ifcontains(t1,t2) ->
	Ifcontains(apply_subst !subst t1, apply_subst !subst t2)
    | Ifnot(Ifequal(t1,t2)) ->
	Ifnot(Ifequal(apply_subst !subst t1, apply_subst !subst t2))
    | Ifnot(Ifleq(t1,t2)) ->
	Ifnot(Ifleq(apply_subst !subst t1, apply_subst !subst t2))
    | Ifnot(Ifcontains(t1,t2)) ->
	Ifnot(Ifcontains(apply_subst !subst t1, apply_subst !subst t2))
    | t -> t)
      !lcond;
  lmsg := List.map (function
      Iknows(t) ->
	Iknows(apply_subst !subst t)
    | Ifdelete(t1,t2) ->
	Ifdelete(apply_subst !subst t1, apply_subst !subst t2)
    | Ifcontains(t1,t2) ->
	Ifcontains(apply_subst !subst t1, apply_subst !subst t2)
    | Unknown(t1,lt2) ->
	Unknown(t1, List.map (apply_subst !subst) lt2)
(*    | Ifcontains(t1,t2) ->
	Ifcontains(apply_subst !subst t1, apply_subst !subst t2)*)
    | t -> t)
      !lmsg;
  (!subst,!lcond,!lmsg)
;;

(*
   We need this function to translate HLPSL conditions into IF conditions.
  \begin{center}
  type:
   [val get_if_cond : Types.predicate list -> Types.if_term list]
  \end{center}
*)
let get_if_cond hlpsl_cond_list = 
  let res = ref [] in
  List.iter
    (function
        (* BUG??? the name of the variable on the left-hand-side is lost!
           no, because the Cons term is copied in the state predicate as
	   new value of the primed variable *)
	Equal(Prime(t1), Cons(t2,t3)) ->
	  res:=!res@[Ifcontains(t2,t3)]
      | Equal(Prime(t1), Delete(t2,t3)) ->
	  res:=!res@[Ifdelete(t2,t3)]
      | New _
      | Event _
      | Equal _ -> () (* equalities are built in state terms: but useful
			   for detecting problems *)
      | hlpsl_cond -> res:=!res@[get_if_term hlpsl_cond])
    hlpsl_cond_list;
  !res
;;

(*
   We need to extract the message terms from the HLPSL perdicates.
  \begin{center}
  type:
   [val get_iknows : Types.predicate list -> Types.if_term list]
  \end{center}
*)
let get_iknows hlpsl_pred_list = 
  List.filter
    (function
	Iknows _
      | Unknown _
      | Ifdelete _
      | Ifcontains _ ->  true
      | _ -> false)
    (List.map get_if_term hlpsl_pred_list)
;;

(*
   The following function is used for detecting variables that have
   two values in a transition: an old one and a new one.  The old one
   is renamed Dummy_..., because either it is not used (it only appear
   in the state term), or it is used to check some information.  The
   variable with the new value has the standard name.
   \begin{center}
   type:
   [val update_dummy_vars :
     Types.predicate list ->
     Types.predicate list ->
     int list ->
     Types.atoms_type Types.term_structure list * Types.predicate list *
     Types.predicate list]
   \end{center}
***
let update_dummy_vars lhs rhs lgid =

  let non_fresh_terms =
    let rec get_non_fresh_terms lt = function
      	Pair(t1,t2)
      | Scrypt(t1,t2)
      | Crypt(t1,t2)
      | Cons(t1,t2)
      | Delete(t1,t2) ->
 	  get_non_fresh_terms (get_non_fresh_terms lt t1) t2
      | Inv(t) ->
 	  get_non_fresh_terms lt t
      | Function(t1,l) ->
 	  List.fold_left
	    get_non_fresh_terms
	    (get_non_fresh_terms lt t1)
	    l
*      | Lst(l)*
      | Set(l) ->
 	  List.fold_left get_non_fresh_terms lt l
      | Base(Var(x)) ->
 	  if List.mem x lt then lt
	  else x::lt
      | _ -> lt
    in
    List.fold_left
      (fun lt ->
	(function
	    Not(Equal(t1,t2))
	  | Not(Leq(t1,t2))
	  | Not(In(t1,t2))
	  | Equal(t1,t2)
	  | Leq(t1,t2)
	  | In(t1,t2) ->
	      get_non_fresh_terms (get_non_fresh_terms lt t1) t2
          | Event(_,lt2) ->
	      List.fold_left get_non_fresh_terms lt lt2
	  | _ -> lt))
      []
      (lhs@rhs)
  in

  let non_used_terms =
    let rec get_fresh_terms lt = function
      	Pair(t1,t2)
      | Scrypt(t1,t2)
      | Crypt(t1,t2)
      | Cons(t1,t2)
      | Delete(t1,t2) ->
 	  get_fresh_terms (get_fresh_terms lt t1) t2
      | Inv(t) ->
 	  get_fresh_terms lt t
      | Function(t1,l) ->
 	  List.fold_left
	    get_fresh_terms
	    (get_fresh_terms lt t1)
	    l
*      | Lst(l)*
      | Set(l) ->
 	  List.fold_left get_fresh_terms lt l
      | Prime(Var(x)) ->
 	  if (List.mem x lt) then lt
	  else x::lt
      | _ -> lt
    in
    List.fold_left
      (fun lt ->
	(function
	    Not(Equal(t1,t2))
	  | Not(Leq(t1,t2))
	  | Not(In(t1,t2))
	  | Equal(t1,t2)
	  | Leq(t1,t2)
	  | In(t1,t2) ->
	      get_fresh_terms (get_fresh_terms lt t1) t2
          | Event(_,lt2) ->
	      List.fold_left get_fresh_terms lt lt2
          | New(x) ->
	      x::lt
	  | _ -> lt))
      []
      (lhs@rhs)
  in

  let subst = ref [] in

  let lnewgid =
    List.map
      (fun gid ->
      	let var_id = global_var_id#get_name_of gid in
      	if (List.mem var_id non_used_terms)
	    && (Globals.is_dummy_term (value_table#get_value (memory_map#get_map gid))) then
          let new_name = "Test_"^(string_id#get_name var_id) in
          let new_id = string_id#register_name new_name 0
	  and typ_var_id = type_table#get_type var_id
	  in
	  (if (type_table#register_type new_id typ_var_id <> typ_var_id) then
	    Semantic_check.add_sem_error
	      (20, "variable "^new_name^" declared with incorrect type (reserved)", "", "","","");
	   subst := (var_id,Base(Var(new_id)))::!subst;
	   Base(Var(new_id)))
        else
          Base(Var(var_id)))
      lgid
  in

  let apply_subst_on_atom =
    let rec apply_subst_on_term = function
      	Pair(t1,t2) ->
	  Pair(apply_subst_on_term t1, apply_subst_on_term t2)
      | Scrypt(t1,t2) ->
	  Scrypt(apply_subst_on_term t1, apply_subst_on_term t2)
      | Crypt(t1,t2) ->
	  Crypt(apply_subst_on_term t1, apply_subst_on_term t2)
      | Cons(t1,t2) ->
 	  Cons(apply_subst_on_term t1, apply_subst_on_term t2)
      | Delete(t1,t2) ->
 	  Delete(apply_subst_on_term t1, apply_subst_on_term t2)
      | Inv(t) ->
 	  Inv(apply_subst_on_term t)
      | Function(t1,l) ->
 	  Function(apply_subst_on_term t1,
		   List.map apply_subst_on_term l)
      | Set(l) ->
 	  Set(List.map apply_subst_on_term l)
*      | Lst(l) ->
 	  Lst(List.map apply_subst_on_term l)*
      | Base(Var(x)) as t ->
 	  if List.mem_assoc x !subst then
	    List.assoc x !subst
	  else t
      | t -> t
    in function
	Equal(t1,t2) ->
	  Equal(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(Equal(t1,t2)) ->
	  Not(Equal(apply_subst_on_term t1, apply_subst_on_term t2))
      | Leq(t1,t2) ->
	  Leq(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(Leq(t1,t2)) ->
	  Not(Leq(apply_subst_on_term t1, apply_subst_on_term t2))
      | In(t1,t2) ->
	  In(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(In(t1,t2)) ->
	  Not(In(apply_subst_on_term t1, apply_subst_on_term t2))
      | Event(a,l) ->
	  Event(a, List.map apply_subst_on_term l)
      | p -> p
  in
  (lnewgid,
   List.map apply_subst_on_atom lhs,
   List.map apply_subst_on_atom rhs)
;;
***)
let update_dummy_vars lhs rhs lgid =

  let primed_terms =
    let rec get_fresh_terms lt = function
      	Pair(t1,t2)
      | Scrypt(t1,t2)
      | Crypt(t1,t2)
      | Cons(t1,t2)
      | Delete(t1,t2) ->
 	  get_fresh_terms (get_fresh_terms lt t1) t2
      | Inv(t) ->
 	  get_fresh_terms lt t
      | Function(t1,l) ->
 	  List.fold_left
	    get_fresh_terms
	    (get_fresh_terms lt t1)
	    l
(*      | Lst(l)*)
      | Set(l) ->
 	  List.fold_left get_fresh_terms lt l
      | Prime(Var(x)) ->
 	  if (List.mem x lt) then lt
	  else x::lt
      | _ -> lt
    in
    List.fold_left
      (fun lt ->
	(function
	    Not(Equal(t1,t2))
	  | Not(Leq(t1,t2))
	  | Not(In(t1,t2))
	  | Equal(t1,t2)
	  | Leq(t1,t2)
	  | In(t1,t2) ->
	      get_fresh_terms (get_fresh_terms lt t1) t2
          | Event(_,lt2) ->
	      List.fold_left get_fresh_terms lt lt2
          | New(x) ->
	      x::lt
	  | _ -> lt))
      []
      (lhs@rhs)
  in

  let subst = ref [] in

  let lnewgid =
    List.map
      (fun gid ->
      	let var_id = global_var_id#get_name_of gid in
      	if (List.mem var_id primed_terms)
(*	    && (Globals.is_dummy_term (value_table#get_value (memory_map#get_map gid))) *)
	then
          let new_name = "Test_"^(string_id#get_name var_id) in
          let new_id = string_id#register_name new_name 0
	  and typ_var_id = type_table#get_type var_id
	  in
	  (if (type_table#register_type new_id typ_var_id <> typ_var_id) then
	    Semantic_check.add_sem_error
	      (20, "variable "^new_name^" declared with incorrect type (reserved)", "", "","","");
	   subst := (var_id,Base(Var(new_id)))::!subst;
	   Base(Var(new_id)))
        else
          Base(Var(var_id)))
      lgid
  in

  let apply_subst_on_atom =
    let rec apply_subst_on_term = function
      	Pair(t1,t2) ->
	  Pair(apply_subst_on_term t1, apply_subst_on_term t2)
      | Scrypt(t1,t2) ->
	  Scrypt(apply_subst_on_term t1, apply_subst_on_term t2)
      | Crypt(t1,t2) ->
	  Crypt(apply_subst_on_term t1, apply_subst_on_term t2)
      | Cons(t1,t2) ->
 	  Cons(apply_subst_on_term t1, apply_subst_on_term t2)
      | Delete(t1,t2) ->
 	  Delete(apply_subst_on_term t1, apply_subst_on_term t2)
      | Inv(t) ->
 	  Inv(apply_subst_on_term t)
      | Function(t1,l) ->
 	  Function(apply_subst_on_term t1,
		   List.map apply_subst_on_term l)
      | Set(l) ->
 	  Set(List.map apply_subst_on_term l)
(*      | Lst(l) ->
 	  Lst(List.map apply_subst_on_term l)*)
      | Base(Var(x)) as t ->
 	  if List.mem_assoc x !subst then
	    List.assoc x !subst
	  else t
      | t -> t
    in function
	Equal(t1,t2) ->
	  Equal(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(Equal(t1,t2)) ->
	  Not(Equal(apply_subst_on_term t1, apply_subst_on_term t2))
      | Leq(t1,t2) ->
	  Leq(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(Leq(t1,t2)) ->
	  Not(Leq(apply_subst_on_term t1, apply_subst_on_term t2))
      | In(t1,t2) ->
	  In(apply_subst_on_term t1, apply_subst_on_term t2)
      | Not(In(t1,t2)) ->
	  Not(In(apply_subst_on_term t1, apply_subst_on_term t2))
      | Event(a,l) ->
	  Event(a, List.map apply_subst_on_term l)
      | p -> p
  in
  (lnewgid,
   List.map apply_subst_on_atom lhs,
   List.map apply_subst_on_atom rhs)
;;

(*
   We determine the received fresh terms in a message in order to update
   their IF value in IF the rules.
  \begin{center}
  type:
   [val get_received_fresh_terms :
     Types.atoms_type Types.term_structure -> Types.atoms_type list]
  \end{center}
*)
let rec get_received_fresh_terms = function
    Pair(t1,t2)
  | Scrypt(t1,t2)
  | Crypt(t1,t2)
  | Cons(t1,t2) ->
      (get_received_fresh_terms t1)@(get_received_fresh_terms t2)
  | Delete(t1,t2) ->
      (get_received_fresh_terms t1)@(get_received_fresh_terms t2)
  | Inv(t) ->
      get_received_fresh_terms t
  | Function(t1,l) ->
      (get_received_fresh_terms t1)
      @(List.flatten (List.map get_received_fresh_terms l))
(*  | Lst(l)*)
  | Set(l) ->
      List.flatten (List.map get_received_fresh_terms l)
  | Base(Var(i))
  | Prime(Var(i)) -> [Var i]
  | _ -> []
;;

(*
   Received fresh terms change their value according to the incoming message.
   This update takes into account the IF fresh values handling.
   We manage the progress of the role state variables through two lists.
   [current_state_vars] contains the list of variables in their current "execution step".
   [generic_state_vars] contains a static model (map) of the variables.
   It can be used to retrieves information that does not exists anymore 
   in the [current_state_vars] because of the changes..
  \begin{center}
  type:
   [val update_received_fresh_terms :
     Types.atoms_type Types.term_structure list ->
     Types.predicate list ->
     Types.atoms_type Types.term_structure list ->
     Types.atoms_type Types.term_structure list]
  \end{center}
*)
let update_received_fresh_terms generic_state_vars lhs current_state_vars = 
  let received_fresh_terms =
    List.flatten
      (List.map
	 (function
	     Iknows(t) -> get_received_fresh_terms t 
	   | Ifequal(t1,t2) -> (get_received_fresh_terms t1)@(get_received_fresh_terms t2)
	   | Ifcontains(t1,t2) -> (get_received_fresh_terms t1)@(get_received_fresh_terms t2)
	   | _ -> [])
	 (List.map get_if_term lhs))
  in
  List.map2
    (fun g_var c_var ->
      if (List.exists (fun x -> Base x = g_var) received_fresh_terms) then
	g_var
      else
	c_var)
    generic_state_vars current_state_vars
;;

(*
   The HLPSL allows the use of fresh terms which values change when they are "primed".
   We use this function to search for them..
  \begin{center}
  type:
   [val get_fresh_terms : Types.if_term list -> Types.atoms_type list]
  \end{center}
*)
let get_fresh_terms msg_term =
  let rec get_fresh_terms_from_msg = function
      Pair(t1,t2)
    | Scrypt(t1,t2)
    | Crypt(t1,t2)
    | Cons(t1,t2) ->
 	(get_fresh_terms_from_msg t1)@(get_fresh_terms_from_msg t2)
    | Delete(t1,t2) ->
 	(get_fresh_terms_from_msg t1)@(get_fresh_terms_from_msg t2)
    | Inv(t) ->
 	get_fresh_terms_from_msg t
    | Function(t1,l) ->
 	(get_fresh_terms_from_msg t1)
	@(List.flatten (List.map get_fresh_terms_from_msg l))
(*    | Lst(l)*)
    | Set(l) ->
 	List.flatten (List.map get_fresh_terms_from_msg l)
    | Base(_) -> []
    | Prime(t) -> [t]
    | Hash(_) ->
	prerr_endline "Semantic_check.check_var: Impossible error: invalid hash type/term"; 
	raise Not_found
  in
  List.flatten
    (List.map
       (function
	   Iknows(t) -> get_fresh_terms_from_msg t
	 | Unknown(_,lt) ->
	     List.fold_left
	       (fun l t -> l@(get_fresh_terms_from_msg t))
	       [] lt
	 | _ -> [])
       msg_term)
;;

(*
   
  \begin{center}
  type:
   [val get_primed_vars : Types.predicate list -> int list]
  \end{center}
*)
let get_primed_vars  =
  let rec get_primed_vars_rec lt = function
      Pair(t1,t2)
    | Scrypt(t1,t2)
    | Crypt(t1,t2)
    | Cons(t1,t2)
    | Delete(t1,t2) ->
 	get_primed_vars_rec (get_primed_vars_rec lt t1) t2
    | Inv(t) ->
 	get_primed_vars_rec lt t
    | Function(t1,l) ->
 	List.fold_left get_primed_vars_rec (get_primed_vars_rec lt t1) l
(*    | Lst(l)*)
    | Set(l) ->
 	List.fold_left get_primed_vars_rec [] l
    | Prime(Var(i)) ->
	if List.mem i lt then lt
	else i::lt
    | _ -> lt
  in
  List.fold_left
    (fun lt ->
      (function
	  Equal(t1,t2)
      	| Not(Equal(t1,t2))
      	| Leq(t1,t2)
      	| Not(Leq(t1,t2))
      	| In(t1,t2)
        | Not(In(t1,t2)) ->
	    get_primed_vars_rec (get_primed_vars_rec lt t1) t2
(*Event:      	| Event(_,t::_) ->
	    get_primed_vars_rec lt t *)
      	| Event(_,lt2) ->
	    List.fold_left get_primed_vars_rec lt lt2
      	| New(x) ->
	    x::lt
      	| _ -> lt))
    []
;;

(*
   This function looks for fresh terms (primed terms) in a given list
   of terms (corresponding to the rhs of a transition) that do not
   also appear as primed terms in the lhs.
   Conventions: a real fresh term can only appear
   - in a message (head of the second element of Event), or
   - in the first argument of an inclusion (In), or
   - in the right-hand side of an equality (Equal).
   Note that a primed term in the lhs has to appear in a received
   message (Event) or as first argument of inclusion (In)..
  \begin{center}
  type:
   [val get_real_fresh_terms :
     Types.predicate list -> Types.predicate list -> Types.atoms_type list]
  \end{center}
*)
let get_real_fresh_terms lhs rhs =

  let get_new_ones =
    List.fold_left
      (fun lt' t' ->
	if List.mem t' lt' then
	  lt'
	else t'::lt')
  in

  let non_real_fresh_terms =
    List.fold_left
      (fun lt ->
	(function
	    In(t,_) ->
	      get_new_ones lt (get_fresh_terms [Iknows(t)])
          | Equal(_,t) ->
	      get_new_ones lt (get_fresh_terms [Iknows(t)])
          | Event(_,lt2) ->
	      List.fold_left
		(fun lt t ->
		  get_new_ones lt (get_fresh_terms [Iknows(t)]))
		lt
		lt2
          | _ -> lt))
      (List.fold_left
	 (fun lt ->
	   (function
	       Equal(t,_) ->
		 get_new_ones lt (get_fresh_terms [Iknows(t)])
             | _ -> lt))
         [] rhs)
      lhs
  in

  let rec get_fresh_terms_from_msg = function
      Pair(t1,t2)
    | Scrypt(t1,t2)
    | Crypt(t1,t2)
    | Cons(t1,t2)
    | Delete(t1,t2) ->
 	(get_fresh_terms_from_msg t1)@(get_fresh_terms_from_msg t2)
    | Inv(t) ->
 	get_fresh_terms_from_msg t
    | Function(t1,l) ->
 	(get_fresh_terms_from_msg t1)
	@(List.flatten (List.map get_fresh_terms_from_msg l))
(*    | Lst(l)*)
    | Set(l) ->
	List.flatten (List.map get_fresh_terms_from_msg l)
    | Base(_) -> []
    | Prime t ->
	if List.mem t non_real_fresh_terms then []
	else [t]
    | Hash(_) ->
	prerr_endline "Semantic_check.check_var: Impossible error: invalid hash type/term"; 
	raise Not_found
  in

  List.fold_left
    (fun lt ->
      (function
	  New(xid) ->
	    get_new_ones lt [Var(xid)]
	| In(t,_)
	| Equal(_,t) ->
	    get_new_ones lt (get_fresh_terms_from_msg t)
	| Event(_,lt2) ->
	    List.fold_left
	      (fun lt t ->
		get_new_ones lt (get_fresh_terms_from_msg t))
	      lt
	      lt2
	| _ -> lt))
    [] rhs
;;

(*
   Fresh terms should be updated at the receipt of a message.
  \begin{center}
  type:
   [val update_fresh_terms :
     Types.if_term list ->
     Types.atoms_type Types.term_structure list ->
     Types.atoms_type Types.term_structure list ->
     Types.atoms_type Types.term_structure list]
  \end{center}
*)
(*
let update_fresh_terms msg_term = 
  let fresh_terms = get_fresh_terms msg_term in*)
let update_fresh_terms fresh_terms = 
  List.map2
    (fun g_var c_var ->
      if (List.exists (fun x -> Base x = g_var) fresh_terms) then
	(match g_var with
	  Base(Var(i)) -> Prime(Var(i))
	| _ -> Base(Var(0)))
      else
	(match c_var with 
	  Prime(_) -> g_var
	| _ -> c_var))
;;

(*
   We define this function to build the list of variables of which we
   should keep track in the state facts..
  \begin{center}
  type:
   [val get_step_var_list :
     Types.atoms_type Types.term_structure list ->
     Types.atoms_type Types.term_structure list ref ->
     Types.atoms_type Types.term_structure list]
  \end{center}
*)
let get_step_var_list current_state_vars res =
  List.iter
    (function
	Base(Var i) | Prime(Var i) ->
	  if (not (List.exists (fun x-> x = Base (Var i))!res)) then
	    res:=!res@[Base(Var i)]
      | _ -> ())
    current_state_vars;
  !res
;; 

(*
   We need to update a variable in a LHS of an equality check by the
   RHS value that it should be equal to.
  \begin{center}
  type:
   [val update_vars_with_cond :
     ('a * 'a) list -> 'a list -> 'a list
  \end{center}
*)
let update_vars_with_cond subst =
  List.map
    (fun state_term ->
      try (List.assoc state_term subst)
      with Not_found -> state_term)
;;

(*
   We change the values of variables that have been assigned.
  \begin{center}
  type:
   [val update_rhs_primed_vars :
     (Types.atoms_type Types.term_structure * Types.term) list ->
     Types.predicate list ->
     Types.term list ->
     Types.atoms_type Types.term_structure list ->
     (Types.atoms_type Types.term_structure * Types.term) list * Types.term list]
  \end{center}
*)
let update_rhs_primed_vars subst rhs cst gsv =
(*BUG: this substitution may contain different assignments of the same variable! *)
(*** old version (May 3, 2005) ***
  let new_subst =
    (List.fold_left
       (fun l -> function
	   Equal(Prime(t1),Cons(_,t2))
	 | Equal(Prime(t1),Delete(_,t2))
	 | Equal(Prime(t1),t2) ->
	     if (List.exists
		   (function
		       (Prime(tt1),_) ->
			 t1=tt1
		     | _ -> false)
		   l) then
	       (displayWarning 10
		  ("multiple assignment of variable "
		   ^(term_to_string (Prime(t1))));
		l)
	     else
	       let l2 = [(Prime(t1), apply_subst l t2)]
	       in
	       (List.map
		  (fun (s,t) ->
		    (s, apply_subst l2 t))
		  l)
               @ l2
	 | _ -> l)
       subst rhs)
  in
***)
  let new_subst =
    let rec compute_subst l = function
	Equal(Prime(t1),
	      Cons(_,(Cons(_)|Delete(_) as t2)))::l2
      | Equal(Prime(t1),
	      Delete(_,(Cons(_)|Delete(_) as t2)))::l2 ->
	  compute_subst l (Equal(Prime(t1),t2)::l2)
      | Equal(Prime(t1),Cons(_,t2))::l2
      | Equal(Prime(t1),Delete(_,t2))::l2 ->
	  compute_subst
	    (let l' = [(Prime(t1), apply_subst l t2)] in
	    (List.map
	       (fun (s,t) ->
		 (s, apply_subst l' t))
	       l)
            @ l')
	    l2
      | Equal(Prime(t1),t2)::l2 ->
	  compute_subst
	    (if (List.exists
		   (function
		       (Prime(tt1),_) ->
			 t1=tt1
		     | _ -> false)
		   l) then
	      (displayWarning 10
		 ("multiple assignment of variable "
		  ^(term_to_string (Prime(t1))));
	       l)
	    else
	      let l' = [(Prime(t1), apply_subst l t2)]
	      in
	      (List.map
		 (fun (s,t) ->
		   (s, apply_subst l' t))
		 l)
              @ l')
	    l2
      | _::l2 -> compute_subst l l2
      | [] -> l
    in
    compute_subst subst rhs
  in
  let if_equal_terms =
    List.map
      (function
	  (Prime(v),t) -> (Base(v),t)
	| p -> p)
      (List.fold_left
	 (fun l -> function
	     (Prime(v),_) as p ->
	       (List.filter
		  (function
		      (Base(v2),_) -> v<>v2
		    | _ -> true)
		  l)
	       @[p]
	   | (Base(v),_) as p ->
	       if List.exists
		   (function
		       (Prime(v2),_) -> v=v2
		     | _ -> false)
		   l then
		 l
	       else
		 l@[p]
	   | p -> l@[p])
	 [] new_subst)
  in
  (new_subst,
   List.map2
     (fun current_state_term generic_state_term ->
       try (List.assoc generic_state_term if_equal_terms)
       with Not_found -> current_state_term)
     cst gsv)
;;

(*
  \begin{center}
  type:
   [val update_rhs_atoms :
     (Types.atoms_type Types.term_structure * Types.term) list ->
     Types.if_term list -> Types.if_term list]
  \end{center}
*)
(*** old version (May 3, 2005) ***
let update_rhs_atoms subst =
  List.map
    (function
 	Iknows(t) ->
	  Iknows(apply_subst subst t)
      | Unknown(t,lt) ->
	  Unknown(t, List.map (apply_subst subst) lt)
      | Ifcontains(t1,t2) ->
	  Ifcontains(apply_subst subst t1, apply_subst subst t2)
      | Ifdelete(t1,t2) ->
	  Ifdelete(apply_subst subst t1, apply_subst subst t2)
      | x -> x)
***)
let update_rhs_atoms subst =
  let rec ura_rec = function
      Iknows(t)::l ->
	Iknows(apply_subst subst t)::(ura_rec l)
    | Unknown(t,lt)::l ->
	Unknown(t, List.map (apply_subst subst) lt)::(ura_rec l)
    | Ifcontains(t1,t2)::l ->
	let l2 = ref [] in
	let rec apply_on_all_set_actions = function
	    Delete(t2,t3) ->
	      let t = apply_on_all_set_actions t3 in
	      l2 := Ifdelete(apply_subst subst t2, t)::!l2;
	      t
	  | Cons(t2,t3) ->
	      let t = apply_on_all_set_actions t3 in
	      l2 := Ifcontains(apply_subst subst t2, t)::!l2;
	      t
	  | t -> apply_subst subst t
	in
	let t = apply_on_all_set_actions t2 in
	(Ifcontains(apply_subst subst t1, t):: !l2)@(ura_rec l)
    | Ifdelete(t1,t2)::l ->
	let l2 = ref [] in
	let rec apply_on_all_set_actions = function
	    Delete(t2,t3) ->
	      let t = apply_on_all_set_actions t3 in
	      l2 := Ifdelete(apply_subst subst t2, t)::!l2;
	      t
	  | Cons(t2,t3) ->
	      let t = apply_on_all_set_actions t3 in
	      l2 := Ifcontains(apply_subst subst t2, t)::!l2;
	      t
	  | t -> apply_subst subst t
	in
	let t = apply_on_all_set_actions t2 in
	(Ifdelete(apply_subst subst t1, t):: !l2)@(ura_rec l)
    | x::l -> x::(ura_rec l)
    | [] -> []
  in
  ura_rec
;;

(*
  \begin{center}
  type:
   [val update_assigned_vars :
     Types.predicate list -> Types.if_term list -> Types.if_term list]
  \end{center}
*
let update_assigned_vars rhs =
  let subst =
    List.fold_left
      (fun l ->
	(function
	    Equal(Prime(_),Cons(_))
          | Equal(Prime(t1),Delete(_)) ->
              l  (* for avoiding "contains" in states *)
	  | Equal(Prime(t1),t2) -> (Base(t1),t2)::l
	  | _ -> l))
      [] rhs
  in
  List.map
    (function
 	Secret(t1,t2) as g ->
	  if List.mem_assoc t1 subst then
	    Secret(List.assoc t1 subst, t2)
	  else g
      | Witness(a1,a2,c,t) as g ->
	  if List.mem_assoc t subst then
	    Witness(a1, a2, c, List.assoc t subst)
	  else g
      | p -> p)
;;*)

(*
   We generate the goal terms that should be added to the RHS of IF rule.
  \begin{center}
  type:
   [val get_secrecy_terms :
     int list -> Types.goals_type list -> Types.term list -> Types.if_term list]
  \end{center}
*
let get_secrecy_terms fresh_vars goals state_terms = 
  let res = ref [] in
  let new_const_name = ref "" in
  List.iter
    (function
	Secrecy(term_list) ->
	  List.iter
	    (fun i ->
	      if (List.exists
		    (function
			Base(Var(xid)) ->
			  string_id#get_name i = string_id#get_name xid
		      | _ ->
			  Error_handler.displayWarning 1 
			    "Section_rules.get_secrecy_terms";
			  false)
		    term_list) then
		let list_of_agents =
		  List.filter
		    (function
			Base(Const(id)) | Base(Var(id))  ->
			  (try (Globals.type_table#get_type id = Base(Agent))
			  with _ -> false)
		      | _ -> false)
		    state_terms
		in
		res :=
		  (List.map
		     (function t -> Secret(Base(Var(i)),t))
		     list_of_agents)
		  @ !res)
	    fresh_vars
      | _ -> ())
    goals;
    !res
;;*)

(*
   We generate the goal terms that should be added to the RHS of IF rule.
  \begin{center}
  type:
   [val get_authentication_terms :
     int ->
     Types.atoms_type list -> Types.goals_type list -> Types.if_term list]
  \end{center}
*
let get_authentication_terms role_id fresh_terms goals = 
  let res = ref [] in
  let new_const_name = ref "" in
  List.iter
    (function
	Authentication(authenticator,authentified,term_list)
      | Weak_authentication(authenticator,authentified,term_list) ->
	  if (authentified = role_id) then
	    let test_if_add_witness t x =
	      match t,x with
		Var(tid), Base(Var(xid)) ->
		  new_const_name := 
		    String.lowercase (string_id#get_name tid);
		  string_id#get_name tid = string_id#get_name xid
	      | _ -> 
		  Error_handler.displayWarning 1 
		    "Section_rules.get_authentication_terms";
		  false
	    in
	    let add_witness_term t =
	      if (List.exists
		    (test_if_add_witness t)
		    term_list) then
 	      	let new_const = string_id#register_name !new_const_name 0 in
		ignore (type_table#register_type new_const (Base(Protocol_id)));
	       	res := 
		  !res@[Witness(authentified, authenticator, 
				Base(Const(new_const)),Base(t))]
	    in
	    List.iter (add_witness_term) fresh_terms

      | _ -> ())
    goals;
  !res
;;*)

(*
   This is the main function of the module. It computes the LHS and
   RHS of the IF rules. In addition to the required type translation
   and special formatting to transform HLSPL structures into IF rules,
   this function computes the different values of the fresh terms during
   the execution of the protocol.
  \begin{center}
  type:
   [val get_if_reactions :
     int -> hlpsl_transition list -> int list ->
     int -> Types.goals_type list -> (int * ('a * Types.basic_if_rule)) list]
  \end{center}
*)
let get_if_reactions role_id hlpsl_reactions gids session_id goals =
  (* Computation of a transformed list of gid, corresponding to gids where
     the player of the role is moved as first element *)
  let player_id = Utils.find_player_in_role role_id in
  let gids =
    let gid_player = ref 0 in
    let rec delete_player = function
	i::li ->
	  let name = global_var_id#get_name_of i in
	  if (name = player_id) then
	    (gid_player := i;
	     li)
	  else i::(delete_player li)
      | _ -> []
    in
    let gids' = delete_player gids in
    !gid_player::gids'
  in
  let generic_state_vars = 
    List.map 
      (fun x -> Base (Var (global_var_id#get_name_of x))) gids 
  in

  List.flatten
    (List.map
       (fun hlpsl_reaction ->
	 if (!debug_level >= 4) then
	   (prerr_string "% initial gids:";
	    List.iter (fun i -> prerr_string " "; prerr_term (Base(Var(global_var_id#get_name_of i)))) gids;
	    prerr_newline());
	 let lhs = hlpsl_reaction#get_lhs in
	 let rhs = hlpsl_reaction#get_rhs in
	 (* detection of dummy variables and renaming in the predicates *)
	 let (current_state_terms,lhs,rhs) =
 	   update_dummy_vars lhs rhs gids
	 in
	 if (!debug_level >= 4) then
	   (prerr_string "% current_state_terms:";
	    List.iter (fun i -> prerr_string " "; prerr_term i) current_state_terms;
	    prerr_newline());

	 let (subst,lhs_cond,lhs_msg) = get_if_lhs lhs in
	 if (!debug_level >= 4) then
	   (prerr_string "  get_if_lhs:";
	    List.iter (fun (i,j) -> prerr_string "  "; prerr_term i; prerr_string "<-"; prerr_term j) subst;
	    prerr_newline();
	    prerr_string "          cond:";
	    List.iter (fun i -> prerr_string " "; prerr_if_term "t" i) lhs_cond;
	    prerr_newline();
	    prerr_string "           msg:";
	    List.iter (fun i -> prerr_string " "; prerr_if_term "t" i) lhs_msg;
	    prerr_newline());

	 (* updating fresh terms changed by the receipt of a message*)
	 (*i current_state_terms:= update_fresh_terms lhs_msg generic_state_vars !current_state_terms; i*)

	 (* computing the list of freshly created terms *)
	 let exists_var_list = get_real_fresh_terms lhs rhs in

	 (* changing variables that were used in conditions by the values they should have to apply transitions *)
	 let current_state_terms = update_vars_with_cond subst current_state_terms in
	 if (!debug_level >= 4) then
	   (prerr_string "  update_vars_with_cond:";
	    List.iter (fun i -> prerr_string " "; prerr_term i) current_state_terms;
	    prerr_newline());

	 let if_var_list = ref (get_step_var_list current_state_terms (ref [])) in

	 let if_lhs =
 	   (State(string_id#get_name role_id,
		  session_id,
		  current_state_terms))
	   ::lhs_msg (* do not add: @lhs_cond *)
	 in

	 (* generating if rhs *)
	 let rhs_msg = get_iknows rhs
	 in
	 let rhs_cond = get_if_cond rhs in

	 (* updating fresh terms changed by the receipt of a message*)
	 let current_state_terms =
           update_received_fresh_terms generic_state_vars lhs current_state_terms
	 in
	 if (!debug_level >= 4) then
	   (prerr_string "  update_received_fresh_terms: ";
	    List.iter (fun i -> prerr_string " "; prerr_term i) current_state_terms;
	    prerr_newline());

	 (* updating fresh terms changed by the generation of a new message to send *)
	 let current_state_terms =
           update_fresh_terms exists_var_list generic_state_vars current_state_terms
(*        update_fresh_terms rhs_msg generic_state_vars current_state_terms *)
	 in
	 if (!debug_level >= 4) then
	   (prerr_string "  update_fresh_terms: ";
	    List.iter (fun i -> prerr_string " "; prerr_term i) current_state_terms;
	    prerr_newline());

	 (* updating the values of variables used as left member of an assignment *)
	 let (subst2,current_state_terms) =
           update_rhs_primed_vars subst rhs current_state_terms generic_state_vars
	 in
	 if (!debug_level >= 4) then
	   (prerr_string "  update_rhs_primed_vars: ";
	    List.iter (fun (i,j) -> prerr_string "  "; prerr_term i; prerr_string "<-"; prerr_term j) subst2;
	    prerr_newline();
	    prerr_string "               ";
	    List.iter (fun i -> prerr_string " "; prerr_term i) current_state_terms;
	    prerr_newline());

	 let rhs_msg = update_rhs_atoms subst2 rhs_msg in
	 let rhs_cond = update_rhs_atoms subst2 rhs_cond in

	 (* computing the goal terms that should be added to the IF RHS (secret,witness...) *)
(*         let goal_terms = 
	   (get_secrecy_terms (get_primed_vars (lhs@rhs)) goals current_state_terms)
	   @(get_authentication_terms role_id exists_var_list goals)
         in*)
	 (* updating the values of variables used as left member of an assignment *)
(*         let goal_terms =
           update_assigned_vars rhs goal_terms
         in*)

	 (* computing the list of variables that will be used as state fact parameters*)
	 if_var_list :=
 	   get_step_var_list current_state_terms
	     if_var_list;

	 let (deleteFacts,rhs_cond) =
	   List.fold_left
	     (fun (d,r) s ->
	       match s with
		 Ifdelete _ -> (d@[s],r)
	       | _ -> (d,r@[s]))
	     ([],[]) rhs_cond
	 and (inFacts,lhs_cond) =
	   List.fold_left
	     (fun (d,r) s ->
	       match s with
		 Ifcontains _ -> (d@[s],r)
	       | _ -> (d,r@[s]))
	     ([],[]) lhs_cond
	 in

	 (* building the RHS of IF rules *)
	 let if_rhs =
 	   (State(string_id#get_name role_id,
		  session_id,
		  current_state_terms))
	   ::rhs_msg@rhs_cond(*@goal_terms*)
	 in

	 let if_reactions =
	   if deleteFacts = [] then
	     [new basic_if_rule (if_lhs@inFacts) (if_rhs@inFacts) lhs_cond !if_var_list exists_var_list]
	   else
(***
	     let rec combineDeletions m c = function
		 s::ls ->
		   (combineDeletions (m@[s]) c ls)
		   @ (combineDeletions m (c@[Ifnot(s)]) ls)
	       | [] ->
		   (* building the IF rule *)
		   [new basic_if_rule m if_rhs c !if_var_list exists_var_list]
	     in
***)
	     let rec combineAllDeletions lf lc rf = function
		 (Ifdelete(s1,s2) as s)::ls ->
		   if inFacts = [] then
		     (combineAllDeletions (remove_duplicate (lf@[s])) lc rf ls)
		     @ (combineAllDeletions lf (remove_duplicate (lc@[Ifnot(s)])) rf ls)
		   else
		     let rec combineAllIns lf lc rf = function
			 (Ifcontains(t1,t2) as t)::lt ->
			   let iss1equalt1 = (s1 = t1)
			   and iss2equalt2 = (s2 = t2)
			   and s1equalt1 = Ifequal(s1,t1)
			   and s2equalt2 = Ifequal(s2,t2)
			   in
			   (combineAllIns (remove_duplicate (lf@[s]))
			      (remove_duplicate (lc
			                         @ (if iss1equalt1 then [] else [s1equalt1])
			                         @ (if iss2equalt2 then [] else [s2equalt2])))
			      rf lt)
			   @ (if iss1equalt1 then []
			      else
			        combineAllIns (remove_duplicate (lf@[s;t]))
                                              (remove_duplicate (lc@[Ifnot(s1equalt1)]))
                                              (remove_duplicate (rf@[t]))
			                      lt)
			   @ (if iss2equalt2 then []
			      else
			        combineAllIns (remove_duplicate (lf@[s;t]))
                                              (remove_duplicate (lc@[Ifnot(s2equalt2)]))
                                              (remove_duplicate (rf@[t]))
			                      lt)
			   @ (if iss1equalt1 && iss2equalt2 then []
			      else
			        combineAllIns (remove_duplicate (lf@[t]))
                                              (remove_duplicate (lc@[Ifnot(s)]))
			                      (remove_duplicate (rf@[t]))
			                      lt)
		       | _ ->
			   combineAllDeletions lf lc rf ls
		     in
		     combineAllIns lf lc rf inFacts
	       | _ ->
		   (* building the IF rule *)
		   if List.exists (fun p -> List.mem (Ifnot(p)) lc) lf then
		     []
		   else
		     [new basic_if_rule lf rf lc !if_var_list exists_var_list]
	     in
	     combineAllDeletions if_lhs lhs_cond if_rhs deleteFacts
	 in
	 List.map
	   (fun if_reaction ->
	     (role_id,((hlpsl_reaction#get_label),if_reaction)))
	 if_reactions)
       hlpsl_reactions)
;;
