
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
(* ..................................................................
    Module.......: Utils_unify
    Initial Date : 18 March 2004
    Last Update..: 09 March 2006
    Developer....: Mathieu Turuani
    Maintenance..: Judson Santiago
    Compiler.....: OCAML 3.07 
    Comments.....: Some util definitions and functions
    Project .....: The CL-Atse Protocol Analyser, AVISPA.
 .................................................................... *)

open Types_unify
open ErrorHandler_unify


(* ==================================
   Variables for command line parsing 
   ================================== *)

let protocol_names : string list ref = ref [];;

let nb_steps = ref 0;;      
let nb_help = "n   Maximum number of role iterations (or loops, default is 3)";;

let light_unif = ref false;;
let light_help = " Use light unification (no partial associativity for pair)";;

let no_type = ref false;;
let no_type_help = "Do not take term types into account";;

let topdown = ref true;;
let td_help = "    Perform depth-first search (default, less memory)";;
let lr_help = "    Perform breath-first search (minimal attack)";;
let short_help = " Perform breath-first search (minimal attack)";;

let out_to_file = ref false;;
let ou_help = "   Write the attack to file.atk";;

let simplifier = ref true;;
let ns_help = "    Do not simplify the IF file (complete trace)";;

let optimiser = ref true;;
let opt_help = "   Try to replace hashing by tokens. (slower or faster dep. on protocol)";;
let noopt_help = " Do not try to replace hashing by tokens. (slower or faster dep. on protocol)";;

let showtable = ref false;;
let ta_help = "   Write the correspondence table (debug)";;

let showparser = ref false;;
let par_help = "   Write the parser output (debug)";;

let verbose = ref false;;
let verbose_help = "     Print more output informations";;

let veryverbose = ref false;;
let veryverbose_help = "    Print more output and term informations";;

let no_execution = ref false;;
let no_execution_help = "Do not search for attacks (debug)";;

let debug_read = ref false;;
let debug_read_help = " Write debuging informations for Read.ml";;

let outdir = ref ".";;
let outdir_help = "d  Output directory to use with -out";;

let bounded_search = ref false;;

let benchmark = ref false;;
let benchmark_help = " Benchmark output format (very consise)."

let nosplit = ref true;;
let nosplit_help = " Use the --split option on hlpsl2if for hlpsl input file."

let column_nb = ref 80;;
let column_nb_help = "n  Number of column in the output (default 80, disabled with -1)."

let states = ref false;;
let states_help = "Shows all internal system states of the analysis."

let version_help = " Print the version number."
let licence_help = " Print the CL-Atse's disclaimer (licence)"


(* ================================
   Storing the properties to verify 
   ================================ *)

let properties : t_attack_state_goal list ref = ref [];;


(* ===========================
   Storing the intruder's atom 
   =========================== *)

(* The function ProtocolIF MUST give the correct value of "intru". *)
let intru_store : term option ref = ref None;;
let intru () = match !intru_store with Some t -> t | None -> raise (Failure "Utils.ml : access to non initialised intru var (bug).\n");;


(* ====================================
   Storing the table of atoms's strings
   ==================================== *)
   
let localtab : tabletype ref = ref [] ;;
let localobjtype : (string * int) list ref = ref [] ;;
let localtabletype : (int * string) list ref = ref [] ;;


(* =======================================================
   Storing the table of role instances (output info only).
   ======================================================= *)

let role_table : (string * string) list ref = ref [] ;;


(* ======================================
   Storing a list of unforgeable elements
   ====================================== *)
   
(* Stores a list of unforgeable keys. This list is optionnal : If unused or unneeded,
   just leave if empty. However, the simplify module will typically fill this list
   with the unforgeable atomic keys found during the simplification process. This
   list is global because it is known from the start and won't be extended.	*)
let not_forgeable : t_unforgeable ref = 
  ref [] ;;


(* Statistics summary *)

let total_states = ref 0 ;;
let reachable_states = ref 0 ;;
let states_per_dot = ref 1000 ;;
let time_sum = ref 0. ;;


(* ===========================================================
   Storing the information that Maybe_Sooner steps are present
   =========================================================== *)
   
let maybe_sooner_present = ref false
;;


(* ===========================================================
   A token used by module optimisation to create new variables
   =========================================================== *)

let token = ref 0
;;


(* =====================================
   name of the protocol beeing processed
   ===================================== *)

let current_protocol = ref "none"
;;



(* ========================================
           Initialisation Function        
   (to call if many protocols are analysed)
   ======================================== *)

let initialisation protocol_name = 
  lineCounter      := 1;
  errorNumber      := 0;
  total_states     := 0;
  reachable_states := 0;
  intru_store      := None;
  properties       := [];
  localtab         := [];
  localobjtype     := [];
  localtabletype   := [];
  role_table       := [];
  not_forgeable    := [];
  maybe_sooner_present := false;
  token            := 0;
  current_protocol := protocol_name;
;;





(* ===============================
   Functions for list manipulation 
   =============================== *)

(*  Remark: 
    let rec f ta i::fin = ...  
    is 2 times faster than 
    it_list (fun ta i -> ... )  *)

(* Merge two given lists. The first list will be the tail and the
   second will be the head of the new list. The elements of the second
   list will be inverted *)
let rec merge tail = function 
  | t::head -> merge (t::tail) head 
  | _ -> tail
;;

(* Interact a function func over two arguments, where the first is an
   initial value and the second is a list. For example If the function
   is the adition operator and the arguments are the number zero and a
   list of numbers, the result will be the addition of all elements of
   the list. This version interacts the list from the first element to
   the last one. This function is the same as List.fold_left *)
let rec it_list func elem = function 
  | head::tail -> it_list func (func elem head) tail  
  | _ -> elem
;;

(* Interact a function func over two arguments, where the first is an
   initial value and the second is a list. For example If the function
   is the adition operator and the arguments are the number zero and a
   list of numbers, the result will be the addition of all elements
   of the list. This version interacts the list from the last element to
   the first one *)
let rec list_it func elem = function 
  | head::tail -> func (list_it func elem tail) head  
  | _ -> elem
;;

(* Apply the given function to every element of the given list,
   generating a new list with the responses *)
let rec flat_map func = function 
  | head::tail -> merge (flat_map func tail) (func head) 
  | _ -> []
;;

(* Apply the given function to every element of the given list,
   generating a new list with the responses in reversed order *)
let rec rev_flat_map func list = function 
  | head::tail -> rev_flat_map func (merge list (func head)) tail 
  | _ -> list
;;

(* it_list2 takes a three argument function "func", an first argument
   "elem", a second argument tas and a pair of lists L1 =
   [a1;b1;c1;d1] and L2 = [a2;b2;c2;d2]. The result will be the
   aplication of "func" to pairs of elements from the lists, like:
   func d1 d2 (func c1 c2 (func b1 b2 (func a1 a2 tas))). The new
   value of tas will depend on the result of the last aplication of
   the function. In fact the function has to return a tas *)
let rec it_list2 func elem tas = function 
  | (a::fa,b::fb) -> it_list2 func elem (func a b tas) (fa,fb) 
  | ([],[]) -> tas 
  | _  -> elem
;;

(* The result of this function is a tuple containg the application of
   the given function func to the tuple (a,b) *)
let double f = fun (a,b) -> (f a, f b);;

(* The function not_filter receives a list of elements "p" that should
   not be filtered, a boolean function "func" that should verify if a
   given element should be filtered or not, and a list of elements to
   verify. It returns a new list containg only the elements that
   should not be filtered *)
let rec not_filter p func = function 
  | t::l -> 
      if func t then 
	not_filter p func l 
      else 
	not_filter (t::p) func l 
  | _ -> p
;;

(* Apply the given function over a list of elements *)
let rec rev_iter func = function 
  | head::tail -> 
      func head; 
      rev_iter func tail 
  | _ -> ()
;;

(* Replace an element "elem" by a list "list" in a list and preserve
   the order. *)
let rec replace elem list = function 
  | head::tail -> 
      if head=elem then 
	list@(replace elem list tail) 
      else 
	head::(replace elem list tail) 
  | [] -> []
;;


(* returns lst if n is in lst and n::lst if n is not in lst *)
let insert (n : 'a) (lst : 'a list) : 'a list =
  if List.mem n lst then 
    lst 
  else 
    n::lst
;;

(* merges l1 and l2 into a new list. 
   elements that are in both lists are only listed once *)
let union l1 l2 =
  let rec help tas l1 l2 =
    match l1 with
    | x::xs -> 
	if List.mem x l2 then 
	  help tas xs l2 
	else 
	  help (x::tas) xs l2
    | [] -> tas
  in
  help l2 l1 l2
;;

(* computes l1 \ l2 as if they were sets *)
let set_minus l1 l2 =
  let rec help tas l1 l2 =
    match l1 with
    | x::xs -> 
	if List.mem x l2 then 
	  help tas xs l2 
	else 
	  help (x::tas) xs l2
    | [] -> tas
  in
  help [] l1 l2
;;




  
(* ======================
   Tools for Exponentials
   ====================== *)
   
let rec merge_first tas = function
  | (a,_)::fin -> merge_first (a::tas) fin
  | []         -> tas
;;

let exp_inverse = function
  | Pos -> Neg
  | Neg -> Pos
;;





(* ===================================
   Modifying and query a partial order
   =================================== *)

exception Contradiction_found;;


(* adds the restriction "n < m" to the already existing po, if the 
   given po follows every chain, i.e. if (n,[..x..]) and (m,[..n..]) is 
   in po then also x is in list of the bigger elements of m, then also
   the result follows every chain. 
   I.e. if a po is constructed only by this function starting with the 
   empty po [] then then you get a valid po *)
let rec add_to_po (po : t_po) n m : t_po =
  let rec help tas n_found = function
    | ((nn,ms) as t)::tail -> 
	if nn=n then
          (* if n is already in the list then add m to the list of the 
	     bigger elements *)
	  help ((nn,insert m ms)::tas) true tail
	else
	  if nn=m then
	    (* if n is in ms we get an error because then nn<n but we want
	       to add n<nn, else we have to add n<mm for every mm in ms *)
	    if List.mem n ms then
	      raise Contradiction_found
	    else
	      it_list (fun po -> add_to_po po n) (help (t::tas) n_found tail) ms
	  else
	    if List.mem n ms then
              (* if nn is less than n, i.e. n is in ms, 
		 then nn is also less than m so add m to ms *)
	      help ((nn,insert m ms)::tas) n_found tail
	    else
	      help (t::tas) n_found tail	
    | [] -> if n_found then tas else (n,[m])::tas
  in
  help [] false po
;;

(* gives the list of elements that are bigger than n *)
let get_majors (po : t_po) n =
  try
    List.assoc n po
  with
    Not_found -> []
;;





(* ================================
   Types and variables/atom numbers
   ================================ *)


(* The number of different atomic types allowed.  The type n 0 is
   reserved for "no type" i.e. the composed type. The type n 1 is
   reserved for "Public/Private Keys" i.e. PInv() *)
let nb_AtmTypes = 16;;

(* Generates a number that will be stored together with the
   Atom/Variable in a tabletype. With this number it is possible to see
   what is the type associated with this Atom/Variable *)
let create_number var_id var_type = 
  var_id * nb_AtmTypes + var_type
;;

(* Function to return the sequence number of a variable in the
   tabletype, given the encoded number var_number *)
let var_id var_number = 
  var_number / nb_AtmTypes
;;

(* Get the type of the given variable *)
let var_type var_number = 
  var_number mod nb_AtmTypes
;;





(* ==============
   Various stuffs
   ============== *)


(* How to print an uper-score. *)
let uper n =
  (String.make n '-') ^ "\n\n" 
(*  (String.make n (char_of_int 175)) ^ "\n" *)
;;






(* =================
   Trace Manadgement
   ================= *)


let add_trace (trace : t_trace) (nom : string) (pas : t_step) (ltr : t_trace) : t_trace = 
  if !veryverbose then 
    let rec get_ifnames : t_trace -> string = function 
	[_,ifstepname,_,_]      -> ifstepname 
      | (_,ifstepname,_,_)::fin -> ifstepname ^ " " ^ (get_ifnames fin) 
      | []                      -> ""
    in (nom,(get_ifnames ltr),pas,[])::trace 
  else 
    merge trace ltr
;;
