/**
Parser generator for AVISPA output (OFMC + ATSE)
@author Erwan HOUSSAY
@since summer 2008
*/

%{
(* Provides access to semantic analysis tools for attack trace generation *)
open Semantics;;
%}

/* Tokens possibly appearing in section "SUMMARY" */
%token SUMMARY
%token SAFE UNSAFE INCONCLUSIVE
%token DETAILS
%token ATTACK_FOUND STRONGLY_TYPED_MODEL TYPED_MODEL UNTYPED_MODEL BOUNDED_NUMBER_OF_SESSIONS BOUNDED_NUMBER_OF_SYMBOLIC_SESSIONS BOUNDED_SEARCH_DEPTH BOUNDED_MESSAGE_DEPTH
%token TIME_OUT MEMORY_OUT NOT_SUPPORTED OVER_APPROXIMATION UNDER_APPROXIMATION

/* Tokens possibly appearing in section "ATTACK_TRACE" */
%token ATTACK_TRACE
%token <string> IDENT ROLE
%token <int> NUMBER
%token <string*int> AGENT
%token <string*int> NONCE_ATSE NONCE_OFMC
%token <string*int> FREE_ATSE FREE_OFMC
%token PROP
%token START EMPTY
%token ARROW COLON
%token L_ROUND R_ROUND L_CURLY R_CURLY
%token DOT UNDERSCORE COMMA
%token EOF

/* Entry point */
%start avispa_output
%type <string> avispa_output

%%

/* Main rule */

avispa_output:
	summary attack_trace	{ (if !dbg_mode then (print_debug $2 ; dbg_mode := false)) ; parsing_reset() ; $2 }
|				{ (if !dbg_mode then (print_debug "NONE" ; dbg_mode := false)) ; parsing_reset() ; "" }
;

/* Rules specifying section "SUMMARY" */

summary:
	SUMMARY result	{  }
;

result:
	conclusive	{  }
|	inconclusive	{  }
;

conclusive:
	SAFE conclusive_details		{ raise Not_unsafe }
|	UNSAFE conclusive_details	{ if not $2 then raise No_attack_found }
;

conclusive_details:
	DETAILS conclusive_explanation more_conclusive_explanation	{ $2 or $3 }
;

more_conclusive_explanation:
	conclusive_explanation more_conclusive_explanation	{ $1 or $2 }
|								{ false }
;

conclusive_explanation:
	ATTACK_FOUND				{ true }
|	STRONGLY_TYPED_MODEL			{ false }
|	TYPED_MODEL				{ false }
|	UNTYPED_MODEL				{ false }
|	BOUNDED_NUMBER_OF_SESSIONS		{ false }
|	BOUNDED_NUMBER_OF_SYMBOLIC_SESSIONS	{ false }
|	BOUNDED_SEARCH_DEPTH			{ false }
|	BOUNDED_MESSAGE_DEPTH			{ false }
;

inconclusive:
	INCONCLUSIVE inconclusive_details	{ raise Not_unsafe }
;

inconclusive_details:
	DETAILS inconclusive_explanation more_inconclusive_explanation	{  }
;

more_inconclusive_explanation:
	inconclusive_explanation more_inconclusive_explanation	{  }
|								{  }
;

inconclusive_explanation:
	TIME_OUT		{  }
|	MEMORY_OUT		{  }
|	NOT_SUPPORTED		{  }
|	OVER_APPROXIMATION	{  }
|	UNDER_APPROXIMATION	{  }
;

/* Rules specifying section "ATTACK TRACE" */

attack_trace:
	ATTACK_TRACE step more_step 	{ $2^$3 }
|					{ "" }
;

more_step:
	step more_step	{ $1^$2 }
|			{ "\n" }
;

step:
	maybe_step_number agents message maybe_properties { string_of_step $2 $3 }
;

maybe_step_number:
	NUMBER DOT	{  }
|			{  }
;

agents:
	agent ARROW agent COLON { string_of_agents $1 $3 }
;

agent:
	AGENT	{ $1 }
|	IDENT	{ match $1 with "i" -> ("i",0) | _ -> raise Parsing.Parse_error }
;

maybe_properties:
	PROP	{  }
|		{  }
;

message:
	specific	{ $1 }
|	general		{ $1 }
;

specific:
	START	{ SPECIFIC }
|	EMPTY	{ SPECIFIC }
;

general:
	maybe_ciphered maybe_paired { GENERAL (string_of_paired $1 $2) }
;

maybe_paired:
	DOT maybe_ciphered maybe_paired	{ PAIRED (string_of_paired $2 $3) }
|					{ END_PAIRED }
;

maybe_ciphered:
	L_CURLY general R_CURLY UNDERSCORE key	{ string_of_ciphered (string_of_general $2) $5 }
|	term					{ $1 }
;

term:
	IDENT args			{ string_of_function $1 $2 }
|	NONCE_ATSE			{ string_of_nonce $1 }
|	FREE_ATSE			{ string_of_free $1 }
|	NONCE_OFMC			{ string_of_nonce $1 }
|	FREE_OFMC			{ string_of_free $1 }
;

args:
	L_ROUND general more_args R_ROUND	{ ARG ((string_of_general $2)^$3) }
|						{ CONST }
;

more_args:
	COMMA general more_args	{ ","^(string_of_general $2)^$3 }
|				{ "" }
;

key:
	term			{ $1 }
|	L_ROUND general R_ROUND	{ (string_of_general $2) }
;
