
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
open Lexing;;

(*

  We define in this module the variables and the functions that will
  be used to display warnings and handle syntax,semantic and compilation errors.
  First, let us define the variables needed to handle errors.
  \enlargethispage{2em}
  
  -- Variable keeping track of the current line number of the specification 
  file being parsed.
  \begin{center}
  type: [val lineCounter : int ref]
  \end{center}
*)
let lineCounter = ref 1;;
let offsetCounter = ref 1;;

(*
   -- Variable keeping track of the number of errors occurred so far
   used in main to stop the execution if some errors have occurred.
  \begin{center}
   type: [val errorsCounter : int ref]
  \end{center}
*)
let errorsCounter = ref 0;; 

(*
   -- Variable keeping track of the number of warnings displayed 
   during the translation and used to remind/notify the user 
   at the end of the IF output.
  \begin{center}
   type: [val warningsCounter : int ref]
  \end{center}
*)
let warningsCounter = ref 0;; 

(*
   -- Variable giving a unique error number to each error.
  \begin{center}
   type: 
   [val errorNumber : int ref]
  \end{center}
*)
let errorNumber = ref 0;; 

(*

   We now define the functions called when handling errors.

  -- Function handling system error, mainly when the input specification file
   cannot be read.
  \begin{center}
   type: 
   [val handleSystemError : string -> 'a]
  \end{center}
*)
let handleSystemError errorMessage =
  prerr_endline ("%% System error: "^errorMessage);
  exit(-1)
;;

(*
   -- Function handling lexer errors. It generates appropriate error messages
   containing the line number where the error has occurred.
  \begin{center}
   type: 
   [val handleLexerError : string -> string -> 'a]
  \end{center}
*)
let handleLexerError lexerBuffer errorMessage =
  prerr_string "%% Lexical error: Line "; prerr_int !lineCounter;
  prerr_string ": ";
  prerr_string errorMessage; prerr_string " ";
  prerr_endline (lexerBuffer);
  exit (-1)
;;

(*
   -- Function handling syntax errors. It  generates error messages
   according to the reported error numbers.
  \begin{center}
   type: 
   [val handleParseError : unit -> 'a]
  \end{center}
*)
let handleParseError() =
  incr errorsCounter;
  prerr_string "%% Syntax error: Line "; prerr_int !lineCounter;
  prerr_string ", Col "; prerr_int (Lexing.lexeme_start !Globals.lexbuf - !offsetCounter);
  prerr_string " (offset "; prerr_int (Lexing.lexeme_start !Globals.lexbuf + 1);
  prerr_string "-"; prerr_int (Lexing.lexeme_end !Globals.lexbuf);
  prerr_endline (", string \""^(Lexing.lexeme !Globals.lexbuf)^"\")");
  prerr_string "%%   Syn.Err("; prerr_int !errorNumber; prerr_string "): ";

  (match !errorNumber with 
    1 -> prerr_endline "invalid beginning of section: missing \"role\" or \"goal\" keyword"
  | 2 -> prerr_endline "missing \"end role\" keyword in transition/composition"
  | 201 -> prerr_endline "invalid composition of roles: missing \"/\\\" or \";\""
  | 3 -> prerr_endline "missing variable identifier"
  | 4 -> prerr_endline "missing \"(\""
  | 5 -> prerr_endline "invalid goal definition: missing \"authentication_on\" or \"weak_authentication_on\""
  | 6 -> prerr_endline "invalid role header: missing \"def=\" keyword"
  | 7 -> prerr_endline "missing \"_\""
  | 701 -> prerr_endline "invalid constants declaration: missing \",\" or \":\""
  | 702 -> prerr_endline "invalid variables declaration: missing \",\" or \":\""
  | 8 -> prerr_endline "missing \"}\""
  | 9 -> prerr_endline "missing \"{\""
  | 901 -> prerr_endline "invalid expression to be encrypted"
  | 10 -> prerr_endline "missing or invalid type in variables declaration"
  | 11 -> prerr_endline "missing \")\""
  | 12 -> prerr_endline "invalid owns declaration in role definition"
  | 13 -> prerr_endline "invalid local declaration in role definition"
  | 14 -> prerr_endline "invalid init declaration in role definition"
  | 15 -> prerr_endline "invalid accept declaration in role definition"
  | 16 -> prerr_endline "invalid transition: missing \"/\\\", label, action or \"end role\" keyword"
  | 161 -> prerr_endline "invalid transition: missing transition label"
  | 17 -> ()
  | 18 -> prerr_endline "invalid transition: missing action (\"--|>\") or reaction (\"=|>\") arrow"
  | 19 -> prerr_endline "missing \".\""
  | 20 -> prerr_endline "invalid expression as second arg of \"in\" (should be a variable)"
  | 202 -> prerr_endline "invalid role header: invalid parameters declaration, or missing \")\""
  | 21 -> prerr_endline "invalid predicate or event in transition"
  | 22 -> prerr_endline "missing \"/\\\" in predicate"
  | 23 -> prerr_endline "missing \"'\" or \"(\" after variable identifier"
  | 241 -> prerr_endline "invalid predicate: missing \"=\", \"<=\" or \"/=\""
  | 242 -> prerr_endline "invalid expression: missing \"=\""
  | 243 -> prerr_endline "invalid assignment expression: missing \":=\""
  | 251 -> prerr_endline "constant sets forbidden in LHS of a transition"
  | 252 -> prerr_endline "constant sets forbidden in goal property"
  | 26 -> prerr_endline "missing right hook"
  | 27 -> prerr_endline "invalid expression or formula"
  | 28 -> prerr_endline "invalid composition declaration"
  | 29 -> prerr_endline "missing \"transition\" keyword"
  | 30 -> prerr_endline "missing \"composition\" keyword"
  | 31 -> prerr_endline "invalid expression/formula after left parenthesis"
  | 311 -> prerr_endline "invalid expression after comma"
  | 312 -> prerr_endline "invalid expression after \"=\""
  | 313 -> prerr_endline "invalid expression after \"<=\""
  | 314 -> prerr_endline "invalid expression after \"/=\""
  | 315 -> prerr_endline "invalid expression after \"}_\""
  | 316 -> prerr_endline "invalid expression after \"{\""
  | 317 -> prerr_endline "invalid expression after \".\""
  | 32 -> prerr_endline "invalid knowledge declaration"
  | 33 -> prerr_endline "invalid played_by declaration"
  | 34 -> prerr_endline "invalid type attribute (\"dy\" or \"ota\" awaited)"
  | 35 -> prerr_endline "invalid declaration of parameters: missing \",\" or \")\""
  | 351 -> prerr_endline "invalid declaration of variables"
  | 352 -> prerr_endline "invalid declaration of constants"
  | 36 -> prerr_endline "invalid const declaration in role definition"
  | 37 -> prerr_endline "missing constant identifier"
  | 38 -> prerr_endline "invalid predicate or variable in (re)action"
  | _ -> prerr_endline "unknown error");
  flush stderr;
;;

(*
   -- Function handling semantic errors. It  generates error messages
   according to the reported error numbers.
  \begin{center}
   type: 
   [val handleSemanticError :
   int * string * string * string * string * string -> unit]
  \end{center}
*)
let handleSemanticError = function
    (0,_,r,_,_,_) -> 
      prerr_endline
      	("%% Semantic error");
      prerr_endline
	("%%   Sem.Err(0): undefined role "^r)
 
  | (1,f,r,_,_,c) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Call "^c);
      prerr_endline
	("%%   Sem.Err(1): "^f^" is not a function")

  | (2,s,r,_,_,c) ->
      (match s with
      	"-1" -> 
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(2a): invalid number of arguments")
      | _ -> 
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c^", Arg "^s);
	  prerr_endline
	    ("%%   Sem.Err(2b): type mismatch"))

  | (4,_,r,tr,_,_) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Transition "^tr);
      prerr_endline
	("%%   Sem.Err(4): missing event")

  | (5,n,r,tr,_,c) ->
      (match n with

	"1i" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Init state");
	  prerr_endline
	    ("%%   Sem.Err(5.1i): type mismatch in assignment")
      | "1c" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(5.1c): type mismatch in cons() arguments")
      | "1t" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Transition "^tr);
	  prerr_endline
	    ("%%   Sem.Err(5.1t): type mismatch in cons() arguments")
      | "1" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err(5.1): type mismatch in cons() arguments")

      | "2i" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Init state");
	  prerr_endline
	    ("%%   Sem.Err(5.2i): invalid number of parameters for function, list or set")
      | "2c" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(5.2c): invalid number of parameters for function, list or set")
      | "2t" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Transition "^tr);
	  prerr_endline
	    ("%%   Sem.Err(5.2t): invalid number of parameters for function, list or set")
      | "2" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err(5.2): invalid number of parameters for function, list or set")

      | "3c" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(5.3c): second argument in cons() should be of type list or set")
      | "3t" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Transition "^tr);
	  prerr_endline
	    ("%%   Sem.Err(5.3t): second argument in cons() should be of type list or set")
      | "3" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err(5.3): second argument in cons() should be of type list or set")
	     
      | "4i" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Init state");
	  prerr_endline
	    ("%%   Sem.Err(5.4i): impossible to check compatibility of types")
      | "4c" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(5.4c): impossible to check compatibility of types")
      | "4t" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Transition "^tr);
	  prerr_endline
	    ("%%   Sem.Err(5.4t): impossible to check compatibility of types")
      | "4" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err(5.4): impossible to check compatibility of types in cons()")

      | "5c" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Call "^c);
	  prerr_endline
	    ("%%   Sem.Err(5.5c): type mismatch in cons()")
      | "5t" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r^", Transition "^tr);
	  prerr_endline
	    ("%%   Sem.Err(5.5t): type mismatch in cons()")
      | "5" ->
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err(5.5): type mismatch in cons()")

      | x -> 
	  prerr_endline
	    ("%% Semantic error: Role "^r);
	  prerr_endline
	    ("%%   Sem.Err("^x^"): unknown error"))
      
  | (10,_,r,tr,_,_) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Transition "^tr);
      prerr_endline
	("%%   Sem.Err(10): \"start\" is not used with a channel")

  | (11,v,"","","","") ->
      prerr_endline
	("%% Semantic error: LTL goal");
      prerr_endline
	("%%   Sem.Err(11g): undeclared variable "^v)
  | (11,v,r,"","","") ->
      prerr_endline
	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(11a): undeclared variable "^v)
  | (11,v,r,"","",c) ->
      prerr_endline
	("%% Semantic error: Role "^r^", Call "^c^
	 ": Sem.Err(11b): undeclared variable "^v)
  | (11,v,r,"",st,"") ->
      prerr_endline
	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(11c): undeclared symbol "^v^" in "^st)
  | (11,v,r,"",st,c) ->
      prerr_endline
	("%% Semantic error: Role "^r^", Call "^c);
      prerr_endline
	("%%   Sem.Err(11d): undeclared symbol "^v^" in "^st)
  | (11,v,r,tr,_,"") ->
      prerr_endline
	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(11e): undeclared variable "^v)
  | (11,v,r,tr,_,c) ->
      prerr_endline
	("%% Semantic error: Role "^r^", Call "^c);
      prerr_endline
	("%%   Sem.Err(11f): undeclared variable "^v)

  | (12,_,r,_,st,_) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Call "^st);
      prerr_endline
	("%%   Sem.Err(12): LHS of assignment is not a variable")

  | (13,var,r,_,_,st) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Call "^st);
      prerr_endline
	("%%   Sem.Err(13): redeclared variable "^var)

  | (14,f,r,_,_,st) -> 
      prerr_endline
      	("%% Semantic error: Role "^r^", Call "^st);
      prerr_endline
	("%%   Sem.Err(14): invalid argument used with "^f)

  | (15,_,r,_,_,_) -> 
      prerr_endline
      	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(15): invalid number of parameters in composition section")

  | (16,msg,line,_,_,_) ->
      prerr_endline
	("%% Semantic error: Line "^line);
      prerr_endline
	("%%   Sem.Err(16): "^msg)

  | (17,_,r,_,_,_) ->
      prerr_endline
 	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(17): player should be declared as agent in parameters")

  | (18,v,r1,r2,_,_) ->
      if r2 <> "" then
	if r1 <> "" then
	  (prerr_endline
	     ("%% Semantic error: Role "^r2);
	   prerr_endline
	     ("%%   Sem.Err(18a): type mismatch with declaration of "^v^" in role "^r1))
	else
	  (prerr_endline
	     ("%% Semantic error: Role "^r2);
	   prerr_endline
	     ("%%   Sem.Err(18b): type mismatch with global declaration of "^v))
      else
	if r1 <> "" then
	  (prerr_endline
	     ("%% Semantic error: Role "^r1);
	   prerr_endline
	     ("%%   Sem.Err(18c): type mismatch with global declaration of "^v))
	else
	  (prerr_endline
	     ("%% Semantic error");
	   prerr_endline
	     ("%%    Sem.Err(18d): type mismatch between global declarations of "^v))

  | (19,v,line,r,_,_) ->
      prerr_endline
	("%% Semantic error: Role "^r);
      prerr_endline
	("%%   Sem.Err(19): Line "^line^": undeclared variable "^v^" in owns section")

  | (20,msg,_,_,_,_) ->
      prerr_endline
	("%% Semantic error");
      prerr_endline
	("%%   Sem.Err(20): "^msg)

  | _ -> ()
;;

(*
   The following function is similar the the previous one except that it does not
   cause the compilation to stop. A warning message is displayed to guide the programmar
   by specifying the problem's context.
  \begin{center}
   type: 
   [val displayWarning : int -> string -> unit]
  \end{center}
*)
let displayWarning msg_type location =
  if !Globals.flag_warnings then
    (incr warningsCounter;
     let string_to_print =
       match msg_type with
	 1->  
	   "%% Internal function "^location
	   ^": warning(1): unhandled use case"
       | 2 -> 
	   "%% Internal function "^location
	   ^": warning(2): invalid variable"
       | 3 -> 
	   "%% Internal function "^location
	   ^": warning(3): invalid use of prime"
       | 4 -> 
	   "%% Internal function "^location^
	   ": warning(4): missing instantiation"
       | 5 -> 
	   "%% Internal function "^location
	   ^": warning(5): invalid fact"
       | 6 -> 
	   "%% System in/out error: warning(6): the prelude file cannot be loaded"
       | 7 -> 
	   "%% HLPSL spec: warning(7): unknown type of constant "^location
       | 8 ->
	   "%% HLPSL spec: warning(8): "^location^": missing event"
       | 9 ->
	   "%% HLPSL spec: warning(9): "^location^": forbidden event"
       | 10 ->
	   "%% Semantical warning: "^location
       | 11 ->
	   "%% HLPSL spec: warning(11): "^location
       | _ ->
	   "%% Warning: "^location
     in
     prerr_endline string_to_print)
;;
