
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
# 7 "lexer.mll"
 

(*

  This module contructs lexical entities and transmits them to the HLPSL parser.
  The keywords are stored in a hashtable. This solution permits to have
  a smaller automaton (see Ocaml Manual 12.7, comon errors). The tokens are
  defined in the \texttt{parser.mly} file.
  \label{lexer:keyword:table}
*)

open Parser;;
exception Eof;;

let keyword_table = Hashtbl.create 68;;


(*
  We now put all reserved word into a hash table, and define functions 
  to access this hash table.
*)

List.iter
  (fun (keyword,token) ->
    Hashtbl.add keyword_table keyword token)
(*
  The following list contains the pairs (keyword,token). For new Ocaml 
  users, note that the tuple constructor is the comma, and the list
  constructor is the semi-colon.
*)
  [
   "accept",ACCEPT;
   "agent",AGENT;
   "authenticates", AUTHENTICATES;
   "authentication_on", AUTHENTICATION_ON;
   "bool",BOOL;
   "channel",CHANNEL;
   "composition",COMPOSITION;
   "cons",CONS;
   "const",CONST;
   "delete",DELETE;
   "dy",DY;
   "end",END;
   "exists",LOCAL;
   "goal",GOAL;
   "hash", HASH;
   "hash_func", HASH_FUNC;
   "function", HASH_FUNC;
   "in",IN;
   "init",INIT;
   "intruder_knowledge",IKNOWLEDGE;
   "inv",INV;
(*   "list",LIST; *)
   "local",LOCAL;
(*   "loop",LOOP;*)
   "message",MESSAGE;
   "nat",NAT;
   "new",NEW;
   "not",NOT;
   "on", ON;
   "ota",OTA;
   "owns",OWNS;
   "played_by",PLAYED_BY;
   "protocol_id",PROTOCOL_ID;
   "public_key",PUBLIC_KEY;
   "role",ROLE;
   "secrecy_of",SECRECY_OF;
   "set",SET;
   "start",START;
   "symmetric_key",SYMMETRIC_KEY;
   "text",TEXT;
   "transition",TRANSITION;
   "weakly",WEAKLY;
   "weak_authentication_on", WEAK_AUTHENTICATION_ON;
 ]
;;


(*
  The two next functions are not optimal, as a keyword will often be
  put twice in uppercase. However, as long as we have no stable
  version, it seems better to ensure that keywords are matched in
  uppercase. 
*)

let is_keyword identifier =
  Hashtbl.mem keyword_table (String.lowercase identifier);;

let get_keyword identifier =
  Hashtbl.find keyword_table (String.lowercase identifier);;
  
(*
  We think this is a good idea to prefix all variables and constants
  in order to avoid confusion with the IF. It could be in principle
  possible to forbid the use of confusing identifiers, but this would
  imply the end-user also have to know the IF, and we do not want that. 

  Thus, we prefix identifiers with the value of
  [identifierPrefix]. We use the [prefixName] fonction to prefix only
  those identifiers which will lead to confusion (and thus permits a
  better readability of the IF in most
  cases.)\label{lexer:identifier:prefix} \label{lexer:prefix:name}
*)

let identifier_prefix = "id";;
  
let prefix_name name =
  match name with
  | "apply"
  | "attack_state"
  | "attack_states"
  | "contains"
  | "crypt"
  | "dummy_agent"
  | "dummy_bool"
  | "dummy_chnl"
  | "dummy_chnl_dy"
  | "dummy_chnl_ota"
  | "dummy_hash"
  | "dummy_msg"
  | "dummy_nat"
  | "dummy_nonce"
  | "dummy_pk"
  | "dummy_set"
  | "dummy_sk"
  | "equal"
  | "equations"
  | "function"
  | "initial_state"
  | "inits"
  | "intruder"
  | "leq"
  | "pair"
  | "properties"
  | "property"
  | "rules"
  | "scrypt"
  | "section"
  | "step"
  | "types" -> (identifier_prefix^name)
  | _ -> name
;;

(*
  We have used a hashtable to store the reserved words 
  (see~\ref{lexer:keyword:table}). First, we look in this
  table whether the parsed word is a reserved keyword. Then, 
  if this not the case, we associate the word to either a 
  variable or a constant following the prolog-convention of
  an uppercase first letter for variables.
*)
let handle_string identifier = 
  if is_keyword identifier then
    get_keyword identifier
  else
    CONST_IDENT(prefix_name identifier);;

# 160 "lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\220\255\221\255\079\000\160\000\224\255\225\255\227\255\
    \228\255\229\255\231\255\232\255\233\255\234\255\236\255\017\000\
    \048\000\016\000\242\255\078\000\079\000\235\000\003\000\080\000\
    \157\000\054\001\127\000\254\255\255\255\249\255\014\000\032\000\
    \251\255\248\255\034\000\250\255\247\255\064\001\139\001\246\255\
    \081\000\245\255\239\255\243\255\244\255\241\255\078\000\240\255\
    \237\255\001\000\238\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\033\000\032\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\020\000\
    \035\000\035\000\255\255\035\000\035\000\033\000\025\000\029\000\
    \035\000\003\000\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\033\000\033\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\026\000\000\000\000\000\000\000\255\255\255\255\
    \000\000\000\000\255\255\000\000\000\000\255\255\255\255\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\028\000\027\000\000\000\000\000\028\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \028\000\000\000\000\000\000\000\000\000\026\000\018\000\006\000\
    \015\000\013\000\050\000\007\000\010\000\024\000\008\000\019\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\025\000\022\000\009\000\020\000\023\000\049\000\045\000\
    \036\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\016\000\017\000\046\000\032\000\005\000\
    \035\000\003\000\003\000\003\000\021\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\012\000\040\000\011\000\014\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\255\255\031\000\044\000\041\000\047\000\033\000\042\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\043\000\048\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\030\000\000\000\034\000\000\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\000\000\029\000\000\000\000\000\000\000\000\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\000\000\000\000\000\000\000\000\004\000\
    \002\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \037\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\025\000\025\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\038\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \039\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\049\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\015\000\017\000\
    \022\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\016\000\031\000\000\000\
    \034\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\020\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\026\000\030\000\019\000\020\000\016\000\023\000\040\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\019\000\046\000\255\255\255\255\003\000\255\255\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\024\000\255\255\023\000\255\255\255\255\255\255\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\255\255\024\000\255\255\255\255\255\255\255\255\
    \255\255\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\255\255\255\255\255\255\255\255\004\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\255\255\255\255\
    \255\255\255\255\021\000\255\255\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\025\000\025\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\255\255\255\255\255\255\255\255\255\255\026\000\
    \255\255\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\255\255\255\255\255\255\255\255\037\000\
    \255\255\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\255\255\255\255\255\255\
    \038\000\255\255\255\255\255\255\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\255\255\255\255\
    \255\255\255\255\038\000\255\255\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "";
  Lexing.lex_backtrk_code = 
   "";
  Lexing.lex_default_code = 
   "";
  Lexing.lex_trans_code = 
   "";
  Lexing.lex_check_code = 
   "";
  Lexing.lex_code = 
   "";
}

let rec token lexbuf =
    __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 176 "lexer.mll"
                       (token lexbuf)
# 373 "lexer.ml"

  | 1 ->
# 177 "lexer.mll"
                  (incr Error_handler.lineCounter;
		   Error_handler.offsetCounter := Lexing.lexeme_end lexbuf;
		   token lexbuf)
# 380 "lexer.ml"

  | 2 ->
# 185 "lexer.mll"
                  (token lexbuf)
# 385 "lexer.ml"

  | 3 ->
# 190 "lexer.mll"
                  (NAT_IDENT(Lexing.lexeme lexbuf))
# 390 "lexer.ml"

  | 4 ->
# 191 "lexer.mll"
                  (ACTION_ARROW)
# 395 "lexer.ml"

  | 5 ->
# 192 "lexer.mll"
                  (REACTION_ARROW)
# 400 "lexer.ml"

  | 6 ->
# 193 "lexer.mll"
                  (FUNCTION_ARROW)
# 405 "lexer.ml"

  | 7 ->
# 194 "lexer.mll"
                  (IMPLIES)
# 410 "lexer.ml"

  | 8 ->
# 195 "lexer.mll"
                  (ASSIGN)
# 415 "lexer.ml"

  | 9 ->
# 196 "lexer.mll"
                  (DEF_EQUAL)
# 420 "lexer.ml"

  | 10 ->
# 197 "lexer.mll"
                  (LEQ)
# 425 "lexer.ml"

  | 11 ->
# 198 "lexer.mll"
                  (NEQ)
# 430 "lexer.ml"

  | 12 ->
# 199 "lexer.mll"
                  (AND)
# 435 "lexer.ml"

  | 13 ->
# 200 "lexer.mll"
                  (AND)
# 440 "lexer.ml"

  | 14 ->
# 201 "lexer.mll"
                  (OR)
# 445 "lexer.ml"

  | 15 ->
# 202 "lexer.mll"
                  (LTL_op1)
# 450 "lexer.ml"

  | 16 ->
# 203 "lexer.mll"
                  (LTL_op2)
# 455 "lexer.ml"

  | 17 ->
# 204 "lexer.mll"
                  (LTL_op3)
# 460 "lexer.ml"

  | 18 ->
# 205 "lexer.mll"
                  (LTL_op4)
# 465 "lexer.ml"

  | 19 ->
# 206 "lexer.mll"
                  (LTL_op5)
# 470 "lexer.ml"

  | 20 ->
# 207 "lexer.mll"
                  (L_PAR)
# 475 "lexer.ml"

  | 21 ->
# 208 "lexer.mll"
                  (R_PAR)
# 480 "lexer.ml"

  | 22 ->
# 213 "lexer.mll"
                  (L_BRACE)
# 485 "lexer.ml"

  | 23 ->
# 214 "lexer.mll"
                  (R_BRACE)
# 490 "lexer.ml"

  | 24 ->
# 215 "lexer.mll"
                  (COMMA)
# 495 "lexer.ml"

  | 25 ->
# 216 "lexer.mll"
                  (COLON)
# 500 "lexer.ml"

  | 26 ->
# 217 "lexer.mll"
                  (SEMI_COLON)
# 505 "lexer.ml"

  | 27 ->
# 218 "lexer.mll"
                  (DOT)
# 510 "lexer.ml"

  | 28 ->
# 219 "lexer.mll"
                  (PLUS)
# 515 "lexer.ml"

  | 29 ->
# 220 "lexer.mll"
                  (EQUAL)
# 520 "lexer.ml"

  | 30 ->
# 221 "lexer.mll"
                  (PRIME)
# 525 "lexer.ml"

  | 31 ->
# 222 "lexer.mll"
                  (UNDERSCORE)
# 530 "lexer.ml"

  | 32 ->
# 224 "lexer.mll"
                  (VAR_IDENT (Lexing.lexeme lexbuf))
# 535 "lexer.ml"

  | 33 ->
# 226 "lexer.mll"
                  (handle_string (Lexing.lexeme lexbuf))
# 540 "lexer.ml"

  | 34 ->
# 231 "lexer.mll"
        (raise Eof)
# 545 "lexer.ml"

  | 35 ->
# 232 "lexer.mll"
                  (Error_handler.handleLexerError 
		     (Lexing.lexeme lexbuf) 
		     " unknown symbol"; 
		   exit (-1))
# 553 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_token_rec lexbuf __ocaml_lex_state

;;

