
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)

(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche, Thomas Genet and Erwan Houssay

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.5. Last modification 23/09/08
*)
(**
Main module for attack trace generation
Makes it possible to launch <report2trace> execution
@author Erwan HOUSSAY
@since summer 2008
*)

open Globals
open Error_handler
open Atkiflexer
open Atkifparser
open Semantics
open Filename


(** Retrieve verification tools binaries pathes *)

(* For using files with pathes containing spaces we need an escape character *)

let os=Sys.os_type;;
let esc= if os="Unix" then "" else "\"";;

(* If avispa tools are already installed then we use path to those tools and AVISPA_PACKAGE variable 
   otherwise we use tools included in span package *)

let path_span = 
  (try 
    Sys.getenv "SPAN"
  with Not_found -> failwith "Environment Variable SPAN should be declared!" )

let path_avispa =
  (try
    Sys.getenv "AVISPA_PACKAGE"
  with
    Not_found -> path_span)

(* backend names are different depending on the execution platform, Unix/Windows *)

let hlpsl2if =
if os="Unix" then
   Filename.quote (path_avispa^"/bin/translator/hlpsl2if" : string)
else
   Filename.quote (path_avispa^"\\bin\\translator\\hlpsl2if.exe" : string);;

(* Cannot use directly hlpsl2if because filenames are quoted *)

if os="Unix" then
  (if not (Sys.file_exists (path_avispa^"/bin/translator/hlpsl2if"))
   then failwith "Check your environment variables ($AVISPA_PACKAGE and $SPAN): hlpsl2if is not reachable")
else
  (if not (Sys.file_exists (path_avispa^"\\bin\\translator\\hlpsl2if.exe")) 
   then failwith "Check your environment variables ($AVISPA_PACKAGE and $SPAN): hlpsl2if is not reachable")


(** Define global variables for execution settings *)

(** Directory for the temporary files generated by the tools and for the output file *)
let output_dir = 
  if os="Unix" then (path_avispa^"/testsuite/results/")
  else (path_avispa^"\\testsuite\\results\\")

(** HLPSL input file path *)
let hlpsl_file = ref "";;
(** Verification input file path *)
let verif_file = ref "";;
(** Trace output file path *)
let trace_file = ref "";;
(** Execution report file path *)
let err_file = ref "";;


(** Parses command line call to trace generator *)
let is_input = ref false in
(Arg.parse
   [
    (
     "--in",
     Arg.String
       (fun path -> verif_file := path
       ),
     " Set verification input file"
    );
    (
     "--out",
     Arg.String
       (fun path -> trace_file := path
       ),
     " Set output file ; default is stdout"
    );
    (
     "--err",
     Arg.String
       (fun path -> err_file := path
       ),
     " Set execution report file ; default is stdout"
    );
    (
     "--dbg",
     Arg.Unit
       (fun () -> dbg_mode := true
       ),
     " Print on standard output information collected and built throughout the parsing"
    )
  ]
   (fun s ->
     if !is_input then
       raise (Arg.Bad "error: multiple input files")
     else
       (is_input := true;
        hlpsl_file := s))
   "Usage: report2trace [options] [file.hlpsl]\n\nOptions:"
);;


(** Converts input file name into output files names *)
let convert_file_name (suffix : string) : string =
	if (Filename.check_suffix (!hlpsl_file) ".hlpsl") 
	then (output_dir^(Filename.basename ((Filename.chop_suffix (!hlpsl_file) ".hlpsl")^suffix)))
	else (output_dir^(Filename.basename ((!hlpsl_file)^suffix)));;


(* *********************************************************************************************** *)

(** Compiles the IF file first, then the verification tool report file,
resulting in the attack trace file to be loaded as an MSC in the animator *)

(************)
(* 1ST STEP *)
(************)

(* Defines output file pathes for avispa verification *)
let if_file = convert_file_name ".if" in

(* Cleans files in case they already exist *)
if (Sys.file_exists if_file) then ignore(Sys.remove if_file) ;

(* Opens input channel and sets global pathes for HLPSL compilation *)
let hlpsl_channel =
	try
		Globals.input_file := !hlpsl_file ;
		Globals.output_dir := (Filename.dirname if_file)^"/" ;
		Globals.output_file := Filename.basename if_file ;
		Globals.tmp_file := if_file ;
		open_in !Globals.input_file
	with
		Sys_error(f) -> Error_handler.handleSystemError(f)
in

(* Compiles HLPSL to IF *)
(try
  prerr_endline ("%% Translation of "^(!Globals.input_file)) ;
  (* Fills in input buffer *)
  Globals.lexbuf := Lexing.from_channel hlpsl_channel ;
  (* Runs compilation *)
  let specif = Parser.specification Lexer.token !Globals.lexbuf in
  (* Invokes semantic check *)
  Semantic_check.check_all specif ;
  (* Instantiates parsed roles *)
  let role_instances = Instantiation.call_instantiation specif in
  (* Maps role player instances *)
  Semantics.build_role_inst_table role_instances ;
  (* Stores identifiers types *)
  Semantics.build_ids_type_table () ;
  (* Prints IF file *)
  Interface.print_if role_instances specif#get_goals ;
  (* Handles errors *)
with
  Parsing.Parse_error ->
    Error_handler.handleParseError();
    exit 1
| Lexer.Eof ->
    prerr_endline "%% Fatal error: unexpected end of file!";
    Error_handler.handleParseError();
    exit 1
| _ ->
    prerr_string "%% ";
    exit 1
) ;

(************)
(* 2ND STEP *)
(************)

(* Cleans trace generation output file in case it already exists *)
if((!trace_file)<>"") then ( if (Sys.file_exists !trace_file) then ignore(Sys.remove !trace_file) ) ;

(* Opens input and output channels for attack trace compilation *)
let atk_channel = if ((!verif_file)<>"") then (open_in !verif_file) else (raise (Arg.Bad "error: no verification file provided"))
and trc_channel = if ((!trace_file)<>"") then (open_out !trace_file) else (stdout)
and err_channel = if ((!err_file)<>"") then (open_out !err_file) else (stdout)
in

(* Cleans temporary channels used for trace generation *)
let (clean_up: unit -> unit) = function () ->
	close_in atk_channel ;
	flush trc_channel ;
	close_out trc_channel ;
	flush err_channel ;
	close_out err_channel
in

(* Compiles attack trace to simulation trace *)
try
	(* Compiles AVISPA output so as to generate simulation trace *)
	let attack_trace = avispa_output adaptative_lexing (Lexing.from_channel atk_channel) in
	(* Prints trace generation result *)
	output_string trc_channel attack_trace ;
	clean_up()
(* Handles errors by printing an error message, cleaning resources, and terminating execution with an error code *)
with x ->
	(match x with
		Atkiflexer.Lexicon_summary x ->
			output_string err_channel ("%% Unknown lexeme: "^(String.make 1 x)^" in section summary\n")
	|	Atkiflexer.Lexicon_attack x ->
			output_string err_channel ("%% Unknown lexeme: "^(String.make 1 x)^" in section attack\n")
	|	Atkiflexer.Tool_undefined ->
			output_string err_channel "%% Tool undefined\n"	
	|	Semantics.Unknown_inst ->
			output_string err_channel "%% Unrecognized agent\n"
	|	Semantics.Not_unsafe ->
			output_string err_channel "%% Protocol verification result was not \"UNSAFE\"\n%% See AVISPA output at section \"SUMMARY\"\n"
	|	Semantics.No_attack_found ->
			output_string err_channel "%% Explanation did not mention \"ATTACK_FOUND\" while result was \"UNSAFE\"\n%% See AVISPA output at section \"DETAILS\"\n"
	|	Semantics.Inconsistent_msg ->
			output_string err_channel "%% Inconsistent message detected\n"
	|	Parsing.Parse_error ->
			let loc = match !Atkiflexer.last_run with
				RUN_NONE	-> "at unknown location"
			|	RUN_SKIP	-> "in AVISPA output file"
			|	RUN_SUMMARY	-> "in verification summary"
			|	RUN_ATTACK	-> "in section attack"
			in
			output_string err_channel ("%% Syntax error: "^loc^"\n")
	|	_ ->
			output_string err_channel ("%% An unhandled error occured during trace generation\n")
	) ;
	clean_up() ;
	exit 1
;;
