
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
(**
Defines global stuff regarding semantic analysis for attack trace generation
@author Erwan HOUSSAY
@since summer 2008
*)

(* EXCEPTIONS *)

(** Points out that some agent could not be found in role instance table *)
exception Unknown_inst;;
(** Points out that a specific message contained in a general message has been detected *)
exception Inconsistent_msg;;
(** Points out that verification result was "SAFE" or "INCONCLUSIVE" in AVISPA output *)
exception Not_unsafe;;
(** Points out that explanation did not mention "ATTACK_FOUND" while result was "UNSAFE" in AVISPA output *)
exception No_attack_found;;

(* TYPES *)

(** Describes agents taking part to secured protocols *)
type agent = string*int

(** Gives messages categories *)
type message =
	GENERAL of string	(** general informative message *)
  |	SPECIFIC		(** specific message to be ignored *)
;;
(** Spots terms to be paired *)
type maybe_paired =
	PAIRED of string	(** term to be paired with the previous one *)
  |	END_PAIRED		(** no more terms to be paired *)
;;
(** Distinguishes simple identifiers from functions *)
type maybe_function =
	ARG of string	(** function with parameters *)
  |	CONST		(** simple identifier *)
;;
(** Role played by the intruder for a step *)
type interlocutors =
	INTRUDER_BOTH		(** intruder is both sender and addressee *)
  |	INTRUDER_ABS of string	(** intruder does not take part to step *)
  |	INTRUDER_RCV of string	(** intruder is addressee *)
  |	INTRUDER_SND of string	(** intruder is sender *)
;;

(* GLOBAL VARIABLES *)

(** Associates to each identifier parsed its corresponding type *)
let ids_type_table = ref [];;

(** Associates role player identifiers and role instance numbers with corresponding role names *)
let role_inst_table = ref [(("i",0),"Intruder_")];;

(** Lists parsed nonces along with their string for printing *)
let nonce_table = ref [];;

(** Lists parsed free variables along with their string for printing *)
let free_table = ref [];;

(** Number to be given to the next nounce detected *)
let nonce_count = ref 1;;

(** Number to be given to the next free variable detected *)
let free_count = ref 1;;

(** Debug mode flag *)
let dbg_mode = ref false;;

(* FUNCTIONS *)

(**
Builds table of role player instances, associating role player identifier and role instance number with corresponding role name.
@param Types.role_instance list : list of instantiated roles
*)
let (build_role_inst_table: Types.role_instance list -> unit) = function role_instances ->
	(* get initial states out of role instances *)
	let init_states = List.flatten (List.map (fun role_instance -> role_instance#get_initial_state) role_instances) in
	(* collect corresponding role names, role instance numbers and role player identifiers out of role instances' initial states *)
	let rec pick_out_players = function
		[]			-> []
	  |	(Types.State (role_name,_,term_list))::tail	->
			(match (List.hd term_list,List.hd (List.rev term_list)) with
				Types.Base(Types.Var(player_id)),Types.Base(Types.Var(player_number))	-> 
					let id = Interface.if_value_to_string Interface.atom_ground_value_to_string (Types.If_address(player_id))
					and nb = Interface.if_value_to_string Interface.atom_ground_value_to_string (Types.If_address(player_number))
					in ((id,int_of_string nb),role_name)::(pick_out_players tail)
			|	_ -> pick_out_players tail)
	  |	_::tail			-> pick_out_players tail
	in
	(* store table of role player instances in a global variable *)
	role_inst_table := (!role_inst_table)@(pick_out_players init_states)
;;

(**
Builds table of identifiers, associating to each identifier its type.
*)
let (build_ids_type_table: unit -> unit) = function () ->
	let role_ids = 0::(List.map snd !Globals.player_to_role) in
	Hashtbl.iter
		(fun ident id_type ->
			if (Globals.not_excluded_type id_type) && (List.mem (Globals.string_id#get_role_id ident) role_ids) then
			ids_type_table := (Globals.string_id#get_name ident,id_type)::(!ids_type_table))
		Globals.type_table#get_tab
;;

(**
Converts identifier types into suitable strings
@param Types.term_type : identifier type
@return string : shortened word corresponding to the identifier type
*)
let (string_of_type: Types.term_type -> string) = function
	Types.Base Types.Text		-> "nonce"
  |	Types.Base Types.Agent		-> "agent"
  |	Types.Base Types.Message	-> "msg"
  |	Types.Base Types.Public_key	-> "pk"
  |	Types.Base Types.Symmetric_key	-> "sk"
  |	_				-> "other"
;;

(**
Converts a nonce into the corresponding string
@param string*int : nonce parsed
@return string : corresponding string
*)
let (string_of_nonce: (string*int) -> string) = fun (id,nb) ->
	try
		List.assoc (id,nb) !nonce_table
	with
		Not_found ->
			let tid = try string_of_type (List.assoc id !ids_type_table) with Not_found -> "nonce" in
			let format = tid^"-"^(string_of_int !nonce_count) in
				(let type_format = List.assoc id !ids_type_table in (*match type_format with AUTRE_TYPE -> () | _ ->*) ids_type_table := (!ids_type_table)@[(format,type_format)]) ;
				nonce_table := ((id,nb),format)::!nonce_table ; nonce_count := !nonce_count + 1 ; format
;;

(**
Converts a free variable into the corresponding string
@param string*int : free variable parsed
@return string : corresponding string
*)
let (string_of_free: (string*int) -> string) = fun (id,nb) ->
	try
		List.assoc (id,nb) !free_table
	with 
		Not_found ->
			let format = "var-"^id^"-"^(string_of_int !free_count) in
				free_table := ((id,nb),format)::!free_table ; free_count := !free_count + 1 ; format
;;

(**
Searches role_inst_table for the role corresponding to a given agent.
@param agent : agent whose role name is looked for
@param (agent*string) list : table to search
@return string*int : role name and instance number corresponding to given agent
*)
(*
In order to deal with OFMC output inconsistency,
the record returned from the table is that which contains the given identifier
associated with the lowest number greater than or equal to the given number,
the role player instances tables being sorted by increasing number.
*)
let rec (search_role: agent -> (agent*string) list -> string*int) = function (a,p) -> function
  	[]					-> raise Unknown_inst
  |	((b,n),role)::l when ( b=a && n>=p )	-> (role,n)
  |	x::l					-> search_role (a,p) l
;;

(**
Converts agent identifier into role name
and builds suitable string for attack trace
@param agent : agent taking part to a step
@return string : attack trace suitable string correponding to the given agent
*)
let (string_of_instance: agent -> string) = fun (id,nb1) ->
	let (role,nb2) = search_role (id,nb1) !role_inst_table  in
		"("^role^", "^(string_of_int nb2)^")"
;;

(**
Pairs sender and addressee according to part taken by intruder
@param agent : sender 
@param agent : addressee
@return interlocutors : paired and tagged agents
*)
let (string_of_agents: agent -> agent -> interlocutors) = fun snd rcv ->
	match (snd,rcv) with
		("i",_),("i",_)	-> INTRUDER_BOTH
	|	("i",0),_	-> INTRUDER_SND  ((string_of_instance snd)^" -> "^(string_of_instance rcv)^" : ")
	|	_,("i",0)	-> INTRUDER_RCV ((string_of_instance snd)^" -> "^(string_of_instance rcv)^" : ")
	|	_		-> INTRUDER_ABS  ((string_of_instance snd)^" -> "^(string_of_instance rcv)^" : ")
;;

(**
Prints term either as a single term or as paired with a second message
@param string : term given
@param maybe_paired : possibly second part
@return string : corresponding string, either simple or paired
*)
let (string_of_paired: string -> maybe_paired -> string) = function term -> function
	END_PAIRED	-> term
  |	PAIRED msg	-> "pair("^term^","^msg^")"
;;

(**
Builds string corresponding to a ciphered message
@param string : content
@param string : key
@return string : ciphered message ready for attack trace
*)
let (string_of_ciphered: string -> string -> string) = fun msg key ->
	try
		match List.assoc key !ids_type_table with
			Types.Base Types.Public_key	-> "crypt("^key^","^msg^")"
		|	_				-> "scrypt("^key^","^msg^")"
	with Not_found ->
		try
			match String.sub key 0 4 with
				"inv("	-> "crypt("^key^","^msg^")"
			|	_	-> "scrypt("^key^","^msg^")"
		with Invalid_argument _ ->
			"scrypt("^key^","^msg^")"
;;

(**
Builds string corresponding to an identifier or a function
@param string : identifier
@param maybe_function : possibly function arguments
@return string : identifier or function ready to be printed for attack trace
*)
let (string_of_function: string -> maybe_function -> string) =  function ident -> function
	CONST		-> ident
  |	ARG msg		-> match ident with
		("inv"|"exp"|"xor")	-> ident^"("^msg^")"
	  |	_			-> "apply("^ident^","^msg^")"
;;


(**
Gets string contained in an informative message
@param message : either informative or not informative message
@return string : content in case given message is general
@raise Inconsistent_msg : in case given message is specific
*)
let (string_of_general: message -> string) = function
	GENERAL msg	-> msg
  |	SPECIFIC		-> raise Inconsistent_msg
;;

(**
Prints line corresponding to an attack trace step in case it contains an informative message
@param interlocutors : well-printed agents taking part to this step
@param message : message sent on this step
@return string : full step for attack simulation trace
*)
let (string_of_step: interlocutors -> message -> string) = function agents -> function
	SPECIFIC	-> ""
  |	GENERAL msg	->
	match agents with
		INTRUDER_BOTH		-> ""
	|	INTRUDER_ABS inter	-> "\n"^inter^"x("^msg^",_gen_trc)"
	|	INTRUDER_RCV inter	-> "\n"^inter^"x("^msg^",_gen_trc)"
	|	INTRUDER_SND inter	-> "\nadd_in_set((Intruder_, 0),"^msg^",(Intruder_, 0),Set_knowledge)\n"^inter^"x("^msg^",_gen_trc)"
;;

(**
Reset global variables
*)
let (parsing_reset: unit -> unit)  = function () ->
	role_inst_table := [(("i",0),"Intruder_")] ;
	nonce_table := [] ;
	free_table := [];
	nonce_count := 1 ;
	free_count := 1 ;
	ids_type_table := [] ;
;;

(**
Prints on standard output information collected and built throughout the parsing.
Might proove useful on debug purposes
@param string : generated attack trace to be printed
*)
let (print_debug: string -> unit) = function trace ->
	print_string "\nROLE PLAYER INSTANCES\n\n" ;
	List.iter (function ((agt,num),rol) -> print_string (agt^"\t"^(string_of_int num)^"\t"^rol^"\n") ) !role_inst_table ;
	print_string "\nIDENTIFIERS TYPES\n\n" ;
	List.iter (function (str,typ) -> print_string (str^"\t\t"^(Interface.atoms_type_to_string false typ)^"\n") ) !ids_type_table ;
	print_string "\nNONCES TABLE\n\n" ;
	List.iter (function ((str,id),rep) -> print_string (str^"\t"^(string_of_int id)^"\t"^rep^"\n") ) !nonce_table ;
	print_string "\nFREE VARIABLES TABLE\n\n" ;
	List.iter (function ((str,id),rep) -> print_string (str^"\t"^(string_of_int id)^"\t"^rep^"\n") ) !free_table ;
	print_string ("\nSIMULATION TRACE\n"^trace^"\n\n")
;;
