
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
open Types;;
open Globals;;


(*
  \section{Miscellaneous}
*)

(*
  \begin{center}
  type:
   [val find_player_in_role : int -> int]
  \end{center}
*)
let find_player_in_role role_id =
  fst
    (List.find
       (fun (_,b) -> b=role_id)
       !Globals.player_to_role)
;;

(*
   \begin{center}
   type: [val is_intruder : int -> bool]
   \end{center}
*)
let is_intruder i =
  (Globals.string_id#get_name i) = intruder_name
;;


(*
  \section{Operations on lists}
*)

(*
   \begin{center}
   type: [val insert_with_no_copy : 'a -> 'a list -> 'a list]
   \end{center}
*)
let insert_with_no_copy elt list =
  if not (List.mem elt list) then
    elt::list
  else
    list
;;

(*
   \begin{center}
   type: [val remove_duplicate : 'a list -> 'a list]
   \end{center}
*)
let rec remove_duplicate = function
    [] -> []
  | head::tail ->
      let purged_list =
 	List.find_all (fun x -> head <> x) tail
      in
      head::(remove_duplicate purged_list)
;;

(*
   \begin{center}
   type: [val concatenate_with_no_copy : 'a list -> 'a list -> 'a list]
   \end{center}
*)
let rec concatenate_with_no_copy list1 list2 =
  match list1 with
    [] -> list2
  | head::tail ->
      insert_with_no_copy head
	(concatenate_with_no_copy tail list2)
;;

(*
   \begin{center}
   type: [val partial_map : ('a -> 'b) -> 'a list -> 'b list]
   \end{center}
*)
let partial_map f list =
  List.flatten
    (List.map
       (fun x ->
	 try
	   [f x]
	 with
	   Match_failure _ -> [])
       list)
;;

(*
  \section{Translation from one type to another}
*)

(*
   \begin{center}
   type: [val variable_in_term_type_to_variable_global_id :
   int -> Types.atoms_type Types.term_structure -> int]
   \end{center}
*)
let variable_in_term_type_to_variable_global_id session_id variable =
  let variable_string_id =
    match variable with
      Base(Var(y)) -> y
    | _ -> raise Not_found
  in
  let variable_global_id =
    Globals.global_var_id#get_id_of session_id variable_string_id
  in
  memory_map#check_map_var variable_global_id;
  variable_global_id
;;


(*
   term type transformation utility
*)

(*
  \section{Functions to find the real value of a term}
*)

(*
   \begin{center}
   type: [val get_value : int -> Types.if_value_type]
   \end{center}
*)
let get_value mmap_id =
  let immediate_value = Globals.value_table#get_value mmap_id in
    match immediate_value with
	If_const maybe_pointer ->
	  begin
	    try
	      begin
		match (Globals.type_table#get_type maybe_pointer) with
(*		  Base(Pointer)*)
		  Set(_) ->
		    let pointer_global_var_id = Globals.global_var_id#get_id_of 0 maybe_pointer in
		    let pointer_mmap_id = Globals.memory_map#get_map pointer_global_var_id in
		    Globals.value_table#get_value pointer_mmap_id 
		| _ -> immediate_value
	      end
	    with
		_ -> immediate_value
	  end
      | _ -> immediate_value
;;


(*
   \begin{center}
   type: [val find_function_value :
   Types.if_value_type -> Types.if_value_type -> Types.if_value_type]
   \end{center}
*)
let rec find_function_value f arg =
  (*
     First, let us find the "real" value of arg if this is a set.
   *)
  let arg_real =
    List.map
      (function
	  If_function(g,l) -> find_function_value g l
	| a -> a)
      arg
  in
  let nb_args = List.length arg_real in
  (*
     We know that the value of [f] is a pointer to a set containing a couple whose first argument
     is [arg_real]. First, let us retrieve this set. We currently handle only the case where
     the function is a variable or a constant. In this last case, we consider the constant to be the pointer itself.
     whenever this processus fails, we raise an exception.
   *)
  let set_of_definition =
    match f with
      If_address set_addr ->
	get_value set_addr
    | If_const pointer ->
	(Globals.value_table#get_value 
	   (Globals.memory_map#get_map
	      (Globals.global_var_id#get_id_of 0 pointer)))
    | _ -> raise Not_found
  in
  match set_of_definition with
(*    If_lst l *)
    If_set l ->
      (*
	 We have found a set [l] that should contain the association between a point
	 and its value. Each element of [l] should be a pair (arg,value).
       *) 
(*      let association_list =
	(List.map
	   (fun element ->
	     match element with
*		 If_lst(l2)*
	       If_set(l2) ->
		 let rec find_val r_a = function
		     [If_function(g,l3)] -> (r_a, find_function_value g l3)
		   | [a] -> (r_a, a)
		   | If_function(g,l3)::la ->
		       find_val (r_a@[find_function_value g l3]) la
		   | a::la ->
		       find_val (r_a@[a]) la
		   | [] -> raise Not_found
		 in
		 find_val [] l2
	     | _ -> raise Not_found)
	   l)*)
      let association_list =
	(List.map
	   (fun element ->
	     let rec find_val r_a = function
	         0 ->
		   (function
		       If_function(g,l2) ->
			 (r_a, find_function_value g l2)
		     | t2 ->
			 (r_a, t2))
	       | n ->
		   (function
		       If_pair(If_function(g,l2),t2) ->
			 find_val (r_a@[find_function_value g l2]) (n-1) t2
		     | If_pair(t1,t2) ->
			 find_val (r_a@[t1]) (n-1) t2
		     | _ -> raise Not_found)
	     in
	     find_val [] nb_args element)
	   l)
      in
      List.assoc arg_real association_list
  | _ -> raise Not_found   
;;


(*
   \begin{center}
   type: [val get_if_value_type :
   int -> Types.atoms_type Types.term_structure -> Types.if_value_type]
   \end{center}
*)
let rec get_if_value_type session_id term =

  match term with 
    Pair(t1,t2) ->
      If_pair(get_if_value_type session_id t1, get_if_value_type  session_id t2)
  | Scrypt(t1,t2) ->
      If_scrypt(get_if_value_type session_id t1, get_if_value_type session_id t2)
  | Crypt(t1,t2) ->
      If_crypt(get_if_value_type session_id t1, get_if_value_type session_id t2)
  | Inv(t) ->
      If_inv(get_if_value_type session_id t)
  | Function(t1,l) ->
      begin
	(*
	   The result of the application depends on the type of [t1]. In a first pass,
	   we only support the case where [t1] is a variable.
	 *)
	match t1 with
	  Base(Var(i))
	| Base(Const(i))
	| Prime(Var(i)) -> 
	    begin
	      let var_type = (Globals.type_table#get_type i) in
	      match var_type with
		Function(_) ->
		  (*
		     In this case, we have a function application. We first translate
		     the two components into the [if_value_type] type, and then compute
		     the value of the function.
		   *)
		  let if_value_of_t1 = get_if_value_type session_id t1 in
		  let if_value_of_args = List.map (get_if_value_type session_id) l in
		  find_function_value if_value_of_t1 if_value_of_args
              | Base(Hash_func) -> 
		  (*
		     In this case, we have a hash function.
		   *)
		  If_function (get_if_value_type session_id t1, List.map (get_if_value_type session_id) l)
	      | _ ->
		  (*
		     In this last case, we have a function application by something with is not
		     a function. We abort.
		   *)
		  if (!debug_level >= 1) then
		    begin
		      prerr_endline ("Application of "^(Globals.string_id#get_name i)^"which is *not* a function.");
		      prerr_endline "Aborting.";
		    end;
		  raise Not_found
	    end
	| _ -> 
	    if (!debug_level >= 1) then
	      begin
		prerr_endline "Function application is currently only supported when the function is a variable or a constant";
		prerr_endline "Aborting.";
	      end;
	    raise Not_found
      end
  | Cons(t1,t2) ->
      If_cons(get_if_value_type session_id t1, get_if_value_type session_id t2)
  | Delete(t1,t2) ->
      If_delete(get_if_value_type session_id t1, get_if_value_type session_id t2)
  | Set(l) ->
      If_set(List.map (get_if_value_type session_id) l)
(*  | Lst(l) ->
      If_lst(List.map (get_if_value_type session_id) l) *)
  | Base(Var(i)) -> 
      Globals.value_table#get_value (memory_map#get_map (global_var_id#get_id_of session_id i))
  | Base(Const(i)) -> 
      If_const(i)
  | Prime(Var(i)) ->
      If_address (global_var_id#get_id_of session_id i)
  | Prime(_) -> (* add a semantic error in this following case ! *)
      prerr_endline "Utils.get_if_value_type: Warning: invalid primed constant"; 
      raise Not_found
  | Hash(_) ->
      prerr_endline "Utils.get_if_value_type: Impossible error: invalid hash type/term"; 
      raise Not_found
;;


(*
   \begin{center}
   type: [val term_to_type : term -> term_type]
   \end{center}
*)
let rec term_to_type = function
    Base(Var(i))
  | Base(Const(i))
  | Prime(Var(i)) ->
      Globals.type_table#get_type i
  | Prime(_) -> (* add a semantic error in this following case ! *)
      prerr_endline "Utils.term_to_type: Warning: invalid primed constant";
      raise Not_found
  | Cons(_,t2)
  | Delete(_,t2) ->
      term_to_type t2
  | Set(l) ->
      let rec find_applied = function
	  t::lt ->
	    (try
	      Set([term_to_type t])
	    with Not_found -> find_applied lt)
	| [] -> raise Not_found
      in
      find_applied l
(*  | Lst(l) ->
      let rec find_applied = function
	  t::lt ->
	    (try
	      Lst([term_to_type t])
	    with Not_found -> find_applied lt)
	| [] -> raise Not_found
      in
      find_applied l*)
  | Pair(t1,t2) ->
      Pair(term_to_type t1, term_to_type t2)
  | Scrypt(t1,t2) ->
      Scrypt(term_to_type t1, term_to_type t2)
  | Crypt(t1,t2) ->
      Crypt(term_to_type t1, term_to_type t2)
  | Inv(t) ->
      Inv(term_to_type t)
  | Function(t1,lt) ->
      Function(term_to_type t1, List.map term_to_type lt)
  | Hash(_) ->
      prerr_endline "Utils.term_to_type: Impossible error: invalid hash type/term"; 
      raise Not_found
;;

let rec if_term_to_type = function
    If_address(i)
  | If_const(i) ->
      Globals.type_table#get_type i
  | If_cons(_,t2)
  | If_delete(_,t2) ->
      if_term_to_type t2
  | If_set(l) ->
      let rec find_applied = function
	  t::lt ->
	    (try
	      Set([if_term_to_type t])
	    with Not_found -> find_applied lt)
	| [] -> raise Not_found
      in
      find_applied l
(*  | If_lst(l) ->
      let rec find_applied = function
	  t::lt ->
	    (try
	      Lst([if_term_to_type t])
	    with Not_found -> find_applied lt)
	| [] -> raise Not_found
      in
      find_applied l*)
  | If_pair(t1,t2) ->
      Pair(if_term_to_type t1, if_term_to_type t2)
  | If_scrypt(t1,t2) ->
      Scrypt(if_term_to_type t1, if_term_to_type t2)
  | If_crypt(t1,t2) ->
      Crypt(if_term_to_type t1, if_term_to_type t2)
  | If_inv(t) ->
      Inv(if_term_to_type t)
  | If_function(t1,lt) ->
      Function(if_term_to_type t1, List.map if_term_to_type lt)
;;



(*
  \section{Translation from one table to another}
*)

(*
   \begin{center}
   type: [val var_id_to_type : int -> Types.term_type]
   \end{center}
*)
let var_id_to_type id =
  type_table#get_type (global_var_id#get_name_of id)
;;


(*
   \begin{center}
   type: [val hlpsl_variable_to_global_var_id : int -> hlpsl_variable -> int]
   \end{center}
*)
let hlpsl_variable_to_global_var_id session_id hlpsl_var = 
  (Globals.global_var_id#get_id_of session_id hlpsl_var#get_id)
;;

(*
   \begin{center}
   type: [val to_if_type : 'a Types.term_structure -> 'a Types.term_structure]
   \end{center}
*
let to_if_type = function
*    Lst _ -> Lst []*
  | Set _ -> Set []
  | t -> t
;;
*)


(*
  type: [vars_consts_from_term: term -> atoms_type list]
*)
let rec vars_consts_from_term = function
    Pair(t1,t2)
  | Scrypt(t1,t2)
  | Crypt(t1,t2)
  | Cons(t1,t2) -> (vars_consts_from_term t1) @ (vars_consts_from_term t2)
  | Inv(t) -> vars_consts_from_term t
  | Function(t,lt) ->
      (vars_consts_from_term t)
      @ (List.flatten (List.map vars_consts_from_term lt))
  | Set(lt) -> List.flatten (List.map vars_consts_from_term lt)
  | Prime(t)
  | Base(t) -> [Base(t)]
  | _ -> []
;;

(*
  type: [vars_consts_from_IF_term: if_term -> atoms_type list]
*)
let rec vars_consts_from_IF_term = function
  | Iknows(t) -> vars_consts_from_term t
  | Ifnot(p) -> vars_consts_from_IF_term p
  | Ifcontains(t1,t2)
  | Ifequal(t1,t2)
  | Ifleq(t1,t2) -> (vars_consts_from_term t1) @ (vars_consts_from_term t2)
  | State(_,_,lt)
  | Unknown(_,lt) -> List.flatten (List.map vars_consts_from_term lt)
  | _ -> []
;;

(*
  type: [vars_consts_from_LTL_goal: ltl_type -> atoms_type list]
*)
let vars_consts_from_LTL_goal g =
  let rec vc_from_ltl = function
      Pred(p) -> vars_consts_from_IF_term p
    | Op1(f)
    | Op2(f)
    | Op3(f)
    | Op4(f)
    | Op5(f) -> vc_from_ltl f
    | Conj(f1,f2)
    | Dij(f1,f2)
    | Impl(f1,f2) ->
	(vc_from_ltl f1) @ (vc_from_ltl f2)
  in
  remove_duplicate (vc_from_ltl g)
;;
