
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2009 Yann Glouche, Thomas Genet, Olivier Heen, 
   	     	 	   Erwan Houssay and Ronan Saillard.

   The CAS+ Parser includes source code from the CASRUL tool
   CASRUL: Copyright (C) 2001 Florent Jacquemard, Michael Rusinowitch and
      		    	      Laurent Vigneron.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.6. Last modification 16/07/09
*)
(**
interface.ml
@author Yann GLOUCHE & Thomas GENET & Erwan Houssay & Ronan Saillard
@version 1.6
@since 16/07/2009
*)

open Tk;;
open Filename;;

let span_version= "SPAN 1.6 - Protocol Verification"
(** Initialize Tk *)
let root = openTk ();;

(** Colors *)

let defcol = (`Color "#dfdfdf" : [> `Color of string ]);;
let selcol = (`Color "#ffffff" : [> `Color of string ]);;
let clicol = (`Color "#ffafaf" : [> `Color of string ]);;
let blue = (`Color "#9999ff" : [> `Color of string ]);;

type bufferType = Hlpsl | Cas | NoSave;;

(** Verification Tools *)


(* For using files with pathes containing spaces we need an escape character *)

let os=Sys.os_type;;
let esc= if os="Unix" then "" else "\"";;

(* Command for redirecting standard and error channels towards f *)
let redirect_towards f= 
  if os="Unix" 
  then (" >& "^f^" ")  
    (* The redirection (" >& "^f^" ") does not work for windows*)
  else 
    (* The redirection (" 1> "^f^" 2>&1 ") does not work for csh *)
    (" 1> "^f^" 2>&1 ");;



(* If avispa tools are already installed then we use path to those tools and AVISPA_PACKAGE variable 
   otherwise we use tools included in span package *)


let path_span = 
  (try 
    Sys.getenv "SPAN"
  with Not_found -> failwith "Environment Variable SPAN should be declared!" )


let path_avispa =
  (try
    Sys.getenv "AVISPA_PACKAGE"
  with
    Not_found -> path_span)


(* backend names are different depending on the execution platform, Unix/Windows *)

let hlpsl2if,ofmc,satmc,atse,ta4sp,hlpsl2scenario,report2trace,alicebob2hlpsl=
if os="Unix" then
   Filename.quote (path_avispa^"/bin/translator/hlpsl2if" : string),
   Filename.quote (path_avispa^"/bin/backends/ofmc/ofmc" : string),
   Filename.quote (path_avispa^"/bin/backends/satmc.bash" : string),
   Filename.quote (path_avispa^"/bin/backends/cl/cl-atse" : string),
   Filename.quote (path_avispa^"/bin/backends/ta4sp.bash" : string),
   Filename.quote (path_span^"/bin/protocol_simulation/hlpsl2scenario" : string),
   Filename.quote (path_span^"/bin/trace_generation/report2trace" : string),
   Filename.quote (path_span^"/bin/hlpsl_generation/alicebob2hlpsl" : string)
else
   Filename.quote (path_avispa^"\\bin\\translator\\hlpsl2if.exe" : string),
   Filename.quote (path_avispa^"\\bin\\backends\\ofmc\\ofmc.exe" : string),
   Filename.quote (path_avispa^"\\bin\\backends\\satmc.bash" : string),
   Filename.quote (path_avispa^"\\bin\\backends\\cl\\clatse.exe" : string),
   Filename.quote (path_avispa^"\\bin\\backends\\TA4SP\\TA4SP.exe" : string),
   Filename.quote (path_span^"\\bin\\protocol_simulation\\hlpsl2scenario.exe" : string),
   Filename.quote (path_span^"\\bin\\trace_generation\\report2trace.exe" : string),
   Filename.quote (path_span^"\\bin\\hlpsl_generation\\alicebob2hlpsl.exe" : string);;

(* Cannot use directly hlpsl2if and hlpsl2scenario because filenames are
   quoted *)

if os="Unix" then
  (if not ((Sys.file_exists (path_avispa^"/bin/translator/hlpsl2if")) &
	  (Sys.file_exists (path_span^"/bin/protocol_simulation/hlpsl2scenario")))
   then failwith "Check your environment variables ($AVISPA_PACKAGE and $SPAN): one of hlpsl2if or hlpsl2scenario is not reachable")
else
  (if not ((Sys.file_exists (path_avispa^"\\bin\\translator\\hlpsl2if.exe")) &
	   (Sys.file_exists (path_span^"\\bin\\protocol_simulation\\hlpsl2scenario.exe"))) 
   then failwith "Check your environment variables ($AVISPA_PACKAGE and $SPAN): one of hlpsl2if or hlpsl2scenario is not reachable")



(** repertory for the tempory files generated by the tools *)
let temp_dir = 
  if os="Unix" then (path_avispa^"/testsuite/results/")
  else (path_avispa^"\\testsuite\\results\\")


(** Read a file and return a list of string where each string is a line of the file 
@return string list
*)
let readFile (fileName : string) : string =
   let channel = open_in fileName in
   let res = ref "" in
   (try
   begin
    while true do
     let line = input_line channel
     in res:=(!res) ^ "\n" ^ line ; 
    done;
   end;
   with
   End_of_file ->
   begin
   close_in channel;
   end);
   !res;;

(** Read a file and return a list of string where each string is a line of the file
@param fileName is the name of the file to read
@return string list
*)
let readFile_to_list (fileName : string) : string list =
   let channel = open_in fileName in
   let res = ref [] in
   try
   begin
    while true do
     let line = input_line channel
     in res:=line::(!res); 
    done;
   end;
   failwith "problem"
   with
   End_of_file ->
   begin
   close_in channel;
   List.rev (!res);
   end;;

 (** Read a file and return the number of lines of the file with a list of string where each string is a line of the file 
@return string list
*)
let readFile_and_give_line_number (fileName : string) : string * int =
   let channel = open_in fileName in
   let res = ref "" in
   let line_number = ref 0 in
   try
   begin
    while true do
     let line = input_line channel
     in res:=(!res) ^ "\n" ^ line ; 
     line_number := (!line_number)+1;
    done;
   end;
   failwith "problem"
   with
   End_of_file ->
   begin
   close_in channel;
   ((!res),(!line_number));
   end;;

(** Save a file 
@param txt is the text to save
@return unit
*)
let saveFile (fileName : string) (txt : string) : unit =
   let channel = open_out fileName in
   output_string channel txt;
   close_out channel;;

let get_output_if (fileName : string) : string =
 if(Filename.check_suffix fileName ".hlpsl") 
 then (temp_dir^(Filename.basename ((Filename.chop_suffix fileName ".hlpsl")^".if")))
 else (temp_dir^(Filename.basename (fileName^".if")));;


(** Remove temporary files created for execution
@param files is a list of names of the temporary files to be removed
*)
let removeFiles (files : string list) : unit =
   List.iter (fun f -> if(Sys.file_exists f) then ignore(Sys.remove f)) files;;

(** Text font *)

(** the current fonts size *)
let current_font_size = ref(
	let file_config_command = (path_span^"/font_size.ini") in
	let string_to_int s = (try 
    				int_of_string s
  				with _ -> -1 ) in
	if(Sys.file_exists file_config_command)
				then(	let rec aux l =
						match l with [] -> 10
							|(x::ls) -> 	if((string_to_int x) > (-1))
									then (string_to_int x)
									else aux ls;
					in aux (readFile_to_list file_config_command)
				)
				else 10
);;

let list_font = (ref [] : Tk.font list Pervasives.ref);;

let get_font (tag : string) (choices : string list) : unit =
 let all = Font.families () in
 let rec find ch =
   match ch with
   | [] -> ""
   | f::tail ->
     ( try (List.find (( = ) f) all)
       with Not_found -> find tail ) in
 match find choices with
 | "" -> list_font:=(Font.create ~name:tag ~size:(!current_font_size) ())::(!list_font)
 | fam -> list_font:=(Font.create ~name:tag ~family:fam ~size:(!current_font_size) ())::(!list_font);;

(** for change fonts size *)
let set_size_of_all_fonts (size : int) : unit =
   List.iter (fun font -> Font.configure font ~size:size) (!list_font);;

let configure () : unit =
 get_font "guifont" ["verdana"; "lucida"; "bitstream vera sans"];
 get_font "textfont"     ["lettergothic"; "lucida typewriter"; "bitstream vera sans mono"; "courier"];
 Option.add ~path:"*Menu.font" "guifont";
 Option.add ~path:"*Label.font" "guifont";
 Option.add ~path:"*Message.font" "guifont";
 Option.add ~path:"*Button.font" "guifont";
 Option.add ~path:"*Radiobutton.font" "guifont";
 Option.add ~path:"*Checkbutton.font" "guifont";
 Option.add ~path:"*Text.font" "textfont";
 Option.add ~path:"*Canvas.font" "textfont";
 Option.add ~path:"*Listbox.font" "textfont";
 Option.add ~path:"*Entry.font" "textfont";
 Option.add ~path:"*Text.borderWidth" "1";
 Option.add ~path:"*Canvas.borderWidth" "1";
 Option.add ~path:"*Button.borderWidth" "1";
 Option.add ~path:"*Scrollbar.borderWidth" "1";
 Option.add ~path:"*Entry.borderWidth" "1";
 Option.add ~path:"*Listbox.borderWidth" "1";
 Option.add ~path:"*Menu.borderWidth" "1";
 Option.add ~path:"*Menu.activeBorderWidth" "1";
 Option.add ~path:"*Menu.tearOff" "false";
 Option.add ~path:"*Listbox.selectBorderWidth" "0";
 Option.add ~path:"*Listbox.selectBackground" "#0000a0";
 Option.add ~path:"*Listbox.selectForeground" "#ffffff";
 Option.add ~path:"*Listbox.background" "#ffffff";
;;

let words = ([
 "anaerobic"; "beneficent"; "corundum"; "diesel";
 "erotic"; "fragment"; "garrulous"; "hexagon";
 "irate"; "jinx"; "kerosene"; "lethargic";
 "malamute"; "normality"; "oblate"; "parsimony";
 "quail"; "rigorous"; "scimitar"; "tuxedo";
 "ululation"; "verify"; "warbler"; "xenon";
 "yawl"; "zither"
] : string list);;

let nop (() : unit) : unit = ();;

(*********************)

class hlpsl2ifOption optionFrame =
object (self)
 val optionLabel : Widget.label Widget.widget = Label.create ~text:"Options" ~width:20 ~height:2 ~background:blue optionFrame
 val types : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Types section  " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable typesString : string = ""
 val inits :  Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Inits section    " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable initString : string = ""
 val rules : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Rules section  " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable rulesString : string = ""
 val goals : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Goals section  " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable goalsString : string = ""
 (** Initialize the objects of this class *)
 method init (attackButton : Widget.button Widget.widget) : unit = (
     Checkbutton.configure types ~command:(fun () -> (if typesString="" then typesString <- " --types " else typesString <- ""));
     Checkbutton.configure inits ~command:(fun () -> (if initString="" then initString <- " --init " else initString <- ""));
     Checkbutton.configure rules ~command:(fun () -> (if rulesString="" then rulesString <- " --rules " else rulesString <- ""));
     Checkbutton.configure goals ~command:(fun () -> (if goalsString="" then goalsString <- " --goals " else goalsString <- ""));
     pack ~side:`Top [optionFrame];
     pack ~side:`Top [optionLabel];
     pack ~side:`Top [types];
     pack ~side:`Top [inits];
     pack ~side:`Top [rules];
     pack ~side:`Top [goals];
		 Button.configure attackButton ~state:`Disabled;)
 (** Destruct the objects of this class, and the graphical objects builds by the objects of this class *)
 method destroy : unit = ( destroy optionLabel; 
		    destroy types; 
		    destroy inits; 
		    destroy rules; 
		    destroy goals;)
 (** Prepare and carry out ystem calls for  execution *)
 method call (fileName : string) : string*string*string*string = (
   let out_name = get_output_if fileName in
   let error_if_file = (out_name^".error") in
	 removeFiles [out_name ; error_if_file];
(* Under windows, don't know why but the Sys.command on hlpsl2if may return a Sys_error when 
   the translation fails on an error... an thus prevent the interface from displaying the 
   hlpsl2if error message *)
   
   (try
   (ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^typesString^initString^rulesString^goalsString^" >"^(Filename.quote out_name)^" 2>"^(Filename.quote error_if_file)^esc)))
    with Sys_error _ -> ());

   (out_name, "", error_if_file, "")
 )
 (** Execute HLPSL2IF with the parameters of the user *)
 method execution (fileName : string) (outText : Widget.text Widget.widget) : unit = (
   let (out_name, _, error_if_file, _) = self#call(fileName) in
   let txt = if(Sys.file_exists out_name) then (
   		let txt = readFile out_name in
		(if(Sys.file_exists error_if_file) 
		then let (out_error, num_of_lines) = readFile_and_give_line_number error_if_file in
			if(num_of_lines >2)
			then ("\n----- Output error of if translator :\n\n"^out_error^(if(txt<>"") then ("\n--------------- \n"^txt) else ""))
			else txt;
		else txt)
		)
	      else (if(Sys.file_exists error_if_file) 
		    then ("\n----- Output error of if translator :\n\n"^(readFile error_if_file))
		    else "")
   in
   Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ;
   Text.insert ~index:(`End,[]) ~text:txt outText;
	 removeFiles [out_name ; error_if_file];
   )
end;;



class ofmcOption optionFrame =
object (self)
 val optionLabel : Widget.label Widget.widget = Label.create ~text:"Options" ~width:20 ~height:2 ~background:blue optionFrame
 val session : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Session Compilation" ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable sessionString : string = ""
 val f : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable hf1 : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable labelDepth : Widget.label Widget.widget = Label.create ~text:"Defth : " ~width:20 ~height:2 optionFrame
 val mutable entryDepth : Widget.entry Widget.widget = Entry.create ~width:20 ~borderwidth:1 optionFrame
 val mutable hf2 : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable labelPath : Widget.label Widget.widget = Label.create ~text:"Path : " ~width:20 ~height:2 optionFrame
 val mutable entryPath : Widget.entry Widget.widget = Entry.create ~width:20 ~borderwidth:1 optionFrame
  (** Initialize the objects of this class *)
 method init (attackButton : Widget.button Widget.widget) : unit = (
 	destroy hf1;
 	hf1 <- Frame.create f;
	 destroy labelDepth;
	 labelDepth <- Label.create ~text:"Defth : " ~width:20 ~height:2 hf1;
	 destroy entryDepth;
	 entryDepth <- Entry.create ~width:20 ~background:selcol hf1;
	 destroy hf2;
	 hf2 <- Frame.create f;
	 destroy labelPath;
	 labelPath <- Label.create ~text:"Path : " ~width:20 ~height:2 hf2;
	 destroy entryPath;
	entryPath <- Entry.create ~width:20 ~background:selcol hf2;
	Checkbutton.configure session ~command:(fun () -> (if sessionString="" then sessionString <- " -sessco " else sessionString <- ""));  
	pack ~side:`Top [optionFrame];
	pack ~side:`Top [optionLabel];
	pack ~side:`Top [session];
	pack ~side:`Top [f];
	pack ~side:`Top [hf1];
	pack ~side:`Left [labelDepth];
	pack ~side:`Left [entryDepth];
	pack ~side:`Top [hf2];
	pack ~side:`Left [labelPath];
  pack ~side:`Left [entryPath];
	Button.configure attackButton ~state:`Normal;)
 (** Destruct the objects of this class, and the graphical objects builds by the objects of this class *)
 method destroy : unit = ( destroy optionLabel; destroy session; 
		    destroy f; destroy hf1; destroy labelDepth; 
		    destroy entryDepth; destroy hf2;
		    destroy labelPath; destroy entryPath;)
 (** Prepare and carry out ystem calls for  execution *)
 method call(fileName : string) : string*string*string*string = (
   let out_name_if = get_output_if fileName in
   let out_name_ofmc = (temp_dir^(Filename.basename fileName)^".ofmc") in
   let error_if_file = (out_name_if^".error") in
   let error_ofmc_file = (out_name_ofmc^".error") in
	 removeFiles [out_name_if ; out_name_ofmc ; error_if_file ; error_ofmc_file];
   
   let entryDepthString = if (Entry.get entryDepth)="" then "" else (" -d "^(Entry.get entryDepth)) in
   let entryPathString = if (Entry.get entryPath)="" then "" else (" -p "^(Entry.get entryPath)) in

   (* Under windows, don't know why but the Sys.command on hlpsl2if may return a Sys_error when 
   the translation fails on an error... an thus prevent the interface from displaying the 
   hlpsl2if error message *)
   
   (try
      (ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^esc)))
    with Sys_error _ -> ());
     ignore(Sys.command(esc^ofmc^" "^(Filename.quote out_name_if)^" "^sessionString^entryDepthString^entryPathString^" > "^(Filename.quote out_name_ofmc)^" 2>"^(Filename.quote error_ofmc_file)^esc));
     (out_name_if, out_name_ofmc, error_if_file, error_ofmc_file)
 )
 (** Execute OFMC with the parameters of the user *)
 method execution (fileName : string) (outText : Widget.text Widget.widget) : unit = (
   let (out_name_if, out_name_ofmc, error_if_file, error_ofmc_file) = self#call(fileName) in
   Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
   if(Sys.file_exists out_name_ofmc) then  (
   	let txt = readFile out_name_ofmc in
	if(txt<>"")
	then(
   		let (out_error, num_of_lines) = readFile_and_give_line_number error_if_file in
	 	let warning= (if(num_of_lines >2)  then ("\n----- Output error of if translator :\n\n"^out_error^"\n--------------- \n") else "") in
	Text.insert ~index:(`End,[]) ~text:(warning^txt) outText
	)
	else(
   		if(Sys.file_exists error_if_file) 
		then Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of if translator :\n\n")^(readFile error_if_file)) outText;
		if(Sys.file_exists error_ofmc_file) 
		then Text.insert ~index:(`End,[]) ~text:(("\n\n\n----- Output error of OFMC :\n\n")^(readFile error_ofmc_file)) outText;
   		);
	);
   removeFiles [out_name_if ; out_name_ofmc ; error_if_file ; error_ofmc_file];
   )
end;;


class atseOption optionFrame =
object (self)
 val optionLabel : Widget.label Widget.widget = Label.create ~text:"Options" ~width:20 ~height:2 ~background:blue optionFrame
 val simplify : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Simplify            " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable simplifyString : string = ""
 val untyped : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Untyped model" ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable untypedString : string = ""
 val verbose : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Verbose mode  " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable verboseString : string = ""
 val search : Widget.label Widget.widget = Label.create ~text:"Search Algorithm" ~width:20 ~height:2 optionFrame
 val lst : Widget.listbox Widget.widget = Listbox.create ~height:2 ~background:selcol ~selectmode:`Single optionFrame
 (** Initialize the objects of this class *)
  method init (attackButton : Widget.button Widget.widget) : unit = (
     Checkbutton.configure simplify ~command:(fun () -> (if simplifyString="" then simplifyString <- " -light " else simplifyString <- ""));
     Checkbutton.configure untyped ~command:(fun () -> (if untypedString="" then untypedString <- " -notype " else untypedString <- ""));
     Checkbutton.configure verbose ~command:(fun () -> (if verboseString="" then (verboseString <- " -vv "; Button.configure attackButton ~state:`Disabled;) else (verboseString <- ""; Button.configure attackButton ~state:`Normal;)));
     Listbox.insert lst ~index:`End ~texts:["Depth first"; "Breadth first"];
     Listbox.selection_set lst ~first:(`Num 0) ~last:(`Num 0);
     pack ~side:`Top [optionFrame];
     pack ~side:`Top [optionLabel];
     pack ~side:`Top [simplify];
     pack ~side:`Top [untyped];
     pack ~side:`Top [verbose];
     pack ~side:`Top [search];
     pack ~side:`Top [lst];
		 Button.configure attackButton ~state:`Normal;)
 (** Destruct the objects of this class, and the graphical objects builds by the objects of this class *)
 method destroy : unit = ( destroy optionLabel; destroy simplify; 
		    destroy untyped; destroy verbose; 
		    destroy search; destroy lst; )
 (** Prepare and carry out ystem calls for  execution *)
 method call(fileName : string) : string*string*string*string = (
   let out_name_if = get_output_if fileName in
   let out_name_atse = (temp_dir^(Filename.basename fileName)^".atse") in
   let error_if_file = (out_name_if^".error") in
   let error_atse_file = (out_name_atse^".error") in
	 removeFiles [out_name_if ; out_name_atse ; error_if_file ; error_atse_file];
   
   let select = match (Listbox.curselection lst)
                with [(`Num num)] -> num
                 | _ -> -1 in
   let algo = match select
	      with 0 -> " -td "
               |    1 -> " -lr "
               |    _ -> "" in 
	       
     (* Under windows, don't know why but the Sys.command on hlpsl2if and atse may
	return a Sys_error when the translation fails on an error... an thus prevent
	the interface from displaying the hlpsl2if error message *)
     
     (try
	(ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^esc)))
      with Sys_error _ -> ());
     (try
	(ignore(Sys.command(esc^atse^" "^(Filename.quote out_name_if)^" "^simplifyString^untypedString^verboseString^algo^" > "^(Filename.quote out_name_atse)^" 2>"^(Filename.quote error_atse_file)^esc)))
      with Sys_error _ -> ());
     (out_name_if, out_name_atse, error_if_file, error_atse_file))
								  (** Execute ATSE with the parameters of the user *)
 method execution (fileName : string) (outText : Widget.text Widget.widget) : unit = (
   let (out_name_if, out_name_atse, error_if_file, error_atse_file) = self#call(fileName) in
   Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
   if(Sys.file_exists out_name_atse) then  (
   	let txt = readFile out_name_atse in
   	if(Sys.file_exists error_if_file) 
	then (let (_,line_number)= (readFile_and_give_line_number error_if_file) in
		if(line_number > 2) 
		then (Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of if translator :\n\n")^(readFile error_if_file)) outText;
			if(Sys.file_exists error_atse_file) 
			then Text.insert ~index:(`End,[]) ~text:(("\n\n\n----- Output error of ATSE :\n\n")^(readFile error_atse_file)) outText;
			if(Sys.file_exists out_name_atse) 
			then Text.insert ~index:(`End,[]) ~text:(("\n\n")^txt) outText;
			)
		else 	(if(Sys.file_exists out_name_atse) 
			then Text.insert ~index:(`End,[]) ~text:txt outText;)
	)
	else (if(Sys.file_exists out_name_atse) 
	    then Text.insert ~index:(`End,[]) ~text:txt outText;)
   );
   removeFiles [out_name_if ; out_name_atse ; error_if_file ; error_atse_file];
   )
end;;


class satmcOption optionFrame =
object (self)
 val optionLabel : Widget.label Widget.widget = Label.create ~text:"Options" ~width:20 ~height:2 ~background:blue optionFrame
 val frame2 : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable solverLabel : Widget.label Widget.widget = Label.create ~text:"Solver : " ~height:2 optionFrame
 val mutable lst : Widget.listbox Widget.widget = Listbox.create ~height:2 ~background:selcol ~borderwidth:1 optionFrame
 val frame3 : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable depthLabel : Widget.label Widget.widget = Label.create ~text:"Depth : " ~height:2 optionFrame
 val mutable depthEntry : Widget.entry Widget.widget = Entry.create ~width:20 ~borderwidth:1 optionFrame
 val abstraction : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Abstraction/Refinement   " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable abstractionString : string = " --mutex=0  "
 val compound : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Compound Types                 " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable compoundString : string = " --typed_model=strongly "
 val optimized : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Optimized intruder            " ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable optimizedString : string = " --oi=false "
 (** Initialize the objects of this class *)
  method init (attackButton : Widget.button Widget.widget) : unit = (
     destroy solverLabel;
     solverLabel <- Label.create ~text:"Solver : "  ~height:2 frame2;
     destroy lst;
     lst <- Listbox.create ~height:2 ~background:selcol frame2;
     destroy depthLabel;
     depthLabel <- Label.create ~text:"Depth : "  ~height:2 frame3;
     destroy depthEntry;
     depthEntry <- Entry.create ~width:20 ~background:selcol frame3;
     Listbox.insert lst ~index:`End ~texts:["Chaff"; "SIM" ; "SATO"];
     Listbox.selection_set lst ~first:(`Num 0) ~last:(`Num 0);
     Checkbutton.configure abstraction ~command:(fun () -> (if abstractionString=" --mutex=0  " then abstractionString <- " --mutex=1 " else abstractionString <- " --mutex=0  "));
     Checkbutton.configure compound ~command:(fun () -> (if compoundString=" --typed_model=strongly " then compoundString <- " --typed_model=yes " else compoundString <- " --typed_model=strongly "));
     Checkbutton.configure optimized ~command:(fun () -> (if optimizedString=" --oi=false " then optimizedString <- " --oi=true " else optimizedString <- " --oi=false "));
     pack ~side:`Top [optionFrame];
     pack ~side:`Top [optionLabel];
     pack ~side:`Top [frame2];
     pack ~side:`Left [solverLabel];
     pack ~side:`Right [lst];
     pack ~side:`Top [frame3];
     pack ~side:`Left [depthLabel];
     pack ~side:`Right [depthEntry];
     pack ~side:`Top [abstraction];
     pack ~side:`Top [compound];
     pack ~side:`Top [optimized];
		 Button.configure attackButton ~state:`Disabled;)
(** Destruct the objects of this class, and the graphical objects builds by the objects of this class *)
 method destroy : unit = ( destroy optionLabel; destroy frame2; 
		    destroy solverLabel; destroy lst; 
		    destroy frame3; destroy depthLabel;
		    destroy depthEntry; destroy abstraction;
		    destroy compound; destroy optimized;)
 (** Prepare and carry out ystem calls for  execution *)
 method call(fileName : string) : string*string*string*string = (
   let out_name_if = get_output_if fileName in
   let out_name_satmc = (temp_dir^(Filename.basename fileName)^".satmc") in
   let error_if_file = (out_name_if^".error") in
   let error_satmc_file = (out_name_satmc^".error") in
	 removeFiles [out_name_if ; out_name_satmc ; error_if_file ; error_satmc_file];
   
   let select = match (Listbox.curselection lst)
                with [(`Num num)] -> num
                 | _ -> -1 in
   let algo = match select
	      with 0 -> " --solver=chaff "
               |    1 -> " --solver=sim "
               |    2 -> " --solver=sato "
               |    _ -> "" in 
   let depthString = if (Entry.get depthEntry)="" then "" else (" --max="^(Entry.get depthEntry)) in
   
   (* Under windows, don't know why but the Sys.command on hlpsl2if may return a Sys_error when 
   the translation fails on an error... an thus prevent the interface from displaying the 
   hlpsl2if error message *)
   
   (try
      (ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^esc)))
    with Sys_error _ -> ());
   ignore(Sys.command(esc^satmc^" "^(Filename.quote out_name_if)^" "^compoundString^" --output="^(Filename.quote (Filename.dirname fileName))^abstractionString^optimizedString^depthString^algo^" > "^(Filename.quote out_name_satmc)^" 2>"^(Filename.quote error_satmc_file)^esc));
  (out_name_if, out_name_satmc, error_if_file, error_satmc_file)
 )
 (** Execute SATMC with the parameters of the user *)
 method execution (fileName : string) (outText : Widget.text Widget.widget) : unit = (
   let (out_name_if, out_name_satmc, error_if_file, error_satmc_file) = self#call(fileName) in
   Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
   if(Sys.file_exists out_name_satmc) then  (
   	let txt = readFile out_name_satmc in
   	if(Sys.file_exists error_if_file) 
	then (let (_,line_number)= (readFile_and_give_line_number error_if_file) in
		if(line_number > 2) 
		then (Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of if translator :\n\n")^(readFile error_if_file)) outText;
			if(Sys.file_exists error_satmc_file) 
			then Text.insert ~index:(`End,[]) ~text:(("\n\n\n----- Output error of SATMC :\n\n")^(readFile error_satmc_file)) outText;
			if(Sys.file_exists out_name_satmc) 
			then Text.insert ~index:(`End,[]) ~text:(("\n\n")^txt) outText;
			)
		else 	(if(Sys.file_exists out_name_satmc) 
			then Text.insert ~index:(`End,[]) ~text:txt outText;)
	)
	else (if(Sys.file_exists out_name_satmc) 
	    then Text.insert ~index:(`End,[]) ~text:txt outText;)
   );
	 removeFiles [out_name_if ; out_name_satmc ; error_if_file ; error_satmc_file];
   )
end;;


class ta4spOption optionFrame =
object (self)
 val optionLabel : Widget.label Widget.widget = Label.create ~text:"Options" ~width:20 ~height:2 ~background:blue optionFrame
 val verificationLabel : Widget.label Widget.widget = Label.create ~text:"Verification Model" ~width:20 ~height:2 optionFrame
 val agent : Widget.checkbutton Widget.widget = Checkbutton.create ~text:"Two Agent Only" ~width:20 ~height:2 ~borderwidth:1 optionFrame
 val mutable agentString : string = ""
 val knowledgeLabel : Widget.label Widget.widget = Label.create ~text:"Intruder Knowledge" ~width:20 ~height:2 optionFrame
 val tv  : Textvariable.textVariable= Textvariable.create ()
 val mutable rads : Widget.radiobutton Widget.widget list = [] 
 val frame2 : Widget.frame Widget.widget = Frame.create optionFrame
 val mutable levelLabel : Widget.label Widget.widget = Label.create ~text:"Level : " ~width:20 ~height:2 optionFrame
 val mutable levelEntry : Widget.entry Widget.widget = Entry.create ~width:20 ~borderwidth:1 ~background:defcol ~state:`Disabled optionFrame
 (** Initialize the objects of this class *)
 method init (attackButton : Widget.button Widget.widget) : unit = (
        Textvariable.set tv "1";
     rads <- List.map
        (fun (t, v) -> Radiobutton.create ~text:t ~value:v ~variable:tv ~borderwidth:1 optionFrame)
        [("Over-Approximation  ", "1"); ("Under-Approximation","2")];
     match rads with []          -> ();
                     |[x]       -> ();
                     |(x::y::_) -> (Radiobutton.select x; 
                                   Radiobutton.configure x ~command:(fun () -> Entry.configure levelEntry  ~background:defcol ~state:`Disabled);
                                   Radiobutton.configure y ~command:(fun () -> Entry.configure levelEntry  ~background:selcol ~state:`Normal););

                                 
     destroy levelLabel;
     levelLabel <- Label.create ~text:"Level : " ~width:20 ~height:2 frame2;
     destroy levelEntry;
     levelEntry <- Entry.create ~width:20 ~background:defcol ~state:`Disabled frame2;
     Checkbutton.configure agent ~command:(fun () -> (if agentString="" then agentString <- " --2AgentsOnly " else agentString <- ""));
     pack ~side:`Top [optionFrame];
     pack ~side:`Top [optionLabel];
     pack ~side:`Top [verificationLabel];
     pack ~side:`Top [agent];
     pack ~side:`Top [knowledgeLabel];
     pack rads;
     pack ~side:`Top [frame2];
     pack ~side:`Left [levelLabel];
     pack ~side:`Right [levelEntry];
		 Button.configure attackButton ~state:`Disabled;)
 (** Destruct the objects of this class, and the graphical objects builds by the objects of this class *)
 method destroy : unit = ( destroy optionLabel; destroy verificationLabel; 
                    List.iter (fun t -> destroy t) rads;
		    destroy agent; destroy knowledgeLabel; 
		    destroy frame2; destroy levelLabel;
		    destroy levelEntry;)
 (** Prepare and carry out ystem calls for  execution *)
 method call(fileName : string) : string*string*string*string = (
   let out_name_if = get_output_if fileName in
   let out_name_ta4sp = (temp_dir^(Filename.basename fileName)^".ta4sp") in
   let error_if_file = (out_name_if^".error") in
   let error_ta4sp_file = (out_name_ta4sp^".error") in
	 removeFiles [out_name_if ; out_name_ta4sp ; error_if_file ; error_ta4sp_file];
   
   if os="Unix"
	 then(   
	  let select = match (Textvariable.get tv)
                with "1" -> ""
                 | "2" -> (if (Entry.get levelEntry)="" then " --level=1 " else (" --level="^(Entry.get levelEntry)))
                 | _ -> "" in
		 
   ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^esc));
    ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^" ; "^ta4sp^" "^(Filename.quote out_name_if)^" "^agentString^select^" > "^(Filename.quote out_name_ta4sp)^" 2>"^(Filename.quote error_ta4sp_file)^esc));
   )
   else(   (* Windows case *)
	  let select = match (Textvariable.get tv)
                with "1" -> ""
                 | "2" -> (if (Entry.get levelEntry)="" then " --level 1 " else (" --level "^(Entry.get levelEntry)))
                 | _ -> "" in
   (* Under windows, don't know why but the Sys.command on hlpsl2if may return a Sys_error when 
   the translation fails on an error... an thus prevent the interface from displaying the 
   hlpsl2if error message *)
   
   (try
      (ignore(Sys.command(esc^hlpsl2if^" --stdout "^(Filename.quote fileName)^" >"^(Filename.quote out_name_if)^" 2>"^(Filename.quote error_if_file)^esc)))
    with Sys_error _ -> ());
   
    let path_tmp = path_avispa^"\\bin\\backends\\TA4SP\\" in
    ignore(Sys.command(esc^ta4sp^" --output "^(Filename.quote path_tmp)^" "^agentString^" "^select^" "^(Filename.quote out_name_if)^" > "^(Filename.quote out_name_ta4sp)^" 2> "^(Filename.quote error_ta4sp_file)^esc));
   );
	 (out_name_if, out_name_ta4sp, error_if_file, error_ta4sp_file)
	)
 (** Execute TA4SP with the parameters of the user *)
 method execution (fileName : string) (outText : Widget.text Widget.widget) : unit = (
  let (out_name_if, out_name_ta4sp, error_if_file, error_ta4sp_file) = self#call(fileName) in
	Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
   if(Sys.file_exists out_name_ta4sp) then  (
   	let txt = readFile out_name_ta4sp in
	if(txt<>"")
	then(
   		let (out_error, num_of_lines) = readFile_and_give_line_number error_if_file in
	 	let warning= (if(num_of_lines >2)  then ("\n----- Output error of if translator :\n\n"^out_error^"\n--------------- \n") else "") in
	Text.insert ~index:(`End,[]) ~text:(warning^txt) outText
	)
	else(
   		if(Sys.file_exists error_if_file) 
		then Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of if translator :\n\n")^(readFile error_if_file)) outText;
		if(Sys.file_exists error_ta4sp_file) 
		then Text.insert ~index:(`End,[]) ~text:(("\n\n\n----- Output error of TA4SP :\n\n")^(readFile error_ta4sp_file)) outText;
   		);
	);
   removeFiles [out_name_if ; out_name_ta4sp ; error_if_file ; error_ta4sp_file];
 )
end;;

(*fichiers ouverts*)
let fileName = (ref "" : string Pervasives.ref);;
let fileNameCas = (ref "" : string Pervasives.ref);;
let synch = ref true;;
let casFileLastModification = ref 0.0;;
(** Attack trace generation debug mode flag *)
let atk_dbg = (ref false : bool Pervasives.ref);;

(* Parsing the arguments of the compiler, for detecting options and
   possible mistakes.
*)

let is_input = ref false in
(Arg.parse
   [
    (
     "--atkdbg",
     Arg.Unit
       (fun () -> atk_dbg := true
       ),
     " Print standard output information collected during attack MSC generation"
    )
   ]
   (fun s ->
     if !is_input then
       raise (Arg.Bad "error: multiple input files")
     else
       (is_input := true;
	fileName := s))
   ("This is SPAN version 1.6\n"^
   "Usage: span [file.hlpsl]")
);;

let _ =
configure();
(* Title setting *)
Wm.title_set root span_version ;
Wm.resizable_set root ~width:true ~height:true ;

(* Base frame *)
let base = Frame.create root in
(*Wm.resizable_set base ~width:true ~height:true ;*)
pack ~expand:true ~fill:`Both [base];

let buffer_type = ref NoSave in

(* Menu bar *)
let bar = Frame.create ~borderwidth:1 ~relief:`Raised base in 
pack ~fill:`X [bar];

let commandFrame = Frame.create ~relief:`Groove ~borderwidth:2  base in
let commandToolFrame = Frame.create commandFrame in

let optionFrame = Frame.create commandToolFrame  in
let gOptionFrame = ref (new hlpsl2ifOption optionFrame) in

let middleFrame = Canvas.create base in
let outText = Text.create ~width:150 ~height:35 ~relief:`Ridge ~borderwidth:1 ~selectborderwidth:3 ~background:selcol middleFrame in

let sb = Scrollbar.create ~background:blue ~borderwidth:1 ~orient:`Vertical middleFrame in
let scroll_link sb outText =
    Text.configure ~yscrollcommand:(Scrollbar.set sb) outText;
    Scrollbar.configure ~command:(Text.yview outText) sb in
  scroll_link sb outText; 


let save = (fun () -> (
	let tmp = (Text.get outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[])) in
   	let txt = if(tmp<>"" && (String.sub tmp 0 1)<>"\n") then tmp else (String.sub tmp 1 ((String.length tmp)-1)) in
	match !buffer_type with
	  Cas -> (
		saveFile (!fileNameCas) txt;
		casFileLastModification := (Unix.stat !fileNameCas).Unix.st_mtime;
		let fileGenHlpsl = temp_dir^"hlpslGenFile.hlpsl" and fileGenErr = temp_dir^"hlpslGenFile.err" in 
		if (Tk.messageBox ~typ:`Yesno ~message:"Generate HLPSL from CAS+ ?" ~icon:`Info ())="yes" then (
			removeFiles [fileGenHlpsl;fileGenErr];
			ignore(Sys.command(esc^alicebob2hlpsl^" -f "^(Filename.quote fileGenHlpsl)^" --secret "^(Filename.quote (!fileNameCas))^(redirect_towards (Filename.quote fileGenErr))^esc));
			let err_log = readFile fileGenErr in
			if err_log = "" then (
				fileName := fileGenHlpsl;
				synch := true
			) else (
				buffer_type := NoSave ; 
				Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
				Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of CAS+ compilation :\n\n")^err_log) outText
			)
		)) 
	| Hlpsl -> saveFile (!fileName) txt;synch := false 
	| NoSave -> ()
)) in

let fileFrame =  Frame.create ~relief:`Ridge ~borderwidth:1  commandFrame  in
let saveButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"Save file" ~command:save in
let casButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"View CAS+" ~command:(fun () -> (
	Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ;
	Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`Linechar(0, 0),[]) ;
	if Sys.file_exists !fileNameCas then (
		(if not !synch then ignore(Tk.messageBox ~message:"Warning: CAS+ file and HLPSL file are no longer synchronized." ~icon:`Info ()));
		buffer_type := Cas ;
		Button.configure saveButton ~state:`Active;
   		Text.insert ~index:(`End,[]) ~text:(readFile (!fileNameCas)) outText;
		if !casFileLastModification<> ((Unix.stat !fileNameCas).Unix.st_mtime) then (
			casFileLastModification := (Unix.stat !fileNameCas).Unix.st_mtime;
			if (Tk.messageBox ~typ:`Yesno ~message:"Modifications detected\nGenerate HLPSL from CAS+ ?" ~icon:`Info ())="yes" then (
				let fileGenHlpsl = temp_dir^"hlpslGenFile.hlpsl" and fileGenErr = temp_dir^"hlpslGenFile.err" in 
				removeFiles [fileGenHlpsl;fileGenErr];
				ignore(Sys.command(esc^alicebob2hlpsl^" -f "^(Filename.quote fileGenHlpsl)^" --secret "^(Filename.quote (!fileNameCas))^(redirect_towards (Filename.quote fileGenErr))^esc));
				let err_log = readFile fileGenErr in
				if err_log = "" then (
					fileName := fileGenHlpsl;
					synch := true
				) else (
					buffer_type := NoSave ; 
					Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
					Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of CAS+ compilation :\n\n")^err_log) outText
				)
			)
		)
	) else (
		buffer_type := NoSave ;
		Button.configure saveButton ~state:`Disabled;
	)
)) in
let hlpslButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"View HLPSL" ~command:(fun () -> (
	Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ;
	Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`Linechar(0, 0),[]) ;
	if Sys.file_exists !fileName then (
		buffer_type := Hlpsl ;
		Button.configure saveButton ~state:`Active;
   		Text.insert ~index:(`End,[]) ~text:(readFile (!fileName)) outText
	) else (
		buffer_type := NoSave ;
		Button.configure saveButton ~state:`Disabled;
	)
)) in

let simulationButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"Protocol\nsimulation" ~command:(fun () -> 
   if ((!fileName)="")
   then ()
   else (
   let error_sim_file = (temp_dir^(Filename.basename (!fileName))^".sim"^".error") in
   if(Sys.file_exists error_sim_file) then ignore(Sys.remove error_sim_file);
   
   ignore(Sys.command(esc^hlpsl2scenario^" "^(Filename.quote (!fileName))^" 2> "^(Filename.quote error_sim_file)^" &"^esc));

  Unix.sleep 1;
   if(Sys.file_exists error_sim_file) then(
      	let txt = readFile error_sim_file in
	if(txt<>"") then ( 
		buffer_type := NoSave ; 
		Button.configure saveButton ~state:`Disabled;
		Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of protocol simulation :\n\n")^txt) outText;
		);
   	ignore(Sys.remove error_sim_file);
   	)
   )) in
let intruderButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"Intruder\nsimulation" ~command:(fun () -> 
   if ((!fileName)="")
   then ()
   else (
   let error_sim_file = (temp_dir^(Filename.basename (!fileName))^".sim"^".error") in
   if(Sys.file_exists error_sim_file) then ignore(Sys.remove error_sim_file);
   
   ignore(Sys.command(esc^hlpsl2scenario^" --intruder "^(Filename.quote (!fileName))^" 2> "^(Filename.quote error_sim_file)^" &"^esc));

   Unix.sleep 1;
   if(Sys.file_exists error_sim_file) then(
      	let txt = readFile error_sim_file in
	if(txt<>"") then ( 
		buffer_type := NoSave ; 
		Button.configure saveButton ~state:`Disabled;
		Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of protocol simulation :\n\n")^txt) outText;
		);
   	ignore(Sys.remove error_sim_file);
   	)
   )) in
(** attackButton makes it possible to call attack trace generation on the current HLPSL specification
   after having selected one of the following AVISPA verification tools : OFMC, ATSE ;
   in case an attack is actually found, the generated trace is launched through the simulator *)
let attackButton = Button.create  ~width:10 ~height:2 ~borderwidth:1 fileFrame ~text:"Attack\nsimulation" ~command:(fun () -> 
   if ((!fileName)="")
   then ()
   else (
			(* call verification tool with appropriate options and store the output from the verification tool *)
			let (out_if, out_tool, err_if, err_tool) = (!gOptionFrame)#call(!fileName) in
			if(not (Sys.file_exists err_tool && ((readFile err_tool)<>"")))
			(* verification worked fine *)
			then(
         (* defines temporary files for trace generation outputs *)
         let genFile = temp_dir^"auto_gen.trc"		(* generated trace *)
         and ifErrFile = temp_dir^"auto_gen.if.err"	(* errors related to translation from HLPSL to IF *)
         and genErrFile = temp_dir^"auto_gen.err" in	(* trace generation execution errors *)
         removeFiles [genFile ; ifErrFile ; genErrFile];
         (* runs trace generation and stores exit status *)
			   let mode = match !atk_dbg with true -> " --dbg" | false -> ""
   			 and input = " --in "^(Filename.quote out_tool)
		   	 and output = " --out "^(Filename.quote genFile) in
         let code = Sys.command(esc^report2trace^mode^input^output^" --err "^(Filename.quote genErrFile)^" "^(Filename.quote !fileName)^" 2> "^(Filename.quote ifErrFile)^esc) in
         Unix.sleep 1;
				 removeFiles [out_tool];
         (* displays execution warnnings and errors from error output if there are any *)
         (* previous error messages flag *)
         let err_msg = ref false in
         if (Sys.file_exists ifErrFile)
         then (
            let (txt,lig) = readFile_and_give_line_number ifErrFile in
					  if(txt<>"" && lig>2)	(* first two lines always exist, are therefore to be ignored *)
            then ( 
               buffer_type := NoSave ; 
               Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
               Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of attack trace generation :\n\n")^txt) outText;
               err_msg := true ;
   	        );
         );
         (* displays execution warnnings and errors from error report if there are any *)
         if (Sys.file_exists genErrFile)
         then (
         	  let txt = readFile genErrFile in
            if(txt<>"")
            then( 
               buffer_type := NoSave ; 
               Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		           Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of attack trace generation :\n\n")^txt) outText;
               err_msg := true ;
	          );
         );
         (* displays MSC iff trace generation worked correctly *)
         if ( code = 0 )
         (* normal termination of attack trace generation *)
         then (
            (* displays a waiting message *)
            buffer_type := NoSave;
	    Button.configure saveButton ~state:`Disabled;
            if (not(!err_msg)) then ( Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ) ;
            Text.insert ~index:(`End,[]) ~text:("\n\nLoading trace, please wait...\n") outText;
            (* defines a temporary file to store error output for protocol simulation *)
            let error_sim_file = (temp_dir^(Filename.basename (!fileName))^".sim"^".error") in
            removeFiles [error_sim_file];
            (* runs protocol simulation *)
            ignore(Sys.command(esc^hlpsl2scenario^" --intruder"^" --trace "^(Filename.quote genFile)^" "^(Filename.quote (!fileName))^" 2> "^(Filename.quote error_sim_file)^" &"^esc));
            (* displays execution warnings and errors if there are any *)
            Unix.sleep 1;
            if(Sys.file_exists error_sim_file)
            then(
               let txt = readFile error_sim_file in
               if(txt<>"")
               then ( 
                  buffer_type := NoSave; 
                  Button.configure saveButton ~state:`Disabled;
                  Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
                  Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of attack simulation :\n\n")^txt) outText;
               );
	          );
            removeFiles [error_sim_file];
         )
		   	 (* abnormal termination of attack trace generation *)
         else (
            buffer_type := NoSave; 
            Button.configure saveButton ~state:`Disabled;
            if (not(!err_msg)) then ( Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ) ;
            Text.insert ~index:(`End,[]) ~text:("\n\n%% report2trace terminated abnormally...\n\n\t\tNot launching simulation\n") outText;
         );
				 removeFiles [ifErrFile ; genErrFile];
			)
			(* verification showed errors *)
      else(
         let txt = readFile err_tool in
         buffer_type := NoSave ; 
         Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		     Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of verification for attack trace generation :\n\n")^txt) outText;
			);
			removeFiles [out_if ; err_if ; err_tool];
   )
) in
(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
let tool = Frame.create commandToolFrame in
let tool1 = Frame.create tool in
let labelHlpsl2if = Label.create ~text:"HLPSL" ~width:8 ~height:2 tool1 in
let tool2 = Frame.create tool in

let toolHlpsl2if = Checkbutton.create ~text:"HLPSL2IF" ~width:8 ~height:2 ~borderwidth:1 ~background:blue  ~indicatoron:false  ~selectcolor:clicol  tool2 in
let tool3 = Frame.create tool in
let labelIf = Label.create ~text:"IF" ~width:8 ~height:2 tool3 in
let tool4 = Frame.create tool in
let toolOFMC = Checkbutton.create ~text:"OFMC" ~width:8 ~height:2 ~borderwidth:1 ~background:blue  ~indicatoron:false ~selectcolor:clicol  tool4 in
let toolATSE = Checkbutton.create ~text:"ATSE" ~width:8 ~height:2 ~borderwidth:1 ~background:blue  ~indicatoron:false ~selectcolor:clicol  tool4 in
let toolSATMC = Checkbutton.create ~text:"SATMC" ~width:8 ~height:2 ~borderwidth:1 ~background:blue  ~indicatoron:false ~selectcolor:clicol  tool4 in
let toolTA4SP = Checkbutton.create ~text:"TA4SP" ~width:8 ~height:2 ~borderwidth:1 ~background:blue  ~indicatoron:false ~selectcolor:clicol  tool4 in

let unchoose = List.iter (fun t-> Checkbutton.configure t ~state:`Normal ~takefocus:false; Checkbutton.deselect t) in
Checkbutton.configure toolHlpsl2if ~command:(fun () -> ignore(unchoose [toolOFMC ; toolATSE ; toolSATMC ; toolTA4SP]; (!gOptionFrame)#destroy; gOptionFrame := new hlpsl2ifOption optionFrame ; (!gOptionFrame)#init(attackButton); ););
Checkbutton.configure toolOFMC ~command:(fun () -> ignore(unchoose [toolHlpsl2if ; toolATSE ; toolSATMC ; toolTA4SP]; (!gOptionFrame)#destroy; gOptionFrame := new ofmcOption optionFrame; (!gOptionFrame)#init(attackButton); ););
Checkbutton.configure toolATSE ~command:(fun () -> ignore(unchoose [toolHlpsl2if ; toolOFMC ; toolSATMC ; toolTA4SP]; (!gOptionFrame)#destroy; gOptionFrame := new atseOption optionFrame; (!gOptionFrame)#init(attackButton); ););
Checkbutton.configure toolSATMC ~command:(fun () -> ignore(unchoose [toolHlpsl2if ; toolOFMC ; toolATSE ; toolTA4SP]; (!gOptionFrame)#destroy; gOptionFrame := new satmcOption optionFrame; (!gOptionFrame)#init(attackButton); );) ;
Checkbutton.configure toolTA4SP ~command:(fun () -> ignore(unchoose [toolHlpsl2if ; toolOFMC ; toolATSE ; toolSATMC]; (!gOptionFrame)#destroy; gOptionFrame := new ta4spOption optionFrame; (!gOptionFrame)#init(attackButton); ););
let blankFrame1 = Frame.create  tool in
let toolLabel = Label.create ~text:"Tools" ~background:blue ~width:20 ~height:2 blankFrame1 in
let blankFrame2 = Frame.create  ~width:120 ~height:50 tool in


let executeFrame = Frame.create ~relief:`Solid ~borderwidth:1  commandToolFrame in
let executeLabel1 = Label.create ~text:"Choose Tool option and" executeFrame in
let executeLabel2 = Label.create ~text:"press execute" executeFrame in
let executeButton = Button.create ~background:clicol ~text:"Execute" ~width:16 ~height:1 ~borderwidth:1 
										~command:(fun () -> (
											     match (!fileName) 
											     with "" -> ()
											     | _ -> (
												buffer_type := NoSave ;
												Button.configure saveButton ~state:`Disabled;
												(!gOptionFrame)#execution (!fileName) outText))) executeFrame in
let typesCAS = [
     {typename="CAS+ Files";	   extensions=[".cas"];mactypes=["TEXT"]};
     {typename="Text Files";       extensions=[".txt"; ".text"]; mactypes=["TEXT"]};
     {typename="All Files";        extensions=["*"];       mactypes=[""]}
 ] 
and typesHLPSL = [
     {typename="Hlpsl Files";      extensions=[".hlpsl"]; mactypes=["TEXT"]};
     {typename="Text Files";       extensions=[".txt"; ".text"]; mactypes=["TEXT"]};
     {typename="All Files";        extensions=["*"];       mactypes=[""]}
 ] 
and types = [
	{typename="CAS+ Files";	   extensions=[".cas"];mactypes=["TEXT"]};
     {typename="Hlpsl Files";      extensions=[".hlpsl"]; mactypes=["TEXT"]};
     {typename="Text Files";       extensions=[".txt"; ".text"]; mactypes=["TEXT"]};
     {typename="All Files";        extensions=["*"];       mactypes=[""]}
 ] in

(* Menu and Menubutton *)
let meb = Menubutton.create ~text:"File" ~font:"guifont" ~activebackground:blue bar in
let men = Menu.create meb ~activebackground:blue ~borderwidth:1 ~relief:`Groove in

Menu.add_command ~label:"Open CAS+ File..." ~command:(fun () -> (
	let file = Tk.getOpenFile() 	~defaultextension:".cas" 
					~filetypes:typesCAS 
					~initialdir: (if !fileNameCas<>"" then (Filename.dirname (!fileNameCas)) else (Filename.dirname (!fileName)))
					~title:"Open a protocol file" in if Sys.file_exists file then (
		fileNameCas := file;
		fileName := "";
		casFileLastModification := (Unix.stat !fileNameCas).Unix.st_mtime;
		buffer_type := Cas;
		Button.configure saveButton ~state:`Active;
		Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		Wm.title_set root (span_version^" : "^(Filename.basename file));
		Text.insert ~index:(`End,[]) ~text:(readFile file) outText;
		
		let fileGenHlpsl = temp_dir^"hlpslGenFile.hlpsl" and fileGenErr = temp_dir^"hlpslGenFile.err" in 
		if (Tk.messageBox ~typ:`Yesno ~message:"Generate HLPSL from CAS+ ?" ~icon:`Question ())="yes" then (
			removeFiles [fileGenHlpsl;fileGenErr];
			ignore(Sys.command(esc^alicebob2hlpsl^" -f "^(Filename.quote fileGenHlpsl)^" --secret "^(Filename.quote (!fileNameCas))^(redirect_towards (Filename.quote fileGenErr))^esc));
			let err_log = readFile fileGenErr in
			if err_log = "" then (
				fileName := fileGenHlpsl;
				synch := true
			) else (
				buffer_type := NoSave ; 
				Button.configure saveButton ~state:`Disabled;
				Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
				Text.insert ~index:(`End,[]) ~text:(("\n----- Output error of CAS+ compilation :\n\n")^err_log) outText
			)
		)		
))) men;

Menu.add_command ~label:"Open HLPSL File..." ~command:(fun () -> (
	let file = Tk.getOpenFile() 	~defaultextension:".hlpsl" 
					~filetypes:typesHLPSL 
					~initialdir:(if !fileNameCas<>"" then (Filename.dirname (!fileNameCas)) else (Filename.dirname (!fileName))) 
					~title:"Open a protocol file" in if Sys.file_exists file then (
		fileName := file;
		fileNameCas := "";
		casFileLastModification := 0.0;
		buffer_type := Hlpsl;
		Button.configure saveButton ~state:`Active;
		Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]);
		Wm.title_set root (span_version^" : "^(Filename.basename file));
		Text.insert ~index:(`End,[]) ~text:(readFile file) outText
))) men;
Menu.add_command ~label:"Save" ~command:save men;
Menu.add_command ~label:"Save as..." ~command:(fun () -> 
   (let file= (!fileName) in
    fileName:=Tk.getSaveFile() 	~defaultextension:".hlpsl" 
				~filetypes:types 
				~initialdir:(if !fileNameCas<>"" then (Filename.dirname (!fileNameCas)) else (Filename.dirname (!fileName))) 
				~title:"Save a protocol file";
    if (!fileName)="" then fileName:=file else ();
    let offset = 1+(String.rindex (!fileName) '/') in
    Wm.title_set root (span_version^" : "^(String.sub (!fileName) offset ((String.length (!fileName))-offset) ));
   let tmp = (Text.get outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[])) in
   let txt = if(tmp<>"" && (String.sub tmp 0 1)<>"\n") then tmp else (String.sub tmp 1 ((String.length tmp)-1)) in
   if ((!fileName)="")
   then ()
   else saveFile (!fileName) txt )) men;

Menu.add_separator men;
(** for change fonts size *)
Menu.add_command ~label:"Font size..." ~command:(fun () ->
	let string_to_int s = (try 
    				int_of_string s
  				with _ -> -1 ) in
	let file_config_command = (path_span^"/font_size.ini") in

	let root_font_size =  Toplevel.create ~borderwidth:1 ~relief:`Groove root in
	Wm.title_set root_font_size ("Font size ...");
	let frame_font_size1 = Frame.create ~borderwidth:1 ~relief:`Sunken root_font_size in
	let frame_font_size2 = Frame.create ~borderwidth:1 ~relief:`Sunken root_font_size in
	let label_font_size = Label.create ~text:"Font size : " ~width:15 ~height:2 frame_font_size1 in
	let entry_font_size = Entry.create ~width:20 ~borderwidth:1 frame_font_size1 ~background:selcol in
	Entry.insert entry_font_size ~index:`End ~text:(string_of_int (!current_font_size)) ;
	let button_ok_font_size = 	Button.create ~text:"OK" 
						~command:(fun () -> 	let tmp_size =  
    										string_to_int (Entry.get entry_font_size)
									in
									if(tmp_size>0 && tmp_size<>(!current_font_size))
									then (	if(Sys.file_exists file_config_command) 
										then ignore(Sys.remove file_config_command);
										current_font_size := tmp_size;
										set_size_of_all_fonts tmp_size;
										saveFile file_config_command (string_of_int tmp_size)
										);
										
									destroy label_font_size ; destroy entry_font_size ; 
									destroy entry_font_size ; destroy root_font_size) 
					~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_font_size2  in
	let button_cancel_font_size = Button.create ~text:"Cancel" 
						~command:(fun () -> 	destroy label_font_size ; destroy entry_font_size ; 
									destroy button_ok_font_size ; destroy entry_font_size ; 
									destroy root_font_size ) 
						~width:19 ~height:1 ~borderwidth:1 ~relief:`Groove frame_font_size2  in
	pack ~side:`Top ~fill:`Y [frame_font_size1];
	pack ~side:`Left ~fill:`Y [label_font_size];
	pack ~side:`Left ~fill:`Y [entry_font_size];
	pack ~side:`Top ~fill:`Y [frame_font_size2];
	pack ~side:`Left ~fill:`Y [button_ok_font_size];
	pack ~side:`Left ~fill:`Y [button_cancel_font_size];
	
) men;
Menubutton.configure ~menu:men meb;
pack ~side:`Left [meb];

Menu.add_separator men;
Menu.add_command ~label:"Quit" ~command:(fun () -> closeTk (); exit 0) men;
Menubutton.configure ~menu:men meb;

if ((!fileName)<>"")
then (   Wm.title_set root (span_version^" : "^(Filename.basename (!fileName)));
	 Text.delete outText ~start:(`Linechar(0, 0),[`Linestart]) ~stop:(`End,[]) ;
	 Text.insert ~index:(`End,[]) ~text:(readFile (!fileName)) outText;)
else();


    pack ~side:`Left [meb];

    pack ~side:`Left ~fill:`Both  [tool];
    pack ~side:`Top [blankFrame1];
    pack ~side:`Top [toolLabel];
    pack ~side:`Top [tool1];
    pack ~side:`Bottom [labelHlpsl2if];
    pack ~side:`Top [tool2];
    pack ~side:`Left [toolHlpsl2if];
    pack ~side:`Top [tool3];
    pack ~side:`Left [labelIf];
    pack ~side:`Top [tool4];
    pack ~side:`Left [toolOFMC];
    pack ~side:`Left [toolATSE];
    pack ~side:`Left [toolSATMC];
    pack ~side:`Left [toolTA4SP];
    
    pack ~side:`Left ~fill:`X [executeFrame];
    pack ~side:`Top [executeLabel1];
    pack ~side:`Top [executeLabel2];
    pack ~side:`Top [executeButton];
    
    pack ~side:`Left  [optionFrame];
    pack ~side:`Bottom [blankFrame2];
    
        
    pack ~side:`Bottom ~fill:`Both   [commandToolFrame];
    
    pack ~side:`Left [saveButton];
    pack ~side:`Left [casButton];
    pack ~side:`Left [hlpslButton];
    pack ~side:`Left [simulationButton];
    pack ~side:`Left [intruderButton];
    pack ~side:`Left [attackButton]; 
    pack ~side:`Bottom [fileFrame];
    
    pack ~side:`Bottom ~fill:`Both    [commandFrame];
    

    pack ~side:`Top ~fill:`Both ~expand:true [middleFrame];
    pack ~side:`Right ~fill:`Y [sb];
    pack ~side:`Left ~fill:`Both ~expand:true [outText];
Printexc.print mainLoop();
